!> \file
!! Contains subroutine conduction_ex()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CONDUCTION_EX
!
!> Computes the explicit thermal gas conduction.
!<
subroutine conduction_ex(imin,imax,jmin,jmax,kmin,kmax)
  
  use var
  use var_loc
  use varold
  use geom
  use divers
  use parameters, only : ndim,nx
  use cl
  use para      , only : mype
  use unites    , only : uma,kb,kg,metre,seconde,degre

  implicit none
  
  integer, intent(in)     :: imin,imax,jmin,jmax,kmin,kmax
  integer, dimension(3,3) :: kron
  integer, dimension(6  ) :: icl_loc
  integer                 :: nnx,nny,nnz,nn
  integer                 :: i,j,k,idim,id
  integer                 :: ip1,jp1,kp1,im1,jm1,km1
  integer                 :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  integer                 :: ixmin,ixmax,iymin,iymax,izmin,izmax
  
  real, dimension(:,:,:,:  ), allocatable :: k_gradT,Bloc
  real, dimension(:,:,:    ), allocatable :: kappa     ! Dt E = Div(kappa Grad T)
  real                                    :: densite,temp,cal_kappa_con
  real                                    :: kap,flux_ener_tot
  real                                    :: dElim,xlim,diffmax,xx

  if (verbose) write(*,*) 'Start of conduction'

  ! some checks
  call check_size(imin,imax,jmin,jmax,kmin,kmax)

  ! Define the local boundaries icl_loc for the sub-domain boundaries
  icl_loc = 0

  if(imin.eq.1    ) icl_loc(1)=icl(1)
  if(imax.eq.nx(1)) icl_loc(2)=icl(2)
  if(jmin.eq.1    ) icl_loc(3)=icl(3)
  if(jmax.eq.nx(2)) icl_loc(4)=icl(4)
  if(kmin.eq.1    ) icl_loc(5)=icl(5)
  if(kmax.eq.nx(3)) icl_loc(6)=icl(6)


  ! Allocate arrays
  nnx = imax - imin + 1 ; nny =  jmax - jmin + 1 ; nnz = kmax - kmin + 1
  if(ndim.eq.1) then
     nxmin = -1 ; nxmax = nnx+2
     nymin =  1 ; nymax = 1
     nzmin =  1 ; nzmax = 1
     ixmin =  0 ; ixmax = nnx+1
     iymin =  1 ; iymax = 1
     izmin =  1 ; izmax = 1
  else if (ndim == 2) then
     nxmin = -1 ; nxmax = nnx+2
     nymin = -1 ; nymax = nny+2
     nzmin =  1 ; nzmax = 1
     ixmin =  0 ; ixmax = nnx+1
     iymin =  0 ; iymax = nny+1
     izmin =  1 ; izmax = 1
  else if (ndim == 3) then
     nxmin = -1 ; nxmax = nnx+2
     nymin = -1 ; nymax = nny+2
     nzmin = -1 ; nzmax = nnz+2
     ixmin =  0 ; ixmax = nnx+1
     iymin =  0 ; iymax = nny+1
     izmin =  0 ; izmax = nnz+1
  endif

  kron = 0
  do idim=1,3
     kron(idim,idim) = 1
  enddo

  nn  = max(nnx,nny,nnz)
  allocate(xloc(-1:nn+3,ndim)) 

  allocate(rholoc (nxmin:nxmax,nymin:nymax,nzmin:nzmax       ))
  allocate(Eloc   (nxmin:nxmax,nymin:nymax,nzmin:nzmax       ),Tloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ))
  allocate(rhouloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim),uloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim))

  allocate(k_gradT(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim))
  allocate(kappa  (nxmin:nxmax,nymin:nymax,nzmin:nzmax       ))


  if(verbose) write(*,*) 'filling the sub-domain'

  ! Filling the sub-domain with grid variables
  if (ndim == 1) then
       rholoc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =  rhoold(imin-2:imax+2,1:1,1:1       )
         Eloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =    Eold(imin-2:imax+2,1:1,1:1       )
      rhouloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) = rhouold(imin-2:imax+2,1:1,1:1,1:ndim)
 else if (ndim ==2) then
      rholoc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =  rhoold(imin-2:imax+2,jmin-2:jmax+2,1:1       )
        Eloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =    Eold(imin-2:imax+2,jmin-2:jmax+2,1:1       )
     rhouloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) = rhouold(imin-2:imax+2,jmin-2:jmax+2,1:1,1:ndim)
 else if (ndim ==3) then
      rholoc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =  rhoold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2       )
        Eloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =    Eold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2       )
     rhouloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) = rhouold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2,1:ndim)
 endif

  if (allocated(B)) then
     allocate(Bloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,3))
     if (ndim == 1) then
        Bloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:3) =    Bold(imin-2:imax+2,1:1,1:1,1:3)
     else if (ndim ==2) then
        Bloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:3) =    Bold(imin-2:imax+2,jmin-2:jmax+2,1:1,1:3)
     else if (ndim ==3) then
        Bloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:3) =    Bold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2,1:3)
     endif
  endif


                xloc(-1:nnx+3,1) = x(imin-2:imax+3,1)   
 if (ndim.gt.1) xloc(-1:nny+3,2) = x(jmin-2:jmax+3,2)   
 if (ndim.gt.2) xloc(-1:nnz+3,3) = x(kmin-2:kmax+3,3)   


  ! Compute the grid geometry
  if(verbose) write(*,*) 'computing the grid geometry'
  call cal_geom(nnx,nny,nnz)

  ! compute T
  if(verbose) write(*,*) 'computing T'

  ! Compute Ein and store it in Eloc
  Eloc = Eloc - 0.5*sum(rhouloc**2,4)/rholoc
  if (allocated(B)) Eloc = Eloc - 0.5*sum(Bloc**2,4)

  call temperature(rholoc,Eloc,Tloc,nxmin,nxmax,nymin,nymax,nzmin,nzmax)
  if(minval(Tloc(:,:,:))<0.) then
     do k = nzmin,nzmax
     do j = nymin,nymax
     do i = nxmin,nxmax
        if(Tloc(i,j,k)<0.) then
           write(*,101) i,j,k,Tloc(i,j,k)
101        format('Tloc negative (i,j,k), T=',I5,I5,I5,2x,E10.3)
           call temperature(rholoc(i,j,k),1.e-8*Eloc(i,j,k),xx,1,1,1,1,1,1)
           Tloc(i,j,k) = xx
        endif
     enddo
     enddo
     enddo
  endif

  ! compute kappa
  if(verbose) write(*,*) 'computing kappa'

  do k = nzmin,nzmax
  do j = nymin,nymax
  do i = nxmin,nxmax

     Temp    =   Tloc(i,j,k)
     Densite = rholoc(i,j,k)
     kappa(i,j,k) =  cal_kappa_con(Densite,Temp)

  enddo
  enddo
  enddo
    
  ! compute k_gradT
  do idim = 1,ndim
  
     do k = izmin,izmax
     do j = iymin,iymax
     do i = 1,nnx+1

        im1 = i - kron(1,idim) ; jm1 = j - kron(2,idim) ; km1 = k - kron(3,idim)

        if(idim.eq.1) then
           id = i
        else if(idim.eq.2) then
           id = j
        else if(idim.eq.3) then
           id = k
        endif

        kap = 0.5*(kappa(im1,jm1,km1) + kappa(i,j,k)) 
        
        k_gradT(i,j,k,idim) = kap * (Tloc(i,j,k) - Tloc(im1,jm1,km1)) / (xloc(id,idim)-xloc(id-1,idim))  
        
     enddo
     enddo
     enddo

  enddo

  ! overwrite if imposed boundary inflow is present
  ! NEEDS TO BE DONE....

  do k = 1,nnz
  do j = 1,nny
  do i = 1,nnx

     flux_ener_tot  = 0.
     do idim = 1,ndim
        ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)
        flux_ener_tot = flux_ener_tot + k_gradT(ip1,jp1,kp1,idim)*ds(ip1,jp1,kp1,idim) - k_gradT(i,j,k,idim)*ds(i,j,k,idim)
     enddo
     flux_ener_tot = flux_ener_tot * dt / dv(i,j,k)
     dElim = 0.9 * Eloc(i,j,k)
     xlim = -flux_ener_tot/dElim
     if (xlim > 1.) then
        do idim = 1,ndim
           ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)
           k_gradT(ip1,jp1,kp1,idim) = k_gradT(ip1,jp1,kp1,idim) / xlim
           k_gradT(i  ,j  ,k  ,idim) = k_gradT(i  ,j  ,k  ,idim) / xlim
        enddo
     endif
        
  enddo
  enddo
  enddo


  diffmax = 0.
  
  do idim = 1,ndim

     do k = 1,nnz
     do j = 1,nny
     do i = 1,nnx

        ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)

        flux_ener_tot = k_gradT(ip1,jp1,kp1,idim)*ds(ip1,jp1,kp1,idim) - k_gradT(i,j,k,idim)*ds(i,j,k,idim)

        Eloc(i,j,k  ) = Eloc(i,j,k  ) + flux_ener_tot * dt / dv(i,j,k)
        
        diffmax = max(diffmax,abs(flux_ener_tot *dt / dv(i,j,k)/Eloc(i,j,k)) )

     enddo
     enddo
     enddo

  enddo


  do k = 1,nnz
  do j = 1,nny
  do i = 1,nnx

     if (Eloc(i,j,k) < 0.) then
        write(*,*) '*******************************'
        write(*,*) 'i,j,k = ',i,j,k
        write(*,*) 'Eloc(i,j,k) =',Eloc(i,j,k  )
        write(*,*)
        write(*,*) 'T(i-1,j+-1) =',Tloc(i-1,j-1,k),Tloc(i-1,j,k),Tloc(i-1,j+1,k)
        write(*,*) 'T(i  ,j+-1) =',Tloc(i  ,j-1,k),Tloc(i  ,j,k),Tloc(i  ,j+1,k)
        write(*,*) 'T(i+1,j+-1) =',Tloc(i+1,j-1,k),Tloc(i+1,j,k),Tloc(i+1,j+1,k)
        write(*,*) 
        write(*,*) 'flux= ', -k_gradT(i,j,k,1)*ds(i,j,k,1)* dt / dv(i,j,k), k_gradT(i+1,j  ,k,1)*ds(i+1,j  ,k,1)* dt / dv(i,j,k)
        write(*,*) 'flux= ', -k_gradT(i,j,k,2)*ds(i,j,k,2)* dt / dv(i,j,k), k_gradT(i  ,j+1,k,2)*ds(i  ,j+1,k,1)* dt / dv(i,j,k)
        write(*,*) '*******************************'
     endif

  enddo
  enddo
  enddo

!  write(*,*) 'diffmax % = ',diffmax*100.
!  read(*,*)
  
  ! Restore Eloc with Etot
  Eloc = Eloc + 0.5*sum(rhouloc**2,4)/rholoc
  if (allocated(B)) Eloc = Eloc + 0.5*sum(Bloc**2,4)

  ! Re-actualise the energy array
  E(imin:imax,jmin:jmax,kmin:kmax) = Eloc(1:nnx,1:nny,1:nnz)

  ! Deallocate the local arrays
  ! geom
  deallocate(dx,dxc)
  deallocate(ds,dv)
  ! the rest
  deallocate(xloc,xcloc)
  deallocate( rholoc,Eloc)
  deallocate(rhouloc,uloc)

  deallocate(Tloc, k_gradT,kappa)

  if (allocated(B)) deallocate(Bloc)

  if (verbose) write(*,*) 'End of conduction'

  return

end subroutine conduction_ex
