!> \file
!! Contains subroutines limites_gra() and cal_barycentre()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine LIMITES_GRA
!
!> Defines the boundary conditions for the gravity.
!<
subroutine limites_gra(Tab)
  
  use parameters, only : ndim,nx,nxmin,nxmax
  use divers    , only : verbose
  use gravity
  use unites    , only : G
  use geom      , only : x
  use cl        , only : icl

  implicit none

  integer, dimension(1:6) :: face
  integer                 :: i,j,k,iface

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: Tab
  real :: xbar,ybar,zbar,pbar,xc,yc,zc

  xbar = 0. ; ybar = 0. ; zbar = 0. ; pbar = 0.
  
  ! ISOLE = 0 : PERIODIC
  ! ISOLE = 1 : ISOLATED SYSTEM
  
  if(verbose) write(*,*) 'Start of limites_gra'

  ! Order in which the faces are treated
  face(1) = 1
  face(2) = 2
  face(3) = 3
  face(4) = 4
  face(5) = 5
  face(6) = 6
 
  if (ISOLE == 1) then
     
     xbar = 0. ; ybar = 0. ; zbar = 0. ; pbar = 0.
     call cal_barycentre(xbar,ybar,zbar,pbar)

  endif

  do iface = 1,2*ndim

     if (ICL(face(iface)).ne.0) then

     ! 1)----------------------------------------

        if (face(iface) == 1) then

           if (ISOLE == 1) then ! isolated system
              do i = -1,0
                 do j = nxmin(2),nxmax(2)
                    do k = nxmin(3),nxmax(3)
             
                       xc = 0.5*(x(i,1)+x(i+1,1))
                       yc = 0.5*(x(j,2)+x(j+1,2))
                       zc = 0.5*(x(k,3)+x(k+1,3))

                       Tab(i,j,k) = -G*pbar/sqrt((xc-xbar)**2+(yc-ybar)**2+(zc-zbar)**2)

                    enddo
                 enddo
              enddo

#if PARA==0

           else if (ISOLE == 0) then ! periodic

              do i = -1,0
                 
                 Tab(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)) = Tab(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3))
                 
              enddo

#endif
    
           endif

        endif

    ! 2)----------------------------------------
        
        if (face(iface) == 2) then

           if (ISOLE == 1) then ! isolated system

              do i = nx(1)+1,nx(1)+2
                 do j = nxmin(2),nxmax(2)
                    do k = nxmin(3),nxmax(3)
       
                       xc = 0.5*(x(i,1)+x(i+1,1))
                       yc = 0.5*(x(j,2)+x(j+1,2))
                       zc = 0.5*(x(k,3)+x(k+1,3))

                       Tab(i,j,k) = -G*pbar/sqrt((xc-xbar)**2+(yc-ybar)**2+(zc-zbar)**2)
       
                    enddo
                 enddo
              enddo

#if PARA==0

           else if (ISOLE == 0) then ! periodic 
              
              do i = 1,2
                 
                 Tab(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)) = Tab(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3))

              enddo

#endif

           endif
           
        endif
    
    !3)----------------------------------------

        if (face(iface) == 3) then

           if (ISOLE == 1) then ! isolated system

              do i = nxmin(1),nxmax(1)
                 do j = -1,0
                    do k = nxmin(3),nxmax(3)
       
                       xc = 0.5*(x(i,1)+x(i+1,1))
                       yc = 0.5*(x(j,2)+x(j+1,2))
                       zc = 0.5*(x(k,3)+x(k+1,3))

                       Tab(i,j,k) = -G*pbar/sqrt((xc-xbar)**2+(yc-ybar)**2+(zc-zbar)**2)
       
                    enddo
                 enddo
              enddo

#if PARA==0
              
           else if (ISOLE == 0) then ! periodic 

              do i = -1,0

                 Tab(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)) = Tab(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3))

              enddo

#endif

           endif
    
        endif

    !4)----------------------------------------

        if (face(iface) == 4) then

           if (ISOLE == 1) then ! isolated system
     
              do i = nxmin(1),nxmax(1)
                 do j = nx(2)+1,nx(2)+2
                    do k = nxmin(3),nxmax(3)
       
                       xc = 0.5*(x(i,1)+x(i+1,1))
                       yc = 0.5*(x(j,2)+x(j+1,2))
                       zc = 0.5*(x(k,3)+x(k+1,3))

                       Tab(i,j,k) = -G*pbar/sqrt((xc-xbar)**2+(yc-ybar)**2+(zc-zbar)**2)
       
                    enddo
                 enddo
              enddo

#if PARA==0

           else if (ISOLE == 0) then ! periodic 

              do i = 1,2
                 
                 Tab(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)) = Tab(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3))
             
              enddo

#endif

           endif

        endif

    !5)----------------------------------------

        if (face(iface) == 5) then

           if (ISOLE == 1) then ! isolated system
          
              do i = nxmin(1),nxmax(1)
                 do j = nxmin(2),nxmax(2)
                    do k = -1,0
       
                       xc = 0.5*(x(i,1)+x(i+1,1))
                       yc = 0.5*(x(j,2)+x(j+1,2))
                       zc = 0.5*(x(k,3)+x(k+1,3))

                       Tab(i,j,k) = -G*pbar/sqrt((xc-xbar)**2+(yc-ybar)**2+(zc-zbar)**2)
       
                    enddo
                 enddo
              enddo

#if PARA==0

           else if (ISOLE == 0) then ! periodic 

              do i = -1,0

                 Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i) = Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i)

              enddo

#endif
              
           endif

        endif

    !6)----------------------------------------

        if (face(iface) == 6) then

           if (ISOLE == 1) then ! isolated system
       
              do i = nxmin(1),nxmax(1)
                 do j = nxmin(2),nxmax(2)
                    do k = nx(3)+1,nx(3)+2
       
                       xc = 0.5*(x(i,1)+x(i+1,1))
                       yc = 0.5*(x(j,2)+x(j+1,2))
                       zc = 0.5*(x(k,3)+x(k+1,3))

                       Tab(i,j,k) = -G*pbar/sqrt((xc-xbar)**2+(yc-ybar)**2+(zc-zbar)**2)
       
                    enddo
                 enddo
              enddo

#if PARA==0
          
           else if (ISOLE == 0) then ! periodic 

              do i = 1,2

                 Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i) = Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i)

              enddo
          
#endif

           endif

        endif
    
     endif

  enddo

  if(verbose) write(*,*) 'End of limites_gra'

  return

end subroutine limites_gra

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CAL_BARYCENTRE
!
!> Computes the barycentre of the grid.
!<
subroutine cal_barycentre(xbar,ybar,zbar,pbar)
  
  use var
  use parameters
  use gravity
  use divers    , only : verbose
  use geom      , only : x
  
  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info
#endif

  integer :: i,j,k
  
  real, dimension(3) :: dx
  real               :: xbar,ybar,zbar,pbar,xp_loc,yp_loc,zp_loc,p_loc,xp,yp,zp,p,dv

  if (verbose) write(*,*) 'Start of cal_barycentre'

  xbar   = 0. ; ybar   = 0. ; zbar   = 0. ; pbar  = 0.
  xp_loc = 0. ; yp_loc = 0. ; zp_loc = 0. ; p_loc = 0.

  xp = 0. ; yp = 0. ; zp = 0.
  p  = 0. ; dx = 0. ; dv = 0.

  do k = 1,nx(3)
     do j = 1,nx(2)
        do i = 1,nx(1)

           dx(1) = x(i+1,1)-x(i,1)
           dx(2) = x(j+1,2)-x(j,2)
           dx(3) = x(k+1,3)-x(k,3)

           dv = dx(1)*dx(2)*dx(3)

           xp_loc = xp_loc+0.5*(x(i,1)+x(i+1,1))*rho(i,j,k)*dv
           yp_loc = yp_loc+0.5*(x(j,2)+x(j+1,2))*rho(i,j,k)*dv
           zp_loc = zp_loc+0.5*(x(k,3)+x(k+1,3))*rho(i,j,k)*dv
           p_loc  = p_loc+rho(i,j,k)*dv

        enddo
     enddo
  enddo

#if PARA==1

  info = 0
  call MPI_ALLREDUCE(xp_loc, xp,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  call MPI_ALLREDUCE(yp_loc, yp,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  call MPI_ALLREDUCE(zp_loc, zp,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  call MPI_ALLREDUCE(p_loc, p,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
 
  xbar = xp/p
  ybar = yp/p
  zbar = zp/p
  pbar = p
  
#else

  xbar = xp_loc/p_loc
  ybar = yp_loc/p_loc
  zbar = zp_loc/p_loc
  pbar = p_loc

#endif

  if (verbose) write(*,*) 'Fin de cal_barycentre'

  return

end subroutine cal_barycentre
