!> \file
!! Contains subroutines output_data(), get_filename(),
!! get_dataexist() and get_nl()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine OUTPUT_DATA
!
!> Contains reading and writing hdf5 subroutines.
!<
#if HDF==1
subroutine output_data(filename,dataexist)

  use hdf5
  use parameters
  use var
  use varold
  use varray
  use divers
  use unites
  use unites_sortie
  use gammas
  use geom         , only : x_glob,x,shift_gr,nshift_gr
  use para         , only : mype,x_start,x_end
  use cellules     , only : ioffset,ifich,Pcell,N_cell
  use diffusion    , only : erd=>Er_dif
  use varray       , er => Eray,fr => Fray
  use rdwrt_h5

  implicit none

  
  integer,dimension(8 ) :: dataexist
  integer,dimension(19) :: para_int
  integer               :: nl1,nl2,nl3
  integer               :: i,j,k,l
  integer               :: n1,n2,n3,n4
  integer(HID_T)        :: file_id        !< File identifiers
  integer               :: error          !< Flag to check operation success 
  
  real, dimension(5)    :: para_real
  
  character(len=*)      :: filename 
  
  call get_nl(ndim,nl1,nl2,nl3)

  !Array sizes
  n1=nx(1)+2*Nbuf*nl1 ; n2=nx(2)+2*Nbuf*nl2 ; n3=nx(3)+2*Nbuf*nl3
  !
  ! Initialize FORTRAN interface. 
  !
  call h5open_f(error)
  
  !
  ! Create  file "create_output_hdf5_file.h5" using default properties.
  ! 
  call h5fcreate_f(trim(filename), H5F_ACC_TRUNC_F, file_id, error)

  ! Array "para_int"; Type integer; dim=1; size=19
  para_int( 1) = nx_glob(1)	  
  para_int( 2) = nx_glob(2)	
  para_int( 3) = nx_glob(3)	    
  para_int( 4) = ndim	  
  para_int( 5) = nx(1)	  
  para_int( 6) = nx(2)	
  para_int( 7) = nx(3)
  para_int( 8) = nfx
  para_int( 9) = x_start(1)	  
  para_int(10) = x_start(2) 
  para_int(11) = x_start(3)	  
  para_int(12) = x_end(1)	
  para_int(13) = x_end(2)	    
  para_int(14) = x_end(3)	  
  para_int(15) = k	  
  para_int(16) = 0
  if(RAY) para_int(16) = 1
  para_int(17) = 0
  if(DIFFU) para_int(17) = 1
  para_int(18) = 0
  if(DETO) para_int(18) = 1
#if MHD==1
  para_int(19) = 1
#else
  para_int(19) = 0
#endif
  call dump_1d_array_int_h5(file_id,"para_int",para_int,19)

  !Array "para_real"; Type real; dim=1; size=5
  para_real=(/ temps/u_T,dt/u_T,dtold/u_T,gamma,mu/)
  call dump_1d_array_h5(file_id,"para_real",para_real,5)

  !Array "x"; Type real; dim=1, size=nx(1)+1
  call dump_1d_array_h5(file_id,"x",x(1:nx(1)+1,1)/u_L,nx(1)+1)

  !Array "y"; Type real; dim=1, size=nx(2)+1
  call dump_1d_array_h5(file_id,"y",x(1:nx(2)+1,2)/u_L,nx(2)+1)

  !Array "z"; Type real; dim=1, size=nx(3)+1
  call dump_1d_array_h5(file_id,"z",x(1:nx(3)+1,3)/u_L,nx(3)+1)

  !Array "shift_gr"; Type real; dim=1, size=ndim
  call dump_1d_array_h5(file_id,"shift_gr",shift_gr(1:ndim),ndim)

  !Array "nshift_gr"; Type integer; dim=1, size=ndim
  call dump_1d_array_int_h5(file_id,"nshift_gr",nshift_gr(1:ndim),ndim)

  !Array "rho"; Type real; dim=3; size=(n1,n2,n3)
  if (dataexist(1)==1) then
     rho = rho/u_dens
     call dump_3d_array_h5(file_id,"rho",rho,n1,n2,n3)
     rho = rho*u_dens
  endif

  !Array "rhou"; Type real; dim=4; size=(n1,n2,n3,n4)
  if (dataexist(2)==1) then
     n4 = N_vit
     rhou = rhou/u_Mom
     call dump_4d_array_h5(file_id,"rhou",rhou,n1,n2,n3,n4)
     rhou = rhou*u_Mom
  endif

  !Array "E"; Type real; dim=3; size=(n1,n2,n3)
  if (dataexist(3)==1) then
     E = E/u_Evol 
     call dump_3d_array_h5(file_id,"E",E,n1,n2,n3)
     E = E*u_Evol 
  endif

#if SCALP==1
  !Array "fx"; Type real; dim=3; size=(n1,n2,n3)
  if (dataexist(4)==1) call dump_3d_array_h5(file_id,"fx",fx,n1,n2,n3)
#endif

  if (RAY) then
     !Array "Eray"; Type real; dim=3, size=(n1,n2,n3)
     if (dataexist(5)==1) then
        er = er/u_Evol 
        call dump_3d_array_h5(file_id,"er",er,n1,n2,n3)
        er = er*u_Evol 
     endif
     !Array "Fray"; Type real; dim=4, size=(n1,n2,n3,n4) 
     if (dataexist(6)==1) then
        n4 = ndim
        fr = fr/u_Fr 
        call dump_4d_array_h5(file_id,"fr",fr,n1,n2,n3,n4)
        fr = fr*u_Fr
     endif
  end if

  !Array "Er_dir"; Type real; dim=3, size=(n1,n2,n3)
  if (DIFFU .and. dataexist(7)==1) then
     erd = erd/u_Evol
     call dump_3d_array_h5(file_id,"erd",erd,n1,n2,n3)
     erd = erd*u_Evol
  end if

#if MHD==1
  !Array "B"; Type real; dim=4; size=(n1,n2,n3)
  if (dataexist(8)==1) then
     n4 = N_vit
     call dump_4d_array_h5(file_id,"B",B,n1,n2,n3,n4)
  end if
#endif

  !Array "supp1"; Type real; dim=3 or 4; size=(n1,n2,n3,suppdim(1))
  if (nsupp>=1) then
     if (suppdim(1)==1) then
        call dump_3d_array_h5(file_id,suppnames(1),supp1,n1,n2,n3)
     else
        call dump_4d_array_h5(file_id,suppnames(1),supp1,n1,n2,n3,suppdim(1))
     endif
  endif

  !Array "supp2"; Type real; dim=3 or 4; size=(n1,n2,n3,suppdim(2))
  if (nsupp>=2) then
     if (suppdim(2)==1) then
        call dump_3d_array_h5(file_id,suppnames(2),supp2,n1,n2,n3)
     else
        call dump_4d_array_h5(file_id,suppnames(2),supp2,n1,n2,n3,suppdim(2))
     endif
  endif

  !Array "supp3"; Type real; dim=3 or 4; size=(n1,n2,n3,suppdim(3))
  if (nsupp>=3) then
     if (suppdim(3)==1) then
        call dump_3d_array_h5(file_id,suppnames(3),supp3,n1,n2,n3)
     else
        call dump_4d_array_h5(file_id,suppnames(3),supp3,n1,n2,n3,suppdim(3))
     endif
  endif

  !Array "supp4"; Type real; dim=3 or 4; size=(n1,n2,n3,suppdim(4))
  if (nsupp>=4) then
     if (suppdim(4)==1) then
        call dump_3d_array_h5(file_id,suppnames(4),supp4,n1,n2,n3)
     else
        call dump_4d_array_h5(file_id,suppnames(4),supp4,n1,n2,n3,suppdim(4))
     endif
  endif

  !Array "supp5"; Type real; dim=3 or 4; size=(n1,n2,n3,suppdim(5))
  if (nsupp>=5) then
     if (suppdim(5)==1) then
        call dump_3d_array_h5(file_id,suppnames(5),supp5,n1,n2,n3)
     else
        call dump_4d_array_h5(file_id,suppnames(5),supp5,n1,n2,n3,suppdim(5))
     endif
  endif
  !
  ! Close file
  ! 
  call h5fclose_f(file_id, error)     
  !
  ! Close FORTRAN interface. 
  !
  call h5close_f(error) 
  
  return
  
end subroutine output_data

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine GET_FILENAME
!
!> Creates a hdf5 output filename.
!<
subroutine get_filename(filename,suffix,toutput,ioutput,mype)

  use divers, only : name_of_job,temps,nout
  use unites, only : seconde

  implicit none
  
  integer           :: ioutput,mype
  real              :: toutput
  character(len=23) :: filename
  character(len=14) :: suffix

  filename = '000/'//trim(name_of_job)//suffix

  if (toutput < temps .and. ioutput <= nout ) then
     if (mype == 0) write(*,*) 'Saving time (in s) :',ioutput,toutput/seconde,temps/seconde
     call mk_dir_out(ioutput,ioutput)
     call make_name(filename,mype,ioutput)
     write(*,*) toutput,temps,ioutput,nout,filename

     ioutput = ioutput + 1

  else if (ioutput <= nout ) then
     call mk_dir_out(0,0)
     call make_name(filename,mype,0)
  endif
  
  return

end subroutine get_filename

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine GET_DATAEXIST
!
!> Determines whether the data to be written actually
!! exists.
!<
subroutine get_dataexist(datanames,ndata,dataexist)

  implicit none

  integer, dimension(20) :: intab
  integer, dimension(8 ) :: dataexist
  integer                :: i,in,itrouv,insup,ndata

  character(80), dimension(20) :: datanames

  in = 0
  dataexist = 0
  intab = 0
  
  do i = 1,ndata
     itrouv=0
     if(datanames(i)=="rho" ) then 
        dataexist(1)=1
        itrouv=1
     endif
     if(datanames(i)=="rhou") then 
        dataexist(2)=1
        itrouv=1
     endif
     if(datanames(i)=="E"   ) then 
        dataexist(3)=1
        itrouv=1
     endif
     if(datanames(i)=="fx"  ) then 
        dataexist(4)=1
        itrouv=1
     endif
     if(datanames(i)=="er"  ) then 
        dataexist(5)=1
        itrouv=1
     endif
     if(datanames(i)=="fr"  ) then 
        dataexist(6)=1
        itrouv=1
     endif
     if(datanames(i)=="erd" ) then 
        dataexist(7)=1
        itrouv=1
     endif
     if(datanames(i)=="B"   ) then 
        dataexist(8)=1
        itrouv=1
     endif
     if(itrouv==0 .and. datanames(i)/="") then 
        in=in+1
        dataexist(8+in) = 1
        intab(in)=i
     endif
  end do
  
  insup = in

  return

end subroutine get_dataexist

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine GET_NL
!
!> Returns the number of lines according to the number of
!! dimensions.
!<
subroutine get_nl(ndim,nl1,nl2,nl3)

  implicit none

  integer :: ndim,nl1,nl2,nl3

  if(ndim == 1) then
     nl1 = 1
     nl2 = 0
     nl3 = 0
  else if(ndim == 2) then 
     nl1 = 1
     nl2 = 1
     nl3 = 0
  else if(ndim == 3) then 
     nl1 = 1
     nl2 = 1
     nl3 = 1 
  endif

  return
  
end subroutine get_nl

#else

subroutine output_data(filename,dataexist)

  use para, only : mype

  implicit none

  integer,dimension(8) :: dataexist
  character(len=*)     :: filename 

  call close_program(mype,'hdf5 dummy: output_data')

end subroutine output_data

#endif
