!> \file
!! Contains subroutine wrt_restart_h5()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine WRT_RESTART_H5
!
!> Writes an hdf5 output file which can be used for
!! performing a simulation restart.
!<
subroutine wrt_restart_h5(k)

#if HDF==1

  use hdf5
  use parameters
  use var
  use varold
  use varray
  use divers
  use divers_ray   , only : ngrp_ray
  use unites
  use unites_sortie
  use gammas
  use geom         , only : x_glob,x,shift_gr,nshift_gr,geometrie
  use para         , only : mype,x_start,x_end
  use cellules     , only : ioffset, ifich,Pcell,N_cell
  use varray       ,        er => Eray,fr => Fray
  use diffusion    , only : erd => Er_dif,ngrp_dif

  use rdwrt_h5

  implicit none

  integer, dimension(23) :: para_int
  integer(HID_T)         :: file_id              !< File identifiers
  integer                :: nl1,nl2,nl3
  integer                :: error                !< Flag to check operation success 
  integer                :: i,j,k,l,n1,n2,n3,n4

  real, dimension(8) :: para_real

  character(len=23) :: filename 

  !call get_filename(filename,'-00000_000.hdf',trepr(irepr),irepr,mype)
  call get_filename(filename,'-00000_000.hdf',tout(iout),iout,mype)
  if (mype == 0) write(*,*) 'Writing file :',filename

  call get_nl(ndim,nl1,nl2,nl3)

  !Array sizes
  n1=nx(1)+2*Nbuf*nl1 ; n2=nx(2)+2*Nbuf*nl2 ; n3=nx(3)+2*Nbuf*nl3

  !
  ! Initialize FORTRAN interface. 
  !
  call h5open_f(error) 
  call h5fcreate_f(filename, H5F_ACC_TRUNC_F, file_id, error)

  ! Array "para_int"; Type integer; dim=1; size=23
  para_int( 1) = nx_glob(1)	  
  para_int( 2) = nx_glob(2)	
  para_int( 3) = nx_glob(3)	    
  para_int( 4) = ndim	  
  para_int( 5) = nx(1)	  
  para_int( 6) = nx(2)	
  para_int( 7) = nx(3)
  para_int( 8) = nfx
  para_int( 9) = x_start(1)	  
  para_int(10) = x_start(2) 
  para_int(11) = x_start(3)	  
  para_int(12) = x_end(1)	
  para_int(13) = x_end(2)	    
  para_int(14) = x_end(3)	  
  para_int(15) = k	  
  para_int(16) = 0
  if(RAY) para_int(16) = 1	
  para_int(17) = 0
  if(DIFFU) para_int(17) = 1	
  para_int(18) = 0
  if(CONDUC) para_int(18) = 1	
#if MHD==1
  para_int(19) = 1
#else
  para_int(19) = 0
#endif
  para_int(20) = 0
  if(GRAV) para_int(20) = 1
  para_int(21) = Nbuf
  para_int(22) = Ngrp_ray
  para_int(23) = Ngrp_dif

  call dump_1d_array_int_h5(file_id,"para_int",para_int,23)

  !String geometry, length=5, dims=1
  call dump_1d_array_string_h5(file_id,"geometry",geometrie,5)

  !Array "para_real"; Type real; dim=1; size=8
  para_real= (/ temps/u_T,dt/u_T,dtold/u_T,gamma,mu,u_M/gramme,u_L/centimetre,u_T/seconde /)
  call dump_1d_array_h5(file_id,"para_real",para_real,8)

  !Array "x_glob"; Type real; dim=1, size=nx_glob(1)+1
  call dump_1d_array_h5(file_id,"x_glob",x_glob(1:nx_glob(1)+1,1)/u_L,nx_glob(1)+1)

  !Array "y_glob"; Type real; dim=1, size=nx_glob(2)+1
  call dump_1d_array_h5(file_id,"y_glob",x_glob(1:nx_glob(2)+1,2)/u_L,nx_glob(2)+1)

  !Array "z_glob"; Type real; dim=1, size=nx_glob(3)+1
  call dump_1d_array_h5(file_id,"z_glob",x_glob(1:nx_glob(2)+1,3)/u_L,nx_glob(3)+1)

  !Array "x"; Type real; dim=1, size=nx(1)+1
  call dump_1d_array_h5(file_id,"x",x(1:nx(1)+1,1)/u_L,nx(1)+1)

  !Array "y"; Type real; dim=1, size=nx(2)+1
  call dump_1d_array_h5(file_id,"y",x(1:nx(2)+1,2)/u_L,nx(2)+1)

  !Array "z"; Type real; dim=1, size=nx(3)+1
  call dump_1d_array_h5(file_id,"z",x(1:nx(3)+1,3)/u_L,nx(3)+1)

  !Array "shift_gr"; Type real; dim=1, size=ndim
  call dump_1d_array_h5(file_id,"shift_gr",shift_gr(1:ndim),ndim)

  !Array "nshift_gr"; Type integer; dim=1, size=ndim
  call dump_1d_array_int_h5(file_id,"nshift_gr",nshift_gr(1:ndim),ndim)

  !Array "rho"; Type real; dim=3; size=(n1,n2,n3)
  rho=rho/u_dens
  call dump_3d_array_h5(file_id,"rho",rho,n1,n2,n3)
  rho=rho*u_dens

  !Array "rhou"; Type real; dim=4; size=(n1,n2,n3,n4)
  n4=N_vit
  rhou = rhou/u_Mom
  call dump_4d_array_h5(file_id,"rhou",rhou,n1,n2,n3,n4)
  rhou = rhou*u_Mom

  !Array "E"; Type real; dim=3; size=(n1,n2,n3)
  E = E/u_Evol 
  call dump_3d_array_h5(file_id,"E",E,n1,n2,n3)
  E = E*u_Evol 

#if SCALP==1
  if (nfx>=1) then
     n4=nfx
     !Array "fx"; Type real; dim=3; size=(n1,n2,n3,n4)
     call dump_4d_array_h5(file_id,"fx",fx,n1,n2,n3,n4)
  endif
#endif

  if (RAY) then
     !Array "Eray"; Type real; dim=3, size=(n1,n2,n3)
     er = er/u_Evol 
     call dump_3d_array_h5(file_id,"er",er,n1,n2,n3)
     er = er*u_Evol 
     !Array "Fray"; Type real; dim=4, size=(n1,n2,n3,n4) 
     n4=ndim
     fr = fr/u_Fr 
     call dump_4d_array_h5(file_id,"fr",fr,n1,n2,n3,n4)
     fr = fr*u_Fr
  end if

  !Array "Er_dir"; Type real; dim=3, size=(n1,n2,n3)
  if (DIFFU) then
     erd = erd/u_Evol
     call dump_3d_array_h5(file_id,"erd",erd,n1,n2,n3)
     erd = erd*u_Evol
  end if

#if MHD==1
  !Array "B"; Type real; dim=4; size=(n1,n2,n3)
  n4=N_vit
  call dump_4d_array_h5(file_id,"B",B,n1,n2,n3,n4)
#endif

  !
  ! Close file, dataset and dataspace identifiers.
  !
  call h5fclose_f(file_id, error)  

  !
  ! Close FORTRAN interface. 
  !
  call h5close_f(error) 

  return

#else

  use para, only : mype
  
  implicit none

  integer :: k

  call close_program(mype,'hdf5 dummy: wrt_restart_h5')
  
#endif

end subroutine wrt_restart_h5
