!> \file
!! Contains subroutines aff_new(),  make_name(),
!! mk_dir_out() and convtoasc()
!!
!! This file contains all the routines used for writing
!! output files and printing outputs to the terminal.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine AFF_NEW
!
!> This routine writes binary data cubes which
!! will be used for data analysis and for
!! simulation restart.
!<
subroutine aff_new(k)

  use parameters
  use var
  use varold
  use varray
  use divers
  use divers_ray , only : ngrp_ray
  use unites
  use unites_sortie
  use gammas
  use geom       , only : x_glob,x,shift_gr,nshift_gr,geometrie
  use para       , only : mype,x_start,x_end,ncpu
  use cellules   , only : ioffset, ifich,Pcell,N_cell
  use diffusion  , only : ngrp_dif,Er_dif

  implicit none

  integer            :: k,it
  character (len=23) :: liste
  logical            :: mhd_log

  liste = 'XXX/liste-00000_000.Bin'

  it = 0

  if (tout(iout) < temps ) then
     it   = iout

     if (mype == 0) write(*,876) iout,tout(iout)/seconde,temps/seconde
876 format('Saving time (s): ',i5,2(ES10.3,2X)) 

     iout = iout + 1

  endif

  call Mk_dir_out(it,it)

  call make_name(liste,mype,it)

  if (mype == 0) write(*,*) 
  if (mype == 0) write(*,*) 'Writing file: ',liste
  if (mype == 0) write(*,*) 

  open (10,file=liste,form='unformatted')
  write(10) nx_glob,ncpu,geometrie
  write(10) x_glob(1:nx_glob(1)+1,1)/u_L,x_glob(1:nx_glob(2)+1,2)/u_L,x_glob(1:nx_glob(3)+1,3)/u_L
  write(10) ndim,nx,nfx,nbuf,ngrp_ray,ngrp_dif
#if MHD==1
  mhd_log=.true.
#else
  mhd_log=.false.
#endif
  write(10) mhd_log,ray,diffu,conduc,grav
  write(10) x_start,x_end
  write(10) x(1:nx(1)+1,1)/u_L,x(1:nx(2)+1,2)/u_L,x(1:nx(3)+1,3)/u_L
  write(10)  shift_gr(1:ndim) 
  write(10) nshift_gr(1:ndim) 
  write(10) k
  write(10) temps/u_T, dt/u_T, dtold/u_T
  write(10) gamma,mu
  write(10) u_M/gramme,u_L/centimetre,u_T/seconde

  rho  = rho  / u_dens
  rhou = rhou / u_Mom
  E    = E    / u_Evol

  ! Performing a division in the writes uses up more memory
  write(10) rho
  write(10) rhou
  write(10) E

  rho  = rho  * u_dens
  rhou = rhou * u_Mom
  E    = E    * u_Evol

#if SCALP==1
  if(nfx.ge.1) write(10) fx  
#endif

  if (RAY) then
     Eray = Eray / u_Evol
     Fray = Fray / u_Fr

     write(10) Eray
     write(10) Fray

     Eray = Eray * u_Evol
     Fray = Fray * u_Fr
  endif
  if (DIFFU) then
     Er_dif = Er_dif / u_Evol

     write(10) Er_dif

     Er_dif = Er_dif * u_Evol
  endif

#if MHD==1
     write(10) B
#endif

  close(10)
  return
  
end subroutine aff_new

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine MAKE_NAME
!
!> Creates a name for the ouput binary file.
!<
subroutine make_name(liste,mype,iout)

  implicit none

  !
  ! CAUTION : liste must contain 23 characters and the following format:
  !              
  !          MMM/XXXXX-NNNNN_MMM.end  where 'XXXXX' : name with 5 characters
  !                                          NNNNN  : processor number
  !                                            MMM  : output time number
  !                                            end  : bin, Bin ou hdf 

  integer, intent(in) :: mype,iout
  character(len=23)   :: liste

  
  call convtoasc(iout,liste(1:3),3)
  liste(17:19)=liste(1:3)
  call convtoasc(mype+1,liste(11:15),5)

  return

end subroutine make_name

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine MK_DIR_OUT
!
!> Creates a new directory for outputting simulation data.
!<
subroutine mk_dir_out(i1,i2)

  use para, only : mype

  implicit none

#if PARA==1
  include 'mpif.h'
  integer       :: info
#endif

  character(3 ) :: dir
  character(20) :: filecmd
  integer       :: i1,i2

  integer       :: i

  do i=i1,i2
     call convtoasc(i,dir,3)
#ifdef SYSTEM
     !if a call system is needed
     filecmd='mkdir -p '//trim(dir)
     if (mype==0) call system(filecmd)
#else
     if (mype==0) call PXFMKDIR(trim(dir),len(trim(dir)),O'755',info)
#endif

  enddo

#if PARA==1
  call MPI_BARRIER(MPI_COMM_WORLD,info)
#endif

 return
 
end subroutine mk_dir_out

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CONVTOASC
!
!> Converts an integer smaller than 999999 to a ndigits
!! characters string.
!<
subroutine convtoasc(number,sstring,ndigits)

  implicit none

  integer                     :: number,istring,ndigits,num,nums10,i
  character(len=ndigits)      :: sstring
  character(len=10),parameter :: nstring="0123456789"
     
  num=10**ndigits
  nums10=num/10
  do i=1,ndigits
      istring=1+mod(number,num)/nums10
      sstring(i:i)=nstring(istring:istring)
      num=num/10
      nums10=nums10/10
  enddo
  
  return
  
end subroutine convtoasc

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PRINT_CONFIGURATION
!
!> Prints the initial configuration of the simulation: size
!! of grid, whether gravity, MHD, rad transfer or MPI are
!! being used, etc...
!<
subroutine print_configuration(format_sortie)

  use divers
  use divers_ray, only : rad_trans_model,matrix_inv_solver,precond_ray,slope_type_ray,varrel_ray
  use diffusion , only : fld_limiter,varrel_dif,precond_dif,dif_imp
  use parameters, only : nfx,ndim,slope_type,nx_glob
  use para      , only : ncpu,ncpu_x,ncpu_y,ncpu_z
  use geom
  use unites    , only : gramme,centimetre,seconde,degre

  implicit none
  
  character(len=30) :: string
  character(len=3 ) :: format_sortie

  write(*,*)
  write(*,*)
  write(*,'(a)') '############################# CONFIGURATION ##############################'

#if PARA==1
     write(*,'(a,i5,a)') ' MPI: ',ncpu,' cpus           ON'
#else
     write(*,'(a)') ' MPI:                           OFF'
#endif

  if(reprise)then
     write(*,'(a)') ' Restart:                       ON'
  else
     write(*,'(a)') ' Restart:                       OFF'
  endif

  if(HYDRO)then
     write(*,'(a)') ' Hydrodynamics:                 ON'
  else
     write(*,'(a)') ' Hydrodynamics:                 OFF'
  endif

#if MHD==1
     write(*,'(a)') ' MHD:                           ON'
#else
     write(*,'(a)') ' MHD:                           OFF'
#endif

  if(RAY)then
     write(*,'(a)') ' Radiative transfer: '//rad_trans_model//'         ON'
  else
     write(*,'(a)') ' Radiative transfer:            OFF'
  endif

  if(CONDUC)then
     write(*,'(a)') ' Thermal conduction:            ON'
  else
     write(*,'(a)') ' Thermal conduction:            OFF'
  endif
  
  if(DIFFU)then
     write(*,'(a)') ' Flux limited diffusion:        ON'
  else
     write(*,'(a)') ' Flux limited diffusion:        OFF'
  endif
  
  if(GRAV)then
     write(*,'(a)') ' Gravity:                       ON'
  else
     write(*,'(a)') ' Gravity:                       OFF'
  endif
  
  if(nfx >= 1)then
     write(*,'(a)') ' Passive scalar:                ON'
  else
     write(*,'(a)') ' Passive scalar:                OFF'
  endif
  
  write(*,'(a)') '--------------------------------------------------------------------------'
  write(*,'(a)') '------------------------------- Grid setup -------------------------------'
  if(Cartesien)then
     write(*,*) '    Cartesian coordinates   : ',geometrie, '  ndim =',ndim
  else if (Cylindrique) then
     write(*,*) '    Cylindrical coordinates : ',geometrie, '  ndim =',ndim
  else if (Spherique  ) then 
     write(*,*) '    Spherical coordinates   : ',geometrie, '  ndim =',ndim
  endif
  write(*,'(a,i6)') '           direction 1 : '//geom_dir(1)//' nx = ',nx_glob(1)
  write(*,'(a,i6)') '           direction 2 : '//geom_dir(2)//' ny = ',nx_glob(2)
  write(*,'(a,i6)') '           direction 3 : '//geom_dir(3)//' nz = ',nx_glob(3)
  write(*,*)
  write(*,'(a,es10.3,a,es10.3,a,es10.3)') '           Lbox(1) =',Lbox(1),' ;  x(0) = ',x_glob(1,1),' ;  x(nx+1) =',x_glob(nx_glob(1)+1,1)
  write(*,'(a,es10.3,a,es10.3,a,es10.3)') '           Lbox(2) =',Lbox(2),' ;  y(0) = ',x_glob(1,2),' ;  y(ny+1) =',x_glob(nx_glob(2)+1,2)
  write(*,'(a,es10.3,a,es10.3,a,es10.3)') '           Lbox(3) =',Lbox(3),' ;  z(0) = ',x_glob(1,3),' ;  z(nz+1) =',x_glob(nx_glob(3)+1,3)
#if PARA==1
  write(*,*)
  write(*,'(a,3(i6))') ' ncpu_x,ncpu_y,ncpu_z : ',ncpu_x,ncpu_y,ncpu_z
#endif
  
  write(*,'(a)') '--------------------------------------------------------------------------'
  write(*,'(a)') '-------------------------- Code units (in cgs) ---------------------------'
  write(*,'(a,es10.2)') ' Mass:        ',gramme
  write(*,'(a,es10.2)') ' Length:      ',centimetre
  write(*,'(a,es10.2)') ' Time:        ',seconde
  write(*,'(a,es10.2)') ' Temperature: ',degre
  
  select case(slope_type)
  case(0)
     string = 'Null'
  case(1)
     string = 'Minmod'
  case(2)
     string = 'Moncen'
  case(3)
     string = 'Positivity preserving'
  case(4)
     string = 'Harmonic average'
  case(5)
     string = 'Van Albada & Van Leer'
  case(6)
     string = 'Superbee'
  end select
  
  write(*,'(a)') '--------------------------------------------------------------------------'
  write(*,'(a)') '------------------------- Simulation parameters --------------------------'
  write(*,'(a)') ' General:'
  write(*,'(a,es9.3)') '    End time:           ',tend/seconde
  write(*,'(a      )') '    Slope type:         '//trim(string)
  if(isotherme)then
     write(*,'(a   )') '    Equation of state:  Ideal gas - isothermal'
  else
     write(*,'(a   )') '    Equation of state:  Ideal gas'
  endif
  write(*,'(a,3(i6))') '    nsx,nsy,nsz:        ',nsx,nsy,nsz
  write(*,'(a,es9.3)') '    Maximum timestep:   ',dt_max/seconde
  write(*,'(a      )') '    Output file format: '//format_sortie
  write(*,'(a,i6   )') '    Number of outputs:  ',nout
  
  if(RAY)then
  
     select case(slope_type_ray)
     case(0)
        string = 'Null'
     case(1)
        string = 'Minmod'
     case(2)
        string = 'Moncen'
     case(3)
        string = 'Positivity preserving'
     case(4)
        string = 'Harmonic average'
     case(5)
        string = 'Van Albada & Van Leer'
     case(6)
        string = 'Superbee'
     end select
     
     write(*,*)
     write(*,'(a      )') 'Radiative transfer:'
     write(*,'(a      )') '    Matrix inversion solver: '//matrix_inv_solver
     write(*,'(a      )') '    Slope type:              '//trim(string)
     write(*,'(a,es9.3)') '    Varrel:                  ',varrel_ray
     if(precond_ray)then
        write(*,'(a   )') '    Preconditionning:        ON'
     else
        write(*,'(a   )') '    Preconditionning:        OFF'
     endif
     
  endif
  
  if(DIFFU)then
     
     write(*,*)
     write(*,'(a      )') 'Flux limited diffusion:'
     if(dif_imp)then
        write(*,'(a   )') '    Solver:                  Implicit'
     else
        write(*,'(a   )') '    Solver:                  Explicit'
     endif
     write(*,'(a      )') '    Flux limiter:            '//fld_limiter
     write(*,'(a,es9.3)') '    Varrel:                  ',varrel_dif
     if(precond_dif)then
        write(*,'(a   )') '    Preconditionning:        ON'
     else
        write(*,'(a   )') '    Preconditionning:        OFF'
     endif
     if(.not.Dif_imp .and. fld_limiter.ne.'none') then
        write(*,'(a   )') 'Explicit diffusion with  flux limiter: possibly irrelevant CFL condition'
     endif
     
  endif
  
  write(*,'(a)') '##########################################################################'
  write(*,*)
  write(*,*)

  return
  
end subroutine print_configuration

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine OUTPUT_TO_TERMINAL
!
!> Prints the timestep information to the terminal.
!<
subroutine output_to_terminal

  use divers, only : terminal_output,noutput

  implicit none

  integer :: n

  do n = 1,noutput
     write(*,'(a)') trim(terminal_output(n))
  enddo

  return

end subroutine output_to_terminal
