!> \file
!! Contains subroutine check_flags()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CHECK_FLAGS
!
!> Performs a safety check on all the major flags
!! (GRAV, RAY, MHD, etc...)
!<
subroutine check_flags(format_sortie)

  use para       , only : mype,ncpu
  use parameters , only : ndim,nx_glob,nfx,N_vit,Nbuf
  use divers     , only : HYDRO,RAY,CONDUC,DIFFU,GRAV
  use cl         , only : icl_glob
  use cl_ray     , only : icl_ray_glob
  use geom       , only : geometrie,geom_dir,Cartesien,Cylindrique,Spherique

  implicit none

  character(len=*)  :: format_sortie

  ! local variables
  integer :: i

#if GRA==0
  if (GRAV) then
     call close_program(mype,'grav = .true. but flag -DGRA not set during compilation...')
  endif
#endif

#if RAYO==0
  if (RAY) then
     call close_program(mype,'ray = .true. but flag -DRAYO not set during compilation...')
  endif
#endif

#if CON==0
  if (CONDUC) then
     call close_program(mype,'Conduc = .true. but flag -DCON not set during compilation...')
  endif
#endif

#if DIF==0
  if (DIFFU) then
     call close_program(mype,'Diffu = .true. but flag -DDIF not set during compilation...')
  endif
#endif

#if SCALP==0
  if (mype == 0) then
     if (nfx >= 1) then
        call close_program(mype,'Nfx >= 1 but flag -DSCALP not set during compilation')
     endif
  endif
#endif

  do i=1,2*ndim
     if(icl_glob(i).ne.1 .and. icl_glob(i).ne.2 .and. icl_glob(i).ne.3 .and.  &
        icl_glob(i).ne.4 .and. icl_glob(i).ne.5                             ) then
        if(mype==0) write(*,*) 'Problem with icl_glob(i) : i,icl_glob(i)=',i,icl_glob(i)
        call close_program(mype,'Check the namelist...')
     endif
#if RAYO==1
     if(icl_ray_glob(i).ne.1 .and. icl_ray_glob(i).ne.2 .and. icl_ray_glob(i).ne.3 .and.  &
        icl_ray_glob(i).ne.4 .and. icl_ray_glob(i).ne.5                                 ) then
        if(mype==0) write(*,*) 'Problem with icl_ray_glob(i) : i,icl_ray_glob(i)=',i,icl_ray_glob(i)
        call close_program(mype,'Check the namelist...')
     endif
#endif
  enddo

  if( (ndim.eq.1 .and. (nx_glob(2).ne.1 .or. nx_glob(3).ne.1)) .or. &
      (ndim.eq.2 .and.                       nx_glob(3).ne.1 )      ) then
     if(mype==0) write(*,003) ndim,nx_glob(1),nx_glob(2),nx_glob(3)
003 format('Unmatched value of ndim vs. nx_glob,ny_glob,nz_glob =',5(i4,2x))
     call close_program(mype,'Check the namelist...')
  endif

#if MHD==1
  if (Nfx > 0) then
     write(*,*) 'Nfx = ',Nfx
     write(*,*) ' Passive scalar unsupported with MHD...'
  endif
#endif

  if (FORMAT_SORTIE == 'bin') then
  else if (FORMAT_SORTIE == 'hdf') then
#if HDF==1
#else
     if (mype == 0) call close_program(mype,'FORMAT_SORTIE = ''hdf'' but flag -DHDF not set during compilation...')
#endif
  else
      call close_program(mype,'Unknow output format: FORMAT_SORTIE = '//FORMAT_SORTIE)
  endif

  ! Determine geometry
  Cartesien = .false. ; Cylindrique = .false. ; Spherique = .false. ; geom_dir = ' '
  if (geometrie == 'x'  .or. geometrie == 'xy'   .or. geometrie == 'xyz'                        ) then
     Cartesien   = .true.
                                                      geom_dir(1) = 'x'
     if (geometrie == 'xy'  .or. geometrie == 'xyz' ) geom_dir(2) = 'y'
     if (                        geometrie == 'xyz' ) geom_dir(3) = 'z'
  endif

  if (geometrie == 'rc' .or. geometrie == 'rz'   .or. geometrie == 'rp'  .or. geometrie == 'rpz') then
     Cylindrique = .true.
                                                      geom_dir(1) = 'rc'
     if (geometrie == 'rp'  .or. geometrie == 'rpz' ) geom_dir(2) = 'phi_c'

     if((ndim==2 .and. geometrie=='rc').or.(ndim==3 .and. geometrie.ne.'rpz')) then
        if(mype==0) write(*,004) geometrie,ndim
004 format('Unmatched value of geometrie vs ndim = ',a5,2x,i4)
        call close_program(mype,'Geometry problem')
     endif
  endif

  if (geometrie == 'rs' .or. geometrie == 'rthe' .or. geometrie == 'rtp'                        ) then
     Spherique   = .true.
                                                        geom_dir(1) = 'rs'
     if (geometrie == 'rthe'  .or. geometrie == 'rtp' ) geom_dir(2) = 'the_s'
     if (                          geometrie == 'rtp' ) geom_dir(3) = 'phi_s'

     if((ndim>=2.and.geometrie=='rs').or.(ndim==3.and.geometrie=='rthe')) then
        if(mype==0) write(*,004) geometrie,ndim
        call close_program(mype,'Geometry problem')
     endif
  endif

  if((.not.Cartesien).and.(.not.Cylindrique).and.(.not.Spherique))then
     call close_program(mype,'Undefined geometry: '//geometrie)
  endif

  !Set N_vit to the right value
  N_vit=ndim

#if MHD==1
  HYDRO=.false.
  N_vit=3 ; Nbuf=3
#endif

  return
  
end subroutine check_flags
