!> \file
!! Contains module communication_gen and subroutines
!! allocate_array_com_gen(), communications_gen(),
!! com_1D_gen(), com_2D_gen() and com_3D()
!!
!! This file contains all the variables and routines
!! necessary to the generic communications between cpus.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module COMMUNICATION_GEN
!
!> Contains variables arrays used to communicate ghost
!! cell data across cpus.
!<
module communication_gen
  implicit none
  real, dimension(:,:,:,:), allocatable :: face1_e        !< Face 1 e
  real, dimension(:,:,:,:), allocatable :: face2_e        !< Face 2 e
  real, dimension(:,:,:,:), allocatable :: face3_e        !< Face 3 e
  real, dimension(:,:,:,:), allocatable :: face4_e        !< Face 4 e
  real, dimension(:,:,:,:), allocatable :: face5_e        !< Face 5 e
  real, dimension(:,:,:,:), allocatable :: face6_e        !< Face 6 e
  real, dimension(:,:,:,:), allocatable :: face1_r        !< Face 1 r
  real, dimension(:,:,:,:), allocatable :: face2_r        !< Face 2 r
  real, dimension(:,:,:,:), allocatable :: face3_r        !< Face 3 r
  real, dimension(:,:,:,:), allocatable :: face4_r        !< Face 4 r
  real, dimension(:,:,:,:), allocatable :: face5_r        !< Face 5 r
  real, dimension(:,:,:,:), allocatable :: face6_r        !< Face 6 r
  real, dimension(:,:,:,:), allocatable :: ar13_e         !< Edge 1 e
  real, dimension(:,:,:,:), allocatable :: ar14_e         !< Edge 2 e
  real, dimension(:,:,:,:), allocatable :: ar23_e         !< Edge 3 e
  real, dimension(:,:,:,:), allocatable :: ar24_e         !< Edge 4 e
  real, dimension(:,:,:,:), allocatable :: ar15_e         !< Edge 5 e
  real, dimension(:,:,:,:), allocatable :: ar25_e         !< Edge 6 e
  real, dimension(:,:,:,:), allocatable :: ar35_e         !< Edge 7 e
  real, dimension(:,:,:,:), allocatable :: ar45_e         !< Edge 8 e
  real, dimension(:,:,:,:), allocatable :: ar16_e         !< Edge 9 e
  real, dimension(:,:,:,:), allocatable :: ar26_e         !< Edge 10 e
  real, dimension(:,:,:,:), allocatable :: ar36_e         !< Edge 11 e
  real, dimension(:,:,:,:), allocatable :: ar46_e         !< Edge 12 e
  real, dimension(:,:,:,:), allocatable :: ar13_r         !< Edge 1 r
  real, dimension(:,:,:,:), allocatable :: ar14_r         !< Edge 2 r
  real, dimension(:,:,:,:), allocatable :: ar23_r         !< Edge 3 r
  real, dimension(:,:,:,:), allocatable :: ar24_r         !< Edge 4 r
  real, dimension(:,:,:,:), allocatable :: ar15_r         !< Edge 5 r
  real, dimension(:,:,:,:), allocatable :: ar25_r         !< Edge 6 r
  real, dimension(:,:,:,:), allocatable :: ar35_r         !< Edge 7 r
  real, dimension(:,:,:,:), allocatable :: ar45_r         !< Edge 8 r
  real, dimension(:,:,:,:), allocatable :: ar16_r         !< Edge 9 r
  real, dimension(:,:,:,:), allocatable :: ar26_r         !< Edge 10 r
  real, dimension(:,:,:,:), allocatable :: ar36_r         !< Edge 11 r
  real, dimension(:,:,:,:), allocatable :: ar46_r         !< Edge 12 r
  real, dimension(:,:,:,:), allocatable :: coin13_e       !< Corner 2D 1 e
  real, dimension(:,:,:,:), allocatable :: coin14_e       !< Corner 2D 2 e
  real, dimension(:,:,:,:), allocatable :: coin23_e       !< Corner 2D 3 e
  real, dimension(:,:,:,:), allocatable :: coin24_e       !< Corner 2D 4 e
  real, dimension(:,:,:,:), allocatable :: coin13_r       !< Corner 2D 1 r
  real, dimension(:,:,:,:), allocatable :: coin14_r       !< Corner 2D 2 r
  real, dimension(:,:,:,:), allocatable :: coin23_r       !< Corner 2D 3 r
  real, dimension(:,:,:,:), allocatable :: coin24_r       !< Corner 2D 4 r
  real, dimension(:,:,:,:), allocatable :: coin135_e      !< Corner 3D 1 e
  real, dimension(:,:,:,:), allocatable :: coin136_e      !< Corner 3D 2 e
  real, dimension(:,:,:,:), allocatable :: coin145_e      !< Corner 3D 3 e
  real, dimension(:,:,:,:), allocatable :: coin146_e      !< Corner 3D 4 e
  real, dimension(:,:,:,:), allocatable :: coin235_e      !< Corner 3D 5 e
  real, dimension(:,:,:,:), allocatable :: coin236_e      !< Corner 3D 6 e
  real, dimension(:,:,:,:), allocatable :: coin245_e      !< Corner 3D 7 e
  real, dimension(:,:,:,:), allocatable :: coin246_e      !< Corner 3D 8 e
  real, dimension(:,:,:,:), allocatable :: coin135_r      !< Corner 3D 1 r
  real, dimension(:,:,:,:), allocatable :: coin136_r      !< Corner 3D 2 r
  real, dimension(:,:,:,:), allocatable :: coin145_r      !< Corner 3D 3 r
  real, dimension(:,:,:,:), allocatable :: coin146_r      !< Corner 3D 4 r
  real, dimension(:,:,:,:), allocatable :: coin235_r      !< Corner 3D 5 r
  real, dimension(:,:,:,:), allocatable :: coin236_r      !< Corner 3D 6 r
  real, dimension(:,:,:,:), allocatable :: coin245_r      !< Corner 3D 7 r
  real, dimension(:,:,:,:), allocatable :: coin246_r      !< Corner 3D 8 r
  integer, parameter                    :: Nmax = 1       !< nmax
end module communication_gen

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine ALLOCATE_ARRAY_COM_GEN
!
!> Allocates the arrays used in the communication process.
!<
subroutine allocate_array_com_gen

  use communication_gen
  use parameters

  implicit none

  integer :: nnx,nny,nnz
  
  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)
  
  if (ndim == 1) then
     allocate(face1_e(Nbuf,1,1,Nmax), face2_e(Nbuf,1,1,Nmax))
     allocate(face1_r(Nbuf,1,1,Nmax), face2_r(Nbuf,1,1,Nmax))
  endif
     
  if (ndim == 2) then
     allocate(face1_e(Nbuf,1:nny,1,Nmax), face2_e(Nbuf,1:nny,1,Nmax))
     allocate(face1_r(Nbuf,1:nny,1,Nmax), face2_r(Nbuf,1:nny,1,Nmax))
     
     allocate(face3_e(1:nnx,Nbuf,1,Nmax), face4_e(1:nnx,Nbuf,1,Nmax))
     allocate(face3_r(1:nnx,Nbuf,1,Nmax), face4_r(1:nnx,Nbuf,1,Nmax))
     
     allocate(coin13_e(Nbuf,Nbuf,1,Nmax),coin14_e(Nbuf,Nbuf,1,Nmax),coin23_e(Nbuf,Nbuf,1,Nmax),coin24_e(Nbuf,Nbuf,1,Nmax))
     allocate(coin13_r(Nbuf,Nbuf,1,Nmax),coin14_r(Nbuf,Nbuf,1,Nmax),coin23_r(Nbuf,Nbuf,1,Nmax),coin24_r(Nbuf,Nbuf,1,Nmax))
  endif
  
  
  if (ndim == 3) then
     allocate(face1_e(Nbuf,1:nny,1:nnz,Nmax), face2_e(Nbuf,1:nny,1:nnz,Nmax))
     allocate(face1_r(Nbuf,1:nny,1:nnz,Nmax), face2_r(Nbuf,1:nny,1:nnz,Nmax))
     
     allocate(face3_e(1:nnx,Nbuf,1:nnz,Nmax), face4_e(1:nnx,Nbuf,1:nnz,Nmax))
     allocate(face3_r(1:nnx,Nbuf,1:nnz,Nmax), face4_r(1:nnx,Nbuf,1:nnz,Nmax))
     
     allocate(face5_e(1:nnx,1:nny,Nbuf,Nmax), face6_e(1:nnx,1:nny,Nbuf,Nmax))
     allocate(face5_r(1:nnx,1:nny,Nbuf,Nmax), face6_r(1:nnx,1:nny,Nbuf,Nmax))
     
     
     allocate(ar13_e(Nbuf,Nbuf,1:nnz,Nmax), ar14_e(Nbuf,Nbuf,1:nnz,Nmax), ar23_e(Nbuf,Nbuf,1:nnz,Nmax), ar24_e(Nbuf,Nbuf,1:nnz,Nmax))
     allocate(ar13_r(Nbuf,Nbuf,1:nnz,Nmax), ar14_r(Nbuf,Nbuf,1:nnz,Nmax), ar23_r(Nbuf,Nbuf,1:nnz,Nmax), ar24_r(Nbuf,Nbuf,1:nnz,Nmax))
     
     allocate(ar15_e(Nbuf,1:nny,Nbuf,Nmax), ar25_e(Nbuf,1:nny,Nbuf,Nmax), ar16_e(Nbuf,1:nny,Nbuf,Nmax), ar26_e(Nbuf,1:nny,Nbuf,Nmax))
     allocate(ar15_r(Nbuf,1:nny,Nbuf,Nmax), ar25_r(Nbuf,1:nny,Nbuf,Nmax), ar16_r(Nbuf,1:nny,Nbuf,Nmax), ar26_r(Nbuf,1:nny,Nbuf,Nmax))
     
     allocate(ar35_e(1:nnx,Nbuf,Nbuf,Nmax), ar45_e(1:nnx,Nbuf,Nbuf,Nmax), ar36_e(1:nnx,Nbuf,Nbuf,Nmax), ar46_e(1:nnx,Nbuf,Nbuf,Nmax))
     allocate(ar35_r(1:nnx,Nbuf,Nbuf,Nmax), ar45_r(1:nnx,Nbuf,Nbuf,Nmax), ar36_r(1:nnx,Nbuf,Nbuf,Nmax), ar46_r(1:nnx,Nbuf,Nbuf,Nmax))
     
     
     allocate(coin135_e(Nbuf,Nbuf,Nbuf,Nmax),coin136_e(Nbuf,Nbuf,Nbuf,Nmax),coin145_e(Nbuf,Nbuf,Nbuf,Nmax),coin146_e(Nbuf,Nbuf,Nbuf,Nmax))
     allocate(coin135_r(Nbuf,Nbuf,Nbuf,Nmax),coin136_r(Nbuf,Nbuf,Nbuf,Nmax),coin145_r(Nbuf,Nbuf,Nbuf,Nmax),coin146_r(Nbuf,Nbuf,Nbuf,Nmax))
     allocate(coin235_e(Nbuf,Nbuf,Nbuf,Nmax),coin236_e(Nbuf,Nbuf,Nbuf,Nmax),coin245_e(Nbuf,Nbuf,Nbuf,Nmax),coin246_e(Nbuf,Nbuf,Nbuf,Nmax))
     allocate(coin235_r(Nbuf,Nbuf,Nbuf,Nmax),coin236_r(Nbuf,Nbuf,Nbuf,Nmax),coin245_r(Nbuf,Nbuf,Nbuf,Nmax),coin246_r(Nbuf,Nbuf,Nbuf,Nmax))
  endif
  return

end subroutine allocate_array_com_gen

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COMMUNICATIONS_GEN
!
!> Perform communications and return cputime used.
!<
subroutine communications_gen(Tab,n,tcomm)

  use parameters, only : nxmin,nxmax,ndim

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: ierr
  real :: tr1,tr2
#endif

  integer :: n
  real    :: tcomm
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),n) :: Tab

#if PARA==1
  tr1 = MPI_Wtime(ierr)
  if (ndim == 1) then
     call com_1D_gen(Tab,n)
  else if (ndim == 2) then
     call com_2D_gen(Tab,n)
  else if (ndim == 3) then
     call com_3D_gen(Tab,n)
  endif
  tr2 = MPI_Wtime(ierr)     
  tcomm = tcomm + tr2-tr1
#endif

  return

end subroutine communications_gen

!###########################################################
!###########################################################
!###########################################################
!###########################################################

#if PARA==1

!  Subroutine COM_1D_GEN
!
!> Perform generic 1D communications.
!<
subroutine com_1D_gen(Tab,n)

  use communication_gen
  use parameters
  use para
  use divers           , only : verbose

  implicit none

  include 'mpif.h'

  integer, parameter                    :: LEFT=1,RIGHT=2
  integer                               :: icount,ierr,n
  integer                               :: destination, emetteur
  integer, dimension(4  )               :: Req
  integer, dimension(MPI_STATUS_SIZE,4) :: Stat

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),n) :: Tab

  integer :: nnx

  if(verbose) write(*,*) 'Start of com_1d_gen'

  if(n>nmax) then
     write(*,*) "n for com_gen is ",n,"but nmax for com allocation is", nmax
     call close_program(mype,'called from com_1D_gen')
  endif

  nnx = nx(1) 

  icount   = Nbuf*n
  Req(1:4) = MPI_REQUEST_NULL

  face1_e(1:Nbuf,1,1,1:n) = Tab(           1:Nbuf ,1,1,1:n) 
  face2_e(1:Nbuf,1,1,1:n) = Tab(nnx-(Nbuf-1):nnx  ,1,1,1:n) 

  ! send data      
  destination = voisin(-1,0,0)
  call MPI_Isend(face1_e,icount,MPI_DOUBLE_PRECISION,destination,LEFT ,MPI_Comm_world,Req(1),ierr)

  destination = voisin(+1,0,0)
  call MPI_Isend(face2_e,icount,MPI_DOUBLE_PRECISION,destination,RIGHT,MPI_COMM_WORLD,Req(2),ierr)

  ! receive data
  emetteur = voisin(+1,0,0)
  call MPI_Irecv(face2_r,icount,MPI_DOUBLE_PRECISION,emetteur   ,LEFT ,MPI_COMM_WORLD,Req(3),ierr)
  
  emetteur = voisin(-1,0,0)
  call MPI_Irecv(face1_r,icount,MPI_DOUBLE_PRECISION,emetteur   ,RIGHT,MPI_COMM_WORLD,Req(4),ierr)
  
  call MPI_Waitall(4,Req,Stat,ierr)

  Tab( 1-Nbuf:0       ,1,1 ,1:n) = face1_r(1:Nbuf,1,1, 1:n)
  Tab(  nnx+1:nnx+Nbuf,1,1 ,1:n) = face2_r(1:Nbuf,1,1, 1:n)
    
  if(verbose) write(*,*) 'End of com_1d_gen'
    
  return
  
end subroutine com_1D_gen

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COM_2D_GEN
!
!> Perform generic 2D communications.
!<
subroutine com_2D_gen(Tab,n)

  use communication_gen
  use parameters
  use para
  use divers           , only : verbose

  implicit none

  include 'mpif.h'

  integer, parameter                     :: LEFT=1,RIGHT=2,DOWN=3, UP=4, RD=13, RU=14, LD=23, LU = 24
  integer                                :: ierr,n
  integer                                :: ic_f12,ic_f34,ic_coin
  integer                                :: destination, emetteur
  integer, dimension(16  )               :: Req
  integer, dimension(MPI_STATUS_SIZE,16) :: Stat

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),n) :: Tab

  integer :: nnx, nny

  if(verbose) write(*,*) 'Start of com_2d_gen'

  if(n>nmax) then
     write(*,*) "n for com_gen is ",n,"but nmax for com allocation is", nmax
     call close_program(mype,'called from com_2D_gen')
  endif

  nnx = nx(1) ; nny = nx(2) 

  ic_f12  = Nbuf*nny*n
  ic_f34  = Nbuf*nnx*n
  ic_coin = Nbuf*Nbuf  *n

  Req(1:16) = MPI_REQUEST_NULL

  ! 4 faces
  face1_e(1:Nbuf,1:nny ,1, 1:n) = Tab(           1:Nbuf ,            1:nny    ,1  , 1:n) 
  face2_e(1:Nbuf,1:nny ,1, 1:n) = Tab(nnx-(Nbuf-1):nnx  ,            1:nny    ,1  , 1:n) 
  face3_e(1:nnx ,1:Nbuf,1, 1:n) = Tab(           1:nnx  ,            1:Nbuf   ,1  , 1:n) 
  face4_e(1:nnx ,1:Nbuf,1, 1:n) = Tab(           1:nnx  , nny-(Nbuf-1):nny    ,1  , 1:n) 

  ! 4 corners
  coin13_e(1:Nbuf,1:Nbuf,1, 1:n) = Tab(           1:Nbuf ,            1:Nbuf  ,1  , 1:n) 
  coin23_e(1:Nbuf,1:Nbuf,1, 1:n) = Tab(nnx-(Nbuf-1):nnx  ,            1:Nbuf  ,1  , 1:n) 
  coin14_e(1:Nbuf,1:Nbuf,1, 1:n) = Tab(           1:Nbuf , nny-(Nbuf-1):nny   ,1  , 1:n) 
  coin24_e(1:Nbuf,1:Nbuf,1, 1:n) = Tab(nnx-(Nbuf-1):nnx  , nny-(Nbuf-1):nny   ,1  , 1:n) 


  ! send data      
  destination = voisin(-1, 0,0)
  call MPI_Isend(face1_e ,ic_f12 ,MPI_DOUBLE_PRECISION,destination,LEFT ,MPI_Comm_world,Req( 1),ierr)

  destination = voisin(+1, 0,0)
  call MPI_Isend(face2_e ,ic_f12 ,MPI_DOUBLE_PRECISION,destination,RIGHT,MPI_COMM_WORLD,Req( 2),ierr)

  destination = voisin( 0,-1,0)
  call MPI_Isend(face3_e ,ic_f34 ,MPI_DOUBLE_PRECISION,destination,DOWN ,MPI_COMM_WORLD,Req( 3),ierr)

  destination = voisin( 0,+1,0)
  call MPI_Isend(face4_e ,ic_f34 ,MPI_DOUBLE_PRECISION,destination,UP   ,MPI_COMM_WORLD,Req( 4),ierr)

  destination = voisin(-1,-1,0)
  call MPI_Isend(coin13_e,ic_coin,MPI_DOUBLE_PRECISION,destination,LD   ,MPI_COMM_WORLD,Req( 5),ierr)

  destination = voisin(-1,+1,0)
  call MPI_Isend(coin14_e,ic_coin,MPI_DOUBLE_PRECISION,destination,LU   ,MPI_COMM_WORLD,Req( 6),ierr)

  destination = voisin(+1,-1,0)
  call MPI_Isend(coin23_e,ic_coin,MPI_DOUBLE_PRECISION,destination,RD   ,MPI_COMM_WORLD,Req( 7),ierr)

  destination = voisin(+1,+1,0)
  call MPI_Isend(coin24_e,ic_coin,MPI_DOUBLE_PRECISION,destination,RU   ,MPI_COMM_WORLD,Req( 8),ierr)

  ! receive data
  emetteur = voisin(+1, 0,0)
  call MPI_Irecv(face2_r ,ic_f12 ,MPI_DOUBLE_PRECISION,emetteur   ,LEFT ,MPI_COMM_WORLD,Req( 9),ierr)
  
  emetteur = voisin(-1, 0,0)
  call MPI_Irecv(face1_r ,ic_f12 ,MPI_DOUBLE_PRECISION,emetteur   ,RIGHT,MPI_COMM_WORLD,Req(10),ierr)
  
  emetteur = voisin( 0,+1,0)
  call MPI_Irecv(face4_r ,ic_f34 ,MPI_DOUBLE_PRECISION,emetteur   ,DOWN ,MPI_COMM_WORLD,Req(11),ierr)
  
  emetteur = voisin( 0,-1,0)
  call MPI_Irecv(face3_r ,ic_f34 ,MPI_DOUBLE_PRECISION,emetteur   ,UP   ,MPI_COMM_WORLD,Req(12),ierr)
  

  emetteur = voisin(+1,+1,0)
  call MPI_Irecv(coin24_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,LD   ,MPI_COMM_WORLD,Req(13),ierr)
  
  emetteur = voisin(+1,-1,0)
  call MPI_Irecv(coin23_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,LU   ,MPI_COMM_WORLD,Req(14),ierr)

  emetteur = voisin(-1,+1,0)
  call MPI_Irecv(coin14_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,RD   ,MPI_COMM_WORLD,Req(15),ierr)

  emetteur = voisin(-1,-1,0)
  call MPI_Irecv(coin13_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,RU   ,MPI_COMM_WORLD,Req(16),ierr)
  

  call MPI_Waitall(16,Req,Stat,ierr)


  ! 4 faces
  Tab( 1-Nbuf:0       ,      1:nny     ,1 , 1:n) = face1_r(1:Nbuf ,1:nny   ,1, 1:n)
  Tab(  nnx+1:nnx+Nbuf,      1:nny     ,1 , 1:n) = face2_r(1:Nbuf ,1:nny   ,1, 1:n)
  Tab(      1:nnx     , 1-Nbuf:0       ,1 , 1:n) = face3_r(1:nnx  ,1:Nbuf  ,1, 1:n)
  Tab(      1:nnx     ,  nny+1:nny+Nbuf,1 , 1:n) = face4_r(1:nnx  ,1:Nbuf  ,1, 1:n)
  
  ! 4 corners
  Tab( 1-Nbuf:0       , 1-Nbuf:0       ,1 , 1:n) = coin13_r(1:Nbuf,1:Nbuf,1, 1:n)
  Tab( 1-Nbuf:0       ,  nny+1:nny+Nbuf,1 , 1:n) = coin14_r(1:Nbuf,1:Nbuf,1, 1:n)
  Tab(  nnx+1:nnx+Nbuf, 1-Nbuf:0       ,1 , 1:n) = coin23_r(1:Nbuf,1:Nbuf,1, 1:n)
  Tab(  nnx+1:nnx+Nbuf,  nny+1:nny+Nbuf,1 , 1:n) = coin24_r(1:Nbuf,1:Nbuf,1, 1:n)
  
  if(verbose) write(*,*) 'End of com_2d_gen'
  
  return
  
end subroutine com_2D_gen

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COM_3D_GEN
!
!> Perform generic 3D communications.
!<
subroutine com_3D_gen(Tab,n)

  use communication_gen
  use parameters
  use para
  use divers           , only : verbose

  implicit none

  include 'mpif.h'

  integer, parameter                     :: F1=1, F2=2, F3=3, F4=4, F5=5, F6=6
  integer, parameter                     :: AR13=13, AR14=14, AR15=15, AR16=16
  integer, parameter                     :: AR23=23, AR24=24, AR25=25, AR26=26
  integer, parameter                     :: AR35=35, AR36=36
  integer, parameter                     :: AR45=45, AR46=46
  integer, parameter                     :: C135=135, C136=136, C145=145, C146=146
  integer, parameter                     :: C235=235, C236=236, C245=245, C246=246
  integer                                :: ierr,n
  integer                                :: ic_f12,ic_f34,ic_f56, ic_arx, ic_ary, ic_arz, ic_coin
  integer                                :: destination, emetteur
  integer, dimension(52  )               :: Req
  integer, dimension(MPI_STATUS_SIZE,52) :: Stat

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),n) :: Tab

  integer :: nnx, nny, nnz

  if(verbose) write(*,*) 'Start of com_3d_gen'

  if(n>nmax) then
     write(*,*) "n for com_gen is ",n,"but nmax for com allocation is", nmax
     call close_program(mype,'called from com_3D_gen')
  endif

  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)

  ic_f12  = Nbuf*nny*nnz*n
  ic_f34  = Nbuf*nnx*nnz*n
  ic_f56  = Nbuf*nnx*nny*n

  ic_arx  = Nbuf*Nbuf*nnx  *n
  ic_ary  = Nbuf*Nbuf*nny  *n
  ic_arz  = Nbuf*Nbuf*nnz  *n

  ic_coin = Nbuf*Nbuf*Nbuf    *n

  Req(1:52) = MPI_REQUEST_NULL
  
  ! 6 faces
  face1_e(1:Nbuf,1:nny ,1:nnz , 1:n) = Tab(           1:Nbuf,            1:nny ,            1:nnz  , 1:n) 
  face2_e(1:Nbuf,1:nny ,1:nnz , 1:n) = Tab(nnx-(Nbuf-1):nnx ,            1:nny ,            1:nnz  , 1:n) 
  face3_e(1:nnx ,1:Nbuf,1:nnz , 1:n) = Tab(           1:nnx ,            1:Nbuf,            1:nnz  , 1:n) 
  face4_e(1:nnx ,1:Nbuf,1:nnz , 1:n) = Tab(           1:nnx , nny-(Nbuf-1):nny ,            1:nnz  , 1:n) 
  face5_e(1:nnx ,1:nny ,1:Nbuf, 1:n) = Tab(           1:nnx ,            1:nny ,            1:Nbuf , 1:n) 
  face6_e(1:nnx ,1:nny ,1:Nbuf, 1:n) = Tab(           1:nnx ,            1:nny , nnz-(Nbuf-1):nnz  , 1:n) 
  
  ! 12 edges
  ar13_e(1:Nbuf, 1:Nbuf, 1:nnz  , 1:n) = Tab(           1:Nbuf,            1:Nbuf,            1:nnz , 1:n)
  ar14_e(1:Nbuf, 1:Nbuf, 1:nnz  , 1:n) = Tab(           1:Nbuf, nny-(Nbuf-1):nny ,            1:nnz , 1:n)
  ar23_e(1:Nbuf, 1:Nbuf, 1:nnz  , 1:n) = Tab(nnx-(Nbuf-1):nnx ,            1:Nbuf,            1:nnz , 1:n)
  ar24_e(1:Nbuf, 1:Nbuf, 1:nnz  , 1:n) = Tab(nnx-(Nbuf-1):nnx , nny-(Nbuf-1):nny ,            1:nnz , 1:n)
  ar15_e(1:Nbuf, 1:nny , 1:Nbuf , 1:n) = Tab(           1:Nbuf,            1:nny ,            1:Nbuf, 1:n)
  ar25_e(1:Nbuf, 1:nny , 1:Nbuf , 1:n) = Tab(nnx-(Nbuf-1):nnx ,            1:nny ,            1:Nbuf, 1:n)
  ar16_e(1:Nbuf, 1:nny , 1:Nbuf , 1:n) = Tab(           1:Nbuf,            1:nny , nnz-(Nbuf-1):nnz , 1:n)
  ar26_e(1:Nbuf, 1:nny , 1:Nbuf , 1:n) = Tab(nnx-(Nbuf-1):nnx ,            1:nny , nnz-(Nbuf-1):nnz , 1:n)
  ar35_e(1:nnx , 1:Nbuf, 1:Nbuf , 1:n) = Tab(           1:nnx ,            1:Nbuf,            1:Nbuf, 1:n)
  ar45_e(1:nnx , 1:Nbuf, 1:Nbuf , 1:n) = Tab(           1:nnx , nny-(Nbuf-1):nny ,            1:Nbuf, 1:n)
  ar36_e(1:nnx , 1:Nbuf, 1:Nbuf , 1:n) = Tab(           1:nnx ,            1:Nbuf, nnz-(Nbuf-1):nnz , 1:n)
  ar46_e(1:nnx , 1:Nbuf, 1:Nbuf , 1:n) = Tab(           1:nnx , nny-(Nbuf-1):nny , nnz-(Nbuf-1):nnz , 1:n)

  ! 8 corners
  coin135_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) = Tab(           1:Nbuf,            1:Nbuf,            1:Nbuf, 1:n) 
  coin136_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) = Tab(           1:Nbuf,            1:Nbuf, nnz-(Nbuf-1):nnz , 1:n) 
  coin145_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) = Tab(           1:Nbuf, nny-(Nbuf-1):nny ,            1:Nbuf, 1:n) 
  coin146_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) = Tab(           1:Nbuf, nny-(Nbuf-1):nny , nnz-(Nbuf-1):nnz , 1:n) 
  coin235_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) = Tab(nnx-(Nbuf-1):nnx ,            1:Nbuf,            1:Nbuf, 1:n) 
  coin236_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) = Tab(nnx-(Nbuf-1):nnx ,            1:Nbuf, nnz-(Nbuf-1):nnz , 1:n) 
  coin245_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) = Tab(nnx-(Nbuf-1):nnx , nny-(Nbuf-1):nny ,            1:Nbuf, 1:n) 
  coin246_e(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) = Tab(nnx-(Nbuf-1):nnx , nny-(Nbuf-1):nny , nnz-(Nbuf-1):nnz , 1:n) 

  ! send data

  ! Faces
  destination = voisin(-1, 0, 0)
  call MPI_Isend(face1_e ,ic_f12, MPI_DOUBLE_PRECISION,destination,F1 ,MPI_Comm_world,Req( 1),ierr)

  destination = voisin(+1, 0, 0)
  call MPI_Isend(face2_e ,ic_f12, MPI_DOUBLE_PRECISION,destination,F2 ,MPI_COMM_WORLD,Req( 2),ierr)

  destination = voisin( 0,-1, 0)
  call MPI_Isend(face3_e ,ic_f34, MPI_DOUBLE_PRECISION,destination,F3 ,MPI_COMM_WORLD,Req( 3),ierr)

  destination = voisin( 0,+1, 0)
  call MPI_Isend(face4_e ,ic_f34, MPI_DOUBLE_PRECISION,destination,F4 ,MPI_COMM_WORLD,Req( 4),ierr)

  destination = voisin( 0, 0,-1)
  call MPI_Isend(face5_e ,ic_f56, MPI_DOUBLE_PRECISION,destination,F5 ,MPI_COMM_WORLD,Req( 5),ierr)

  destination = voisin( 0, 0,+1)
  call MPI_Isend(face6_e ,ic_f56, MPI_DOUBLE_PRECISION,destination,F6 ,MPI_COMM_WORLD,Req( 6),ierr)



  ! Edges
  destination = voisin(-1,-1,0)
  call MPI_Isend(ar13_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR13   ,MPI_COMM_WORLD,Req( 7),ierr)

  destination = voisin(-1,+1,0)
  call MPI_Isend(ar14_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR14   ,MPI_COMM_WORLD,Req( 8),ierr)

  destination = voisin(+1,-1,0)
  call MPI_Isend(ar23_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR23   ,MPI_COMM_WORLD,Req( 9),ierr)

  destination = voisin(+1,+1,0)
  call MPI_Isend(ar24_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR24   ,MPI_COMM_WORLD,Req(10),ierr)


  destination = voisin(-1, 0,-1)
  call MPI_Isend(ar15_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR15   ,MPI_COMM_WORLD,Req(11),ierr)

  destination = voisin(-1, 0,+1)
  call MPI_Isend(ar16_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR16   ,MPI_COMM_WORLD,Req(12),ierr)

  destination = voisin(+1, 0,-1)
  call MPI_Isend(ar25_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR25   ,MPI_COMM_WORLD,Req(13),ierr)

  destination = voisin(+1, 0,+1)
  call MPI_Isend(ar26_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR26   ,MPI_COMM_WORLD,Req(14),ierr)


  destination = voisin( 0,-1,-1)
  call MPI_Isend(ar35_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR35   ,MPI_COMM_WORLD,Req(15),ierr)

  destination = voisin( 0,-1,+1)
  call MPI_Isend(ar36_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR36   ,MPI_COMM_WORLD,Req(16),ierr)

  destination = voisin( 0,+1,-1)
  call MPI_Isend(ar45_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR45   ,MPI_COMM_WORLD,Req(17),ierr)

  destination = voisin( 0,+1,+1)
  call MPI_Isend(ar46_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR46   ,MPI_COMM_WORLD,Req(18),ierr)


  ! Corners
  destination = voisin( -1,-1,-1)
  call MPI_Isend(coin135_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C135   ,MPI_COMM_WORLD,Req(19),ierr)

  destination = voisin( -1,-1,+1)
  call MPI_Isend(coin136_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C136   ,MPI_COMM_WORLD,Req(20),ierr)

  destination = voisin( -1,+1,-1)
  call MPI_Isend(coin145_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C145   ,MPI_COMM_WORLD,Req(21),ierr)

  destination = voisin( -1,+1,+1)
  call MPI_Isend(coin146_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C146   ,MPI_COMM_WORLD,Req(22),ierr)

  destination = voisin( +1,-1,-1)
  call MPI_Isend(coin235_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C235   ,MPI_COMM_WORLD,Req(23),ierr)

  destination = voisin( +1,-1,+1)
  call MPI_Isend(coin236_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C236   ,MPI_COMM_WORLD,Req(24),ierr)

  destination = voisin( +1,+1,-1)
  call MPI_Isend(coin245_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C245   ,MPI_COMM_WORLD,Req(25),ierr)

  destination = voisin( +1,+1,+1)
  call MPI_Isend(coin246_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C246   ,MPI_COMM_WORLD,Req(26),ierr)

 

  ! receive data
  

  ! Faces
  emetteur = voisin(+1, 0, 0)
  call MPI_Irecv(face2_r ,ic_f12, MPI_DOUBLE_PRECISION,emetteur,F1 ,MPI_Comm_world,Req(27),ierr)

  emetteur = voisin(-1, 0, 0)
  call MPI_Irecv(face1_r ,ic_f12, MPI_DOUBLE_PRECISION,emetteur,F2 ,MPI_COMM_WORLD,Req(28),ierr)

  emetteur = voisin( 0,+1, 0)
  call MPI_Irecv(face4_r ,ic_f34, MPI_DOUBLE_PRECISION,emetteur,F3 ,MPI_COMM_WORLD,Req(29),ierr)

  emetteur = voisin( 0,-1, 0)
  call MPI_Irecv(face3_r ,ic_f34, MPI_DOUBLE_PRECISION,emetteur,F4 ,MPI_COMM_WORLD,Req(30),ierr)

  emetteur = voisin( 0, 0,+1)
  call MPI_Irecv(face6_r ,ic_f56, MPI_DOUBLE_PRECISION,emetteur,F5 ,MPI_COMM_WORLD,Req(31),ierr)

  emetteur = voisin( 0, 0,-1)
  call MPI_Irecv(face5_r ,ic_f56, MPI_DOUBLE_PRECISION,emetteur,F6 ,MPI_COMM_WORLD,Req(32),ierr)



  ! Edges
  emetteur = voisin(+1,+1,0)
  call MPI_Irecv(ar24_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR13   ,MPI_COMM_WORLD,Req(33),ierr)

  emetteur = voisin(+1,-1,0)
  call MPI_Irecv(ar23_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR14   ,MPI_COMM_WORLD,Req(34),ierr)

  emetteur = voisin(-1,+1,0)
  call MPI_Irecv(ar14_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR23   ,MPI_COMM_WORLD,Req(35),ierr)

  emetteur = voisin(-1,-1,0)
  call MPI_Irecv(ar13_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR24   ,MPI_COMM_WORLD,Req(36),ierr)


  emetteur = voisin(+1, 0,+1)
  call MPI_Irecv(ar26_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR15   ,MPI_COMM_WORLD,Req(37),ierr)

  emetteur = voisin(+1, 0,-1)
  call MPI_Irecv(ar25_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR16   ,MPI_COMM_WORLD,Req(38),ierr)

  emetteur = voisin(-1, 0,+1)
  call MPI_Irecv(ar16_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR25   ,MPI_COMM_WORLD,Req(39),ierr)

  emetteur = voisin(-1, 0,-1)
  call MPI_Irecv(ar15_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR26   ,MPI_COMM_WORLD,Req(40),ierr)


  emetteur = voisin( 0,+1,+1)
  call MPI_Irecv(ar46_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR35   ,MPI_COMM_WORLD,Req(41),ierr)

  emetteur = voisin( 0,+1,-1)
  call MPI_Irecv(ar45_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR36   ,MPI_COMM_WORLD,Req(42),ierr)

  emetteur = voisin( 0,-1,+1)
  call MPI_Irecv(ar36_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR45   ,MPI_COMM_WORLD,Req(43),ierr)

  emetteur = voisin( 0,-1,-1)
  call MPI_Irecv(ar35_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR46   ,MPI_COMM_WORLD,Req(44),ierr)


  ! Corners
  emetteur = voisin( +1,+1,+1)
  call MPI_Irecv(coin246_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C135   ,MPI_COMM_WORLD,Req(45),ierr)

  emetteur = voisin( +1,+1,-1)
  call MPI_Irecv(coin245_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C136   ,MPI_COMM_WORLD,Req(46),ierr)

  emetteur = voisin( +1,-1,+1)
  call MPI_Irecv(coin236_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C145   ,MPI_COMM_WORLD,Req(47),ierr)

  emetteur = voisin( +1,-1,-1)
  call MPI_Irecv(coin235_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C146   ,MPI_COMM_WORLD,Req(48),ierr)

  emetteur = voisin( -1,+1,+1)
  call MPI_Irecv(coin146_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C235   ,MPI_COMM_WORLD,Req(49),ierr)

  emetteur = voisin( -1,+1,-1)
  call MPI_Irecv(coin145_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C236   ,MPI_COMM_WORLD,Req(50),ierr)

  emetteur = voisin( -1,-1,+1)
  call MPI_Irecv(coin136_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C245   ,MPI_COMM_WORLD,Req(51),ierr)

  emetteur = voisin( -1,-1,-1)
  call MPI_Irecv(coin135_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C246   ,MPI_COMM_WORLD,Req(52),ierr)

  call MPI_Waitall(52,Req,Stat,ierr)

  call MPI_BARRIER(MPI_COMM_WORLD,ierr)


  ! 6 faces
  Tab( 1-Nbuf:0       ,     1:nny     ,      1:nnz     , 1:n) = face1_r(1:Nbuf,1:nny ,1:nnz , 1:n) 
  Tab(  nnx+1:nnx+Nbuf,     1:nny     ,      1:nnz     , 1:n) = face2_r(1:Nbuf,1:nny ,1:nnz , 1:n) 
  Tab(      1:nnx     ,1-Nbuf:0       ,      1:nnz     , 1:n) = face3_r(1:nnx ,1:Nbuf,1:nnz , 1:n)  
  Tab(      1:nnx     , nny+1:nny+Nbuf,      1:nnz     , 1:n) = face4_r(1:nnx ,1:Nbuf,1:nnz , 1:n) 
  Tab(      1:nnx     ,     1:nny     , 1-Nbuf:0       , 1:n) = face5_r(1:nnx ,1:nny ,1:Nbuf, 1:n) 
  Tab(      1:nnx     ,     1:nny     ,  nnz+1:nnz+Nbuf, 1:n) = face6_r(1:nnx ,1:nny ,1:Nbuf, 1:n) 
  
  ! 12 edges
  Tab( 1-Nbuf:0       , 1-Nbuf:0       ,      1:nnz      , 1:n) = ar13_r(1:Nbuf, 1:Nbuf, 1:nnz  , 1:n)
  Tab( 1-Nbuf:0       ,  nny+1:nny+Nbuf,      1:nnz      , 1:n) = ar14_r(1:Nbuf, 1:Nbuf, 1:nnz  , 1:n)
  Tab(  nnx+1:nnx+Nbuf, 1-Nbuf:0       ,      1:nnz      , 1:n) = ar23_r(1:Nbuf, 1:Nbuf, 1:nnz  , 1:n)
  Tab(  nnx+1:nnx+Nbuf,  nny+1:nny+Nbuf,      1:nnz      , 1:n) = ar24_r(1:Nbuf, 1:Nbuf, 1:nnz  , 1:n)
  Tab( 1-Nbuf:0       ,      1:nny     , 1-Nbuf:0        , 1:n) = ar15_r(1:Nbuf, 1:nny , 1:Nbuf , 1:n) 
  Tab(  nnx+1:nnx+Nbuf,      1:nny     , 1-Nbuf:0        , 1:n) = ar25_r(1:Nbuf, 1:nny , 1:Nbuf , 1:n)
  Tab( 1-Nbuf:0       ,      1:nny     ,  nnz+1:nnz+Nbuf , 1:n) = ar16_r(1:Nbuf, 1:nny , 1:Nbuf , 1:n)
  Tab  (nnx+1:nnx+Nbuf,      1:nny     ,  nnz+1:nnz+Nbuf , 1:n) = ar26_r(1:Nbuf, 1:nny , 1:Nbuf , 1:n)
  Tab(      1:nnx     , 1-Nbuf:0       , 1-Nbuf:0        , 1:n) = ar35_r(1:nnx , 1:Nbuf, 1:Nbuf , 1:n)
  Tab(      1:nnx     ,  nny+1:nny+Nbuf, 1-Nbuf:0        , 1:n) = ar45_r(1:nnx , 1:Nbuf, 1:Nbuf , 1:n) 
  Tab(      1:nnx     , 1-Nbuf:0       ,  nnz+1:nnz+Nbuf , 1:n) = ar36_r(1:nnx , 1:Nbuf, 1:Nbuf , 1:n) 
  Tab(      1:nnx     ,  nny+1:nny+Nbuf,  nnz+1:nnz+Nbuf , 1:n) = ar46_r(1:nnx , 1:Nbuf, 1:Nbuf , 1:n)

  ! 8 corners
  Tab(1-Nbuf:0       , 1-Nbuf:0       , 1-Nbuf:0          , 1:n) = coin135_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n)  
  Tab(1-Nbuf:0       , 1-Nbuf:0       ,  nnz+1:nnz+Nbuf   , 1:n) = coin136_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n)
  Tab(1-Nbuf:0       ,  nny+1:nny+Nbuf, 1-Nbuf:0          , 1:n) = coin145_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n)  
  Tab(1-Nbuf:0       ,  nny+1:nny+Nbuf,  nnz+1:nnz+Nbuf   , 1:n) = coin146_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) 
  Tab( nnx+1:nnx+Nbuf, 1-Nbuf:0       , 1-Nbuf:0          , 1:n) = coin235_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) 
  Tab( nnx+1:nnx+Nbuf, 1-Nbuf:0       ,  nnz+1:nnz+Nbuf   , 1:n) = coin236_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) 
  Tab( nnx+1:nnx+Nbuf,  nny+1:nny+Nbuf, 1-Nbuf:0          , 1:n) = coin245_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n) 
  Tab( nnx+1:nnx+Nbuf,  nny+1:nny+Nbuf,  nnz+1:nnz+Nbuf   , 1:n) = coin246_r(1:Nbuf, 1:Nbuf, 1:Nbuf, 1:n)

  if(verbose) write(*,*) 'End of com_3d_gen'
  
  return
  
end subroutine com_3D_gen

#endif
