!> \file
!! Contains subroutine restart_new()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine RESTART_NEW
!
!> Performs simulation restart. The data cube is read
!! in and the variables are set to the values from
!! the previous run.
!<
subroutine restart_new(k)

  use parameters
  use var
  use varold
  use varray
  use divers     , only : temps, dt, dtold, mu, RAY, DIFFU, CONDUC, GRAV, i_restart
  use divers_ray , only : ngrp_ray
  use unites
  use unites_sortie
  use gammas
  use geom       , only : x_glob, x,shift_gr, nshift_gr, geometrie
  use para       , only : mype,x_start,x_end,ncpu
  use cellules   , only : ioffset, ifich,Pcell, N_cell
  use diffusion  , only : ngrp_dif,Er_dif

  implicit none

  integer, dimension(3) :: nx_r
  integer               :: i,k,it,ncpu_r,nfx_r,nbuf_r,ngrp_ray_r,ngrp_dif_r
  character(len=23)     :: liste
  character(len= 5)     :: geometrie_r
  logical               :: mhd_log,mhd_log_r,ray_r,diffu_r,conduc_r,grav_r

  it = i_restart
  liste = 'XXX/liste-00000_000.Bin'
  
  call make_name(liste,mype,it)

  if (mype == 0) write(*,*) 'Reading file:',liste
  open(10,file=liste,form='unformatted')
  read(10) nx_glob,ncpu_r,geometrie_r

  if(ncpu.ne.ncpu_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'ncpu   =',ncpu
     write(*,*) 'ncpu_r =',ncpu_r
     call close_program(mype,'ncpu NOT EQ ncpu_r')
  endif

  if(geometrie.ne.geometrie_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'geometrie   = ',geometrie
     write(*,*) 'geometrie_r = ',geometrie_r
     call close_program(mype,'geometrie NOT EQ geometrie_r')
  endif

  read(10) x_glob(1:nx_glob(1)+1,1),x_glob(1:nx_glob(2)+1,2),x_glob(1:nx_glob(3)+1,3)
  read(10) ndim,nx_r,nfx_r,nbuf_r,ngrp_ray_r,ngrp_dif_r

  if(nfx.ne.nfx_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'nfx   = ',nfx
     write(*,*) 'nfx_r = ',nfx_r
     call close_program(mype,'nfx NOT EQ nfx_r')
  endif

  if(nbuf.ne.nbuf_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'nbuf   = ',nbuf
     write(*,*) 'nbuf_r = ',nbuf_r
     call close_program(mype,'nbuf NOT EQ nbuf_r')
  endif

  if(ngrp_ray.ne.ngrp_ray_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'ngrp_ray   = ',ngrp_ray
     write(*,*) 'ngrp_ray_r = ',ngrp_ray_r
     call close_program(mype,'ngrp_ray NOT EQ ngrp_ray_r')
  endif

  if(ngrp_dif.ne.ngrp_dif_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'ngrp_dif   = ',ngrp_dif
     write(*,*) 'ngrp_dif_r = ',ngrp_dif_r
     call close_program(mype,'ngrp_dif NOT EQ ngrp_dif_r')
  endif

  read(10) mhd_log_r,ray_r,diffu_r,conduc_r,grav_r

#if MHD==1
  mhd_log=.true.
#else
  mhd_log=.false.
#endif
  if(mhd_log.neqv.mhd_log_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'mhd   =',mhd_log
     write(*,*) 'mhd_r =',mhd_log_r
     call close_program(mype,'mhd_log NOT EQ mhd_log_r')
  endif

  if(ray.neqv.ray_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'ray   =',ray
     write(*,*) 'ray_r =',ray_r
     call close_program(mype,'ray NOT EQ ray_r')
  endif

  if(diffu.neqv.diffu_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'diffu   =',diffu
     write(*,*) 'diffu_r =',diffu_r
     call close_program(mype,'diffu NOT EQ diffu_r')
  endif

  if(conduc.neqv.conduc_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'conduc   =',conduc
     write(*,*) 'conduc_r =',conduc_r
     call close_program(mype,'conduc NOT EQ conduc_r')
  endif

  if(grav.neqv.grav_r) then
     write(*,*) 'Pb. in the restart file.'
     write(*,*) 'grav   =',grav
     write(*,*) 'grav_r =',grav_r
     call close_program(mype,'grav NOT EQ grav_r')
  endif

  read(10) x_start,x_end
  read(10) x(1:nx(1)+1,1),x(1:nx(2)+1,2),x(1:nx(3)+1,3)
  read(10)  shift_gr(1:ndim)
  read(10) nshift_gr(1:ndim)
  read(10) k
  read(10) temps, dt, dtold
  read(10) gamma,mu
  read(10) u_M,u_L,u_T
  read(10) rho 
  read(10) rhou
  read(10) E   
#if SCALP==1
  if(nfx.ge.1) read(10) fx  
#endif  
  if (RAY) then
     read(10) Eray 
     read(10) Fray 
  endif
  if (DIFFU) then
     read(10) Er_dif 
  endif
#if MHD==1
     read(10) B
#endif
  close(10)
  if (mype == 0) write(*,*) 'time (Myr) = ',temps/Mans
  if (mype == 0) write(*,*) 'End of reading'

  do i=1,3
     if (nx(i) .ne. nx_r(i)) then
        write(*,*) 'Pb. in the restart file.'
        write(*,*) 'idim  =',i
        write(*,*) 'nx    =', nx  (i)
        write(*,*) 'nx_r  =', nx_r(i)
        call close_program(mype,'nx NOT EQ nx_r')
     endif
  enddo

  ! Go back to code units
  u_m = u_M * gramme
  u_L = u_L * centimetre
  u_T = u_T * seconde

  u_Vol  = u_L**3
  u_dens = u_M/u_Vol
  u_Vit  = u_L/u_T
  u_E    = u_M * u_Vit**2
  u_Mom  = u_dens * u_Vit
  u_Evol = u_E/u_Vol
  u_Fr   = u_vit * u_Evol

  x      = x      * u_L
  x_glob = x_glob * u_L
  rho    = rho    * u_dens
  rhou   = rhou   * u_Mom
  E      = E      * u_Evol

  if (RAY) then
     Eray =Eray   * u_Evol
     Fray =Fray   * u_Fr
  endif
  if (DIFFU) then
     Er_dif =Er_dif * u_Evol
  endif
  
  temps=temps*u_T ; dt=dt*u_T ; dtold=dtold*u_T

  return

end subroutine restart_new
