!> \file
!! Contains subroutines init_diffusion() and
!! allocate_array_dif() and functions source_dif() and
!! derive_source_dif()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine INIT_DIFFUSION
!
!> Initialization for the FLD routines.
!<
subroutine init_diffusion

  use para     , only : mype
  use diffusion

  implicit none

  character(len=70) :: string

  namelist/param_dif/dif_imp,varrel_dif,vardt_dif,precond_dif,fld_limiter,ngrp_dif

  Dif_imp = .true.

  varrel_dif = 0.2
  vardt_dif  = 10.**(1./20.)
  
  fld_limiter = 'levermore'
  precond_dif     = .true.

  ngrp_dif = 1

  open  (1,file='liste.nml')
  rewind(1                 )
  read  (1,NML=param_dif   )
  close (1                 )

  if(ngrp_dif.ne.1) then
     write(string,'(a,i3)' ) ' M1 Multigroup not yet implemented, ngrp_dif=: ',ngrp_dif
     call close_program(mype,string)
  endif

  if(fld_limiter.ne.'minerbo'.and.fld_limiter.ne.'levermore'.and.fld_limiter.ne.'none') then
     call close_program(mype,'Unknown flux limiter for diffusion: '//fld_limiter)
  endif

  nitetot_dif = 0

  return

end subroutine init_diffusion

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine ALLOCATE_ARRAY_DIF
!
!> Allocates the arrays used in the FLD calculations.
!<
subroutine allocate_array_dif

  use parameters
  use diffusion
  use cl_dif

  !Boundary conditions allocations
  allocate(Er_dif_l(1:2*ndim))
  
  allocate(Er_dif(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)))

  return
  
end subroutine allocate_array_dif

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Function SOURCE_DIF
!
!> Energy ource term for the FLD.
!<
function source_dif(Temp)
  
  use unites, only : a_r
  
  implicit none

  real :: Source_dif,Temp

  source_dif = a_r*Temp**4

end function source_dif

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Function DERIVE_SOURCE_DIF
!
!> Derivative of the energy ource term for the FLD.
!<
function derive_source_dif(Temp)
  
  use unites, only : a_r
  
  implicit none

  real :: derive_Source_dif, Temp

  derive_Source_dif = 4.*a_r*Temp**3

end function derive_source_dif
