!> \file
!! Contains subroutine cal_b3d()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CAL_B3D
!
!> Computes the B member in 3D for the GMRES inversion.
!! Note: equations are written for alpha = 1.
!<
subroutine cal_b3d

  use parameters, only : ndim,nxmin,nxmax,nx,nvar_ray,pi
  use divers    , only : dt,verbose,hydro
  use cputime   , only : temps_comm
  use divers_ray, only : alpha,precond_ray,cal_valp,valp_min,rad_trans_model,slope_type_ray
  use unites    , only : c,c2
  use geom      , only : x,dx,ds,dv,xcloc,geometrie,cylindrique,spherique,geom_dir,dxc
  use var_loc   , only : xloc
  use varray    , only : eray,fray,erayold,frayold,kappa_emi_ray,kappa_abs_ray,kappa_sca_ray
  use var       , only : tgaz,rho
  use varold    , only : tgazold
  use tabgmres  , only : b,imin,imax,jmin,jmax,kmin,kmax
  use cl_ray    , only : icl_ray

  implicit none

#if PARA==1
  include 'mpif.h'
  real    :: tr1,tr2
  integer :: ierr
#endif

  integer, dimension(2,3,2,3) :: iv
  integer                     :: i,j,k,ivar,nn ,iimin,iimax,iipas,jjmin,jjmax,jjpas,iloc
  integer                     :: iface,idp,idm
  integer                     :: ip,jp,kp,im,jm,km,idim,jdim,kdim
  
  real, dimension(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+(kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray) :: xin
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim       ) :: peray,pfray1,pdummy
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim,1:ndim) :: pfray
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)              ) :: eray0,tgaz0
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim       ) :: fray0
  real, dimension(     2,ndim)         :: eraym,erayp,deray
  real, dimension(ndim,2,ndim)         :: fraym,frayp,dfray
  real, dimension(     2,ndim)         :: ffm,ffp
  real, dimension(1:3,1:3,2,ndim)      :: praym, prayp, deddp,deddm
  real, dimension(        2,ndim)      :: eps,ap,am,lm,lp,lmp
  real, dimension(3,3       )          :: Dedd,Dedd_dE
  real, dimension(3,3,1:ndim)          :: Dedd_dF
  real, dimension(3,3,       2,1:ndim) :: DeddM_dE, DeddP_dE
  real, dimension(3,3,1:ndim,2,1:ndim) :: DeddM_dF, DeddP_dF
  real, dimension(3,3       ,2)        :: Dedd_1d,Dedd_dE_1d,Dedd_2d,Dedd_dE_2d,Dedd_3d,Dedd_dE_3d
  real, dimension(3,3,1:ndim,2)        :: Dedd_dF_1d,Dedd_dF_2d,Dedd_dF_3d
  real, dimension(2,1:ndim)            :: dSloc
  real, dimension(       2,1:ndim)     :: flux_F
  real, dimension(1:ndim,2,1:ndim)     :: flux_P
  real, dimension(     3,2,3     )     :: Rayon
  real                                 :: flux_F_tot
  real, dimension(1:ndim         )     :: flux_P_tot
  real, dimension(       2       )     :: signe
  real, dimension(nvar_ray,2*nvar_ray) :: a1d,a2d,a3d
  real, dimension(1:5,1:5)             :: mat,inv
  real, dimension(1:5    )             :: vect
  real                                 :: drp,drm
  real                                 :: ff1,ff2
  real                                 :: thetaM,thetaP,interpol_valp
  real                                 :: xx,conv,source_ray,derive_source_ray

  logical :: chi_uniforme,derive_chi

!*********************************************************************************************************************************
!  Dedd,Dedd_dE (i,j)
!  Dedd_dF      (i,j,k)
!
!  DeddM_dE, DeddP_dE  (i,j  ,iface,idim)
!  DeddM_dF, DeddP_dF  (i,j,k,iface,idim)
!  
!  Dedd_1d,Dedd_dE_1d,Dedd_2d,Dedd_dE_2d,Dedd_3d,Dedd_dE_3d  (i,j  ,iface)
!  Dedd_dF_1d,Dedd_dF_2d,Dedd_dF_3d                          (i,j,k,iface)
!
!   iface : face index            (=1:   2) with 1 = back ; 2 = front
!   idim  : direction considered  (=1:ndim)
!
!   i,j   : tensorial coordinates       (=1:3   )
!   k     : flux direction considered   (=1:ndim)
!
!*********************************************************************************************************************************

  if (verbose) write(*,*) 'Start of cal_b3d'

  ! In order to have a uniform chi in the cells
  Chi_uniforme = .false.

  ! In order to include the derivatives in the matrix
  derive_chi = .true. ! besoin dans cal_Dedd

  ! incoming and outgoing fluxes
  signe(1) = -1. ; signe(2) = 1.

  ! geometry
  nn = max(nx(1),nx(2),nx(3))
  allocate(xloc(-1:nn+3,ndim))
  
                 xloc(-1:nx(1)+3,1) = x(-1:nx(1)+3,1)   
  if (ndim.gt.1) xloc(-1:nx(2)+3,2) = x(-1:nx(2)+3,2)   
  if (ndim.gt.2) xloc(-1:nx(3)+3,3) = x(-1:nx(3)+3,3)   

  call cal_geom(nx(1),nx(2),nx(3))

  ! neighbour offsets
  ! Iv(i,j,k,l) :: i = front or back face -- j = face direction -- k = front or back cell -- l = cell coordinate
  Iv = 0
  do idim = 1,3
     Iv(1,idim,1,idim) = -1
     Iv(2,idim,2,idim) = +1
  enddo

  b = 0.

#if PARA==1
  tr1 = MPI_Wtime(ierr)
  ! Communications
  if (ndim.eq.1) then
     call com_1D_xin(xin)
  else if (ndim.eq.2) then
     call com_2D_xin(xin)
  else if (ndim.eq.3) then
     call com_3D_xin(xin)
  endif
  tr2 = MPI_Wtime(ierr)
  temps_comm = temps_comm + tr2-tr1
#endif

  ! Boundary conditions
  call limites_xin(xin)

  call cal_kappa_ray(rho,Tgaz,kappa_emi_ray,kappa_sca_ray,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  kappa_abs_ray = kappa_emi_ray

  Eray0 = Erayold
  Fray0 = Frayold
  Tgaz0 = Tgazold


  if (verbose) write(*,*) '     -- Eray slope --'
  Call slopes(Eray0,peray,pdummy,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3),slope_type_ray)
  if (verbose) write(*,*) '     -- Fray slope --'
  do idim = 1,ndim
     Call slopes(Fray0(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),idim),pfray1,pdummy, &
                 nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3),slope_type_ray)
     pfray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),idim,1:ndim) = pfray1(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim)
  enddo


  do k = 1,nx(3)
     do j = 1,nx(2)
        do i = 1,nx(1)

           PrayM = 0. ; PrayP = 0.

           if (verbose) write(*,*) '     -- Interface values --'
   
           do idim  = 1,ndim
           do iface = 1,2
   
              im = i + Iv(iface,idim,1,1) ; jm = j + Iv(iface,idim,1,2); km = k + Iv(iface,idim,1,3)
              ip = i + Iv(iface,idim,2,1) ; jp = j + Iv(iface,idim,2,2); kp = k + Iv(iface,idim,2,3)
   
              if(idim.eq.1) then
                 idm = im
                 idp = ip
              else if(idim.eq.2) then
                 idm = jm
                 idp = jp
              else if(idim.eq.3) then
                 idm = km
                 idp = kp
              endif
   
              drp = 0.5*(x(idp  ,idim)-x(idm,idim))
              drm = 0.5*(x(idp+1,idim)-x(idp,idim)) 
   
              ErayM(iface,idim) = Eray0(im,jm,km) + drp*peray(im,jm,km,idim)  
              ErayP(iface,idim) = Eray0(ip,jp,kp) - drm*peray(ip,jp,kp,idim)
   
              if(iface.eq.1) then
                 dEray(iface,idim) = ErayP(iface,idim) - Eray0(i,j,k) 
              else
                 dEray(iface,idim) = ErayM(iface,idim) - Eray0(i,j,k) 
              endif
   
              ffM(iface,idim) = 0. ; ffP(iface,idim) = 0.
              do jdim = 1,ndim
                 FrayM(jdim,iface,idim) = Fray0(im,jm,km,jdim) + drp*pfray(im,jm,km,jdim,idim)
                 FrayP(jdim,iface,idim) = Fray0(ip,jp,kp,jdim) - drm*pfray(ip,jp,kp,jdim,idim)
   
                 if(iface.eq.1) then
                    dFray(jdim,iface,idim) = FrayP(jdim,iface,idim)- Fray0(i,j,k,jdim)
                 else
                    dFray(jdim,iface,idim) = FrayM(jdim,iface,idim)- Fray0(i,j,k,jdim)
                 endif
   
                 ffM(iface,idim) = ffM(iface,idim) + FrayM(jdim,iface,idim)**2
                 ffP(iface,idim) = ffP(iface,idim) + FrayP(jdim,iface,idim)**2
              enddo
              ffM(iface,idim) = sqrt(ffM(iface,idim))/(c*ErayM(iface,idim))
              ffP(iface,idim) = sqrt(ffP(iface,idim))/(c*ErayP(iface,idim))
   
              ! Calculate theta incidence angle to interpolate eigenvalues
              ! theta needs to be calculated here before ffM is changed if it is greater than 1
              if(ffM(iface,idim).gt.1.e-5) then
                 xx=FrayM(idim,iface,idim)/(c*ErayM(iface,idim))/ffM(iface,idim)
                 if(abs(xx).gt.1.) then
                    thetaM=pi/2.*(1.-sign(1.,xx))
                 else
                    thetaM=acos(xx)
                 endif
              else
                 thetaM=0.
              endif
              if(ffP(iface,idim).gt.1.e-5) then
                 xx=FrayP(idim,iface,idim)/(c*ErayP(iface,idim))/ffP(iface,idim)
                 if(abs(xx).gt.1.) then
                    thetaP=pi/2.*(1.-sign(1.,xx))
                 else
                    thetaP=acos(xx)
                 endif
              else
                 thetaP=0.
              endif
   
              ! If ffM > 1 then set it to 1 - WARNING: NOT VERY SATISFACTORY!! Ed
              if (ffM(iface,idim).gt. 1.) ffM(iface,idim)= 1.
              if (ffP(iface,idim).gt. 1.) ffP(iface,idim)= 1.
   
   
              ! Compute Eddington tensor and its derivatives at the cell centre and interfaces
              if (Chi_uniforme) then 
                 call cal_Dedd(Eray0(ip,jp,kp),Fray0(ip,jp,kp,1:ndim),DeddP(:,:,iface,idim),DeddP_dE(:,:,iface,idim),DeddP_dF(:,:,:,iface,idim) , Derive_chi)
                 call cal_Dedd(Eray0(im,jm,km),Fray0(im,jm,km,1:ndim),DeddM(:,:,iface,idim),DeddM_dE(:,:,iface,idim),DeddM_dF(:,:,:,iface,idim) , Derive_chi)
              else
                 call cal_Dedd(ErayP(iface,idim),FrayP(1:ndim,iface,idim),DeddP(:,:,iface,idim),DeddP_dE(:,:,iface,idim),DeddP_dF(:,:,:,iface,idim) , Derive_chi)
                 call cal_Dedd(ErayM(iface,idim),FrayM(1:ndim,iface,idim),DeddM(:,:,iface,idim),DeddM_dE(:,:,iface,idim),DeddM_dF(:,:,:,iface,idim) , Derive_chi)
              endif
   
              do jdim = 1,ndim
              do kdim = 1,ndim
                 PrayM(jdim,kdim,iface,idim) = DeddM(jdim,kdim,iface,idim)*ErayM(iface,idim)
                 PrayP(jdim,kdim,iface,idim) = DeddP(jdim,kdim,iface,idim)*ErayP(iface,idim)
              enddo
              enddo


              ! Compute epsilon for asymptotic preserving scheme
              if (verbose) write(*,*) '     -- compute epsilon --'
   
              ff1 = ffM(iface,idim)
              ff2 = ffP(iface,idim)
   
              if( ((kappa_abs_ray(im,jm,km)+kappa_sca_ray(im,jm,km))*rho(im,jm,km)).ne.0. .and. &
                  ((kappa_abs_ray(ip,jp,kp)+kappa_sca_ray(ip,jp,kp))*rho(ip,jp,kp)).ne.0.       ) then
                 eps(iface,idim) = max( 1./(((kappa_abs_ray(im,jm,km)+kappa_sca_ray(im,jm,km))*rho(im,jm,km))*(2.*drm)) &
                      &                ,1./(((kappa_abs_ray(ip,jp,kp)+kappa_sca_ray(ip,jp,kp))*rho(ip,jp,kp))*(2.*drp)) ) 
                 eps(iface,idim) = max(eps(iface,idim),2.*abs(ff1),2.*abs(ff2)) 
                 eps(iface,idim) = min(1.,eps(iface,idim)) 
              else
                 eps(iface,idim) = 1.
              endif
              
              ! Compute eigenvalues at the cell interfaces
              if (verbose)  write(*,*) '     -- compute eigenvalues --'
              
              if (cal_valp) then
   
                 select case(rad_trans_model)
         
                 case('P1')
         
                    am(iface,idim)=min(-valp_min,-1./sqrt(3.))
                    ap(iface,idim)=max(+valp_min, 1./sqrt(3.))
            
                 case('M1')
         
                    am(iface,idim)=min(-valp_min,interpol_valp(ffM(iface,idim),thetaM,eps(iface,idim),1), &
                                                 interpol_valp(ffP(iface,idim),thetaP,eps(iface,idim),1))
                    ap(iface,idim)=max(+valp_min,interpol_valp(ffM(iface,idim),thetaM,eps(iface,idim),4), &
                                                 interpol_valp(ffP(iface,idim),thetaP,eps(iface,idim),4))
         
                 end select
         
              else
                 am(iface,idim) = -1.
                 ap(iface,idim) =  1.
              endif
              
              if(ap(iface,idim).ne.am(iface,idim)) then
                 lm (iface,idim) = am(iface,idim)               /(ap(iface,idim)-am(iface,idim))
                 lp (iface,idim) =                ap(iface,idim)/(ap(iface,idim)-am(iface,idim))
                 lmp(iface,idim) = am(iface,idim)*ap(iface,idim)/(ap(iface,idim)-am(iface,idim))
              else !alors ap=am=0.
                 lm (iface,idim) = -1./2
                 lp (iface,idim) =  1./2
                 
                 lmp(iface,idim) =  0.
              endif
   
           enddo
           enddo

           !*************************************
           !         Chi derivatives
           !*************************************
           if (verbose)  write(*,*) '     -- Compute chi derivatives --'
   
           call cal_Dedd(Eray0(i-1,j,k),Fray0(i-1,j,k,1:ndim),Dedd_1d(:,:,1),Dedd_dE_1d(:,:,1),Dedd_dF_1d(:,:,:,1),Derive_chi )
           call cal_Dedd(Eray0(i  ,j,k),Fray0(i  ,j,k,1:ndim),Dedd          ,Dedd_dE          ,Dedd_dF            ,Derive_chi )
           call cal_Dedd(Eray0(i+1,j,k),Fray0(i+1,j,k,1:ndim),Dedd_1d(:,:,2),Dedd_dE_1d(:,:,2),Dedd_dF_1d(:,:,:,2),Derive_chi )
           if(ndim.gt.1) then
              call cal_Dedd(Eray0(i,j-1,k),Fray0(i,j-1,k,1:ndim),Dedd_2d(:,:,1),Dedd_dE_2d(:,:,1),Dedd_dF_2d(:,:,:,1),Derive_chi )
              call cal_Dedd(Eray0(i,j+1,k),Fray0(i,j+1,k,1:ndim),Dedd_2d(:,:,2),Dedd_dE_2d(:,:,2),Dedd_dF_2d(:,:,:,2),Derive_chi )
              if(ndim.gt.2) then
                 call cal_Dedd(Eray0(i,j,k-1),Fray0(i,j,k-1,1:ndim),Dedd_3d(:,:,1),Dedd_dE_3d(:,:,1),Dedd_dF_3d(:,:,:,1),Derive_chi )
                 call cal_Dedd(Eray0(i,j,k+1),Fray0(i,j,k+1,1:ndim),Dedd_3d(:,:,2),Dedd_dE_3d(:,:,2),Dedd_dF_3d(:,:,:,2),Derive_chi )
              endif
           endif

           ! Initialise geometry
           do idim = 1,ndim
              ip = i ; jp = j ; kp = k
              if(idim==1) ip = i+1
              if(idim==2) jp = j+1
              if(idim==3) kp = k+1
              dSloc(1,idim) = dS(i ,j ,k ,idim)
              dSloc(2,idim) = dS(ip,jp,kp,idim)
           enddo
   
           rayon = 1.
           if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') then
              rayon(2,1:2,1:3) = xcloc(i  ,1)
              rayon(2,1  ,1  ) =  xloc(i  ,1)
              rayon(2,2  ,1  ) =  xloc(i+1,1)
           endif
   
           if (geom_dir(3) == 'phi_s' ) then
              rayon(3,1  ,1  ) =  xloc(i  ,1) * sin(xcloc(j  ,2))
              rayon(3,2  ,1  ) =  xloc(i+1,1) * sin(xcloc(j  ,2))
              rayon(3,1  ,2  ) = xcloc(i  ,1) * sin( xloc(j  ,2))
              rayon(3,2  ,2  ) = xcloc(i  ,1) * sin( xloc(j+1,2))
              rayon(3,1  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
              rayon(3,2  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
           endif

           !******************************************************************
           !                     Compute fluxes
           !******************************************************************
   
           flux_F = 0.
           flux_P = 0.
           do idim=1,ndim
              flux_F(1,idim) = (+  lp (1,idim)            *FrayM(idim,1,idim)                 &
                                -  lm (1,idim)            *dFray(idim,1,idim)                 &
                                -c*lmp(1,idim)*eps(1,idim)*ErayM(     1,idim)                 &
                                +c*lmp(1,idim)*eps(1,idim)*dEray(     1,idim) )*dSloc(1,idim)
              flux_F(2,idim) = (-  lm (2,idim)            *FrayP(idim,2,idim)                 &
                                +  lp (2,idim)            *dFray(idim,2,idim)                 &
                                +c*lmp(2,idim)*eps(2,idim)*ErayP(     2,idim)                 &
                                -c*lmp(2,idim)*eps(2,idim)*dEray(     2,idim) )*dSloc(2,idim)
           enddo
   
           do jdim = 1,ndim
              iimin=1
              iimax=ndim
              iipas=1
              if (jdim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 iimin=2
                 idim=1
                 flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                         -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                         +c *  lmp(1,idim)*eps(1,idim)                          * dFray(jdim     ,1,idim) &
                                         +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)*(PrayP(jdim,idim,1,idim)-Dedd(jdim,idim)*Eray0(i,j,k)) ) *dV(i,j,k)/(xloc(i+1,1)-xloc(i,1))
                                     !ed +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)*(DeddP(jdim,idim,1,idim)*dEray(1,idim)               ) ) *dV(i,j,k)/(xloc(i+1,1)-xloc(i,1))
                 flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                         +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                         -c *  lmp(2,idim)*eps(2,idim)                          * dFray(jdim     ,2,idim) &
                                         +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)*(PrayM(jdim,idim,2,idim)-Dedd(jdim,idim)*Eray0(i,j,k)) ) *dV(i,j,k)/(xloc(i+1,1)-xloc(i,1))
                                     !ed +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)*(DeddM(jdim,idim,2,idim)*dEray(2,idim)               ) ) *dV(i,j,k)/(xloc(i+1,1)-xloc(i,1))
              endif
              if (jdim.eq.2 .and. geom_dir(2) == 'the_s') then
                 iipas=2
                 idim=2
                 flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                         -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                         +c *  lmp(1,idim)*eps(1,idim)                          * FrayP(jdim     ,1,idim) &
                                         +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayP(jdim,idim,1,idim) ) *dV(i,j,k)/(xloc(j+1,2)-xloc(j,2)) !/(xcloc(i,1)
                 flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                         +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                         -c *  lmp(2,idim)*eps(2,idim)                          * FrayM(jdim     ,2,idim) &
                                         +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayM(jdim,idim,2,idim) ) *dV(i,j,k)/(xloc(j+1,2)-xloc(j,2)) !/(xcloc(i,1)
              endif
              do idim=iimin,iimax,iipas
                 flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                         -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                         +c *  lmp(1,idim)*eps(1,idim)                          * dFray(jdim     ,1,idim) &
                                         +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)*(PrayP(jdim,idim,1,idim)-Dedd(jdim,idim)*Eray0(i,j,k)) ) *dSloc(1,idim) *rayon(jdim,1,idim)
                                     !ed +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)*(DeddP(jdim,idim,1,idim)*dEray(1,idim)               ) ) *dSloc(1,idim) *rayon(jdim,1,idim)
                 flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                         +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                         -c *  lmp(2,idim)*eps(2,idim)                          * dFray(jdim     ,2,idim) &
                                         +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)*(PrayM(jdim,idim,2,idim)-Dedd(jdim,idim)*Eray0(i,j,k)) ) *dSloc(2,idim) *rayon(jdim,2,idim)
                                     !ed +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)*(DeddM(jdim,idim,2,idim)*dEray(2,idim)               ) ) *dSloc(2,idim) *rayon(jdim,2,idim)
   
              enddo
           enddo


           flux_F_tot = 0.
           flux_P_tot = 0.
           do idim=1,ndim
              do iface=1,2
                 flux_F_tot = flux_F_tot + signe(iface)*dt*flux_F(iface,idim)
                 do jdim=1,ndim
                    flux_P_tot(jdim) = flux_P_tot(jdim) + signe(iface)*dt*flux_P(jdim,iface,idim) !*rayon(jdim,iface,idim)
                 enddo
              enddo
           enddo

           flux_F_tot = flux_F_tot/dV(i,j,k)
           do idim=1,ndim
              flux_P_tot(idim) = flux_P_tot(idim)/dV(i,j,k)/rayon(idim,2,3)
           enddo
   
           if (verbose)  write(*,*) '     -- compute matrix --'
   
           vect = 0.
           mat  = 0.
           do idim=1,5
              mat(idim,idim) = 1.
           enddo

           ! Er equation without the source term
           mat(2,1) = 0.
           mat(2,2) = 1. 
           do idim = 1,ndim
              mat(2,2     ) = mat(2,2) - dt/dV(i,j,k)*c*(lmp(1,idim)*eps(1,idim)*dSloc(1,idim)+lmp(2,idim)*eps(2,idim)*dSloc(2,idim))
              mat(2,2+idim) =            dt/dV(i,j,k)  *(lm (1,idim)            *dSloc(1,idim)+lp (2,idim)            *dSloc(2,idim))
           enddo
           
           vect(2) = -Erayold(i,j,k) + mat(2,2)*Eray0(i,j,k) + flux_F_tot
           do idim = 1,ndim
              vect(2) = vect(2) + mat(2,2+idim)*Fray0(i,j,k,idim)
           enddo

           ! Tgaz equation (using the coefficients from the Er equation without the source term)
           call heat_capacity(rho(i,j,k),Tgaz0(i,j,k),conv,1,1,1,1,1,1)
           conv=1./conv

           mat(1,1  ) = 1.
           mat(1,2:5) = mat(2,2:5)* conv 
           
           vect(1)    = vect(2)*conv -Tgazold(i,j,k)+mat(1,1)*Tgaz0(i,j,k)

           
           ! Finally include the energy source term: rho c (kappa_em aT^4-kappa_abs E)
            
           !  T4 = Tgazold(i,j,k)**4 *(1. + 4.*(Tgaz(i,j,k)-tgazold(i,j,k))/Tgazold(i,j,k)) 
           !  T4 = -3.*Tgazold(i,j,k)**4 + 4.* Tgazold(i,j,k)**3*Tgaz(i,j,k)

           mat(2,1) =          - dt*kappa_emi_ray(i,j,k)*rho(i,j,k)*c*derive_source_ray(Tgaz0(i,j,k))
           mat(2,2) = mat(2,2) + dt*kappa_abs_ray(i,j,k)*rho(i,j,k)*c
           vect (2) = vect (2) - dt*kappa_emi_ray(i,j,k)*rho(i,j,k)*c*source_ray(Tgaz0(i,j,k))
           vect (2) = vect (2) + dt*kappa_abs_ray(i,j,k)*rho(i,j,k)*c*Eray0(i,j,k) 
           

           ! Fr equation using only dtFr + c2(Div . Pr) = 0 terms
           do idim = 1,ndim
              mat(2+idim,1     ) = 0.
              mat(2+idim,2     ) = 0.
              mat(2+idim,2+idim) = 1.
              jjmin=1
              jjmax=ndim
              jjpas=1
              if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 jjmin=2
                 jdim=1
                 mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd(idim,jdim)
                 mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/(xloc(i+1,1)-xloc(i,1))* &
                      (lmp(1,jdim)*eps(1,jdim) &
                      +lmp(2,jdim)*eps(2,jdim) )  
              endif
              if(idim.eq.2 .and. geom_dir(2) == 'the_s') then
                 jjpas=2
                 jdim=1
                 mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd(idim,jdim)
                 mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                      (lmp(1,jdim)*eps(1,jdim) &
                      +lmp(2,jdim)*eps(2,jdim) )  
              endif
              do jdim = jjmin,jjmax,jjpas
                 mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/dV(i,j,k)* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd(idim,jdim)
                 !ed   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) * DeddM(idim,jdim,2,jdim ) &
                 !ed   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) * DeddP(idim,jdim,1,jdim ) ) 
                 
                 mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/dV(i,j,k)* &
                      (lmp(1,jdim)*eps(1,jdim)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) &
                      +lmp(2,jdim)*eps(2,jdim)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) )  
              enddo
           enddo


           ! Geometric sources
           if (cylindrique) then
              iloc=3
              if(geom_dir(2) == 'phi_c') iloc=2
              mat(3,2) = mat(3,2) + c2*dt/dV(i,j,k)* (Dedd(1,1) - Dedd(iloc,iloc)) * (dSloc(2,1)-dSloc(1,1))
           endif
           if (spherique) then
                                         mat(3,2) = mat(3,2) + c2*dt/dV(i,j,k)*                                  &
                   (2.*Dedd(1,1) - Dedd(2,2) - Dedd(3,3)) * 0.5*(dSloc(2,1)-dSloc(1,1))
              if(geom_dir(2) == 'the_s') mat(4,2) = mat(4,2) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))* &
                   (               Dedd(2,2) - Dedd(3,3)) * 0.5*(dSloc(2,1)-dSloc(1,1))
           endif


           do idim=1,ndim
              vect(2+idim) = -Frayold(i,j,k,idim) + flux_P_tot(idim)
              
              ! Include the flux source term: -kappa_abs rho c F
              mat(2+idim,2+idim) = mat(2+idim,2+idim) + alpha*c*dt*(kappa_abs_ray(i,j,k)+kappa_sca_ray(i,j,k))*rho(i,j,k)
              vect(2+idim)  = vect(2+idim)  !+ dt*(kappa_abs_ray(i,j,k)+kappa_sca_ray(i,j,k))*rho(i,j,k)*c*(1.-alpha)*Frayold(i,j,k,idim)

              vect(2+idim) = vect(2+idim) + mat(2+idim,2)*Eray0(i,j,k)
              do jdim=1,ndim
                 vect(2+idim) = vect(2+idim) + mat(2+idim,2+jdim)*Fray0(i,j,k,jdim)
              enddo
           enddo

           !*************************************************
           !     Derivative terms in the matrix
           !
           !      !!!!       CAUTION       !!!!
           ! To be done after mat(*,*)*Eray0 + mat(*,*)*Fray0
           !*************************************************
           ! Fr equation using only dtFr + c2(Div . Pr) = 0 terms
           do idim = 1,ndim
              jjmin=1
              jjmax=ndim
              jjpas=1
              if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 jjmin=2
                 jdim=1
                 mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dE(idim,jdim     )  ! d/dEray
                 do kdim=1,ndim
                    mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                         ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                         - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
                 enddo
              endif
              if(idim.eq.2 .and. geom_dir(2) == 'the_s') then
                 jjpas=2
                 jdim=2
                 mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dE(idim,jdim     )  ! d/dEray
                 do kdim=1,ndim
                    mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                         ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                         - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
                 enddo
              endif
              do jdim = jjmin,jjmax,jjpas
                 ! The Dedd derivatives
                 mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/dV(i,j,k)* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd_dE(idim,jdim     )  ! d/dEray
                 !ed  ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) * DeddM_dE(idim,jdim,2,jdim )&
                 !ed  - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) * DeddP_dE(idim,jdim,1,jdim ) ) ! d/dEray
                 do kdim=1,ndim
                    mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/dV(i,j,k)* &
                         ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                         - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
                    !ed    ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) * DeddM_dF(idim,jdim,kdim,2,jdim )&
                    !ed    - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) * DeddP_dF(idim,jdim,kdim,1,jdim ) )  ! d/dFray
                 enddo
              enddo
           enddo

           ! Geometric sources
           if(cylindrique) then
              iloc=3
              if(geom_dir(2) == 'phi_c') iloc=2
              mat(3,2     ) = mat(3,2     ) + c2*dt/dV(i,j,k)* (Dedd_dE(1,1) - Dedd_dE(iloc,iloc)) * (dSloc(2,1)-dSloc(1,1))
              do idim=1,ndim
                 mat(3,2+idim) = mat(3,2+idim) + c2*dt/dV(i,j,k)* (Dedd_dF(1,1,idim) - Dedd_dF(iloc,iloc,idim)) * (dSloc(2,1)-dSloc(1,1))
              enddo
           endif
           if (spherique) then
              mat(3,2     ) = mat(3,2     ) + c2*dt/dV(i,j,k)* (2.*Dedd_dE(1,1     ) - Dedd_dE(2,2     ) - Dedd_dE(3,3     )) * 0.5*(dSloc(2,1)-dSloc(1,1))
              do idim=1,ndim
                 mat(3,2+idim) = mat(3,2+idim) + c2*dt/dV(i,j,k)* (2.*Dedd_dF(1,1,idim) - Dedd_dF(2,2,idim) - Dedd_dF(3,3,idim)) * 0.5*(dSloc(2,1)-dSloc(1,1))
              enddo
              if(geom_dir(2) == 'the_s') then
                 mat(4,2     ) = mat(4,2     ) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*(Dedd_dE(2,2     ) - Dedd_dE(3,3     )) * 0.5*(dSloc(2,1)-dSloc(1,1))
                 do idim=1,ndim
                    mat(4,2+idim) = mat(4,2+idim) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*(Dedd_dF(2,2,idim) - Dedd_dF(3,3,idim)) * 0.5*(dSloc(2,1)-dSloc(1,1))
                 enddo
              endif
           endif

           !******************************************************************
           !                     End of Flux computations
           !******************************************************************
           
           ! non-diagonal terms
           a1d = 0. ; a2d = 0. ; a3d = 0.

           ! Temperature equation
           a1d(1,2)  =  dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*conv*dSloc(1,1) !*ErayM(1)   
           a1d(1,3)  = -dt/dV(i,j,k)*   lp(1,1)         *conv*dSloc(1,1) !*FrayM(1,1) 

           ! Radiative energy equation
           a1d(2,2)  =  dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*dSloc(1,1)!*ErayM(1)                
           a1d(2,3)  = -dt/dV(i,j,k)*   lp(1,1)         *dSloc(1,1)!*FrayM(1,1)

           ! Radiative flux equation
           do idim=1,ndim
              if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 a1d(2+idim,2) = -dt/(xloc(i+1,1)-xloc(i,1))*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) * (DeddM(idim,1,1,1)+Dedd_dE_1d(idim,1,1)) *Rayon(idim,1,1)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a1d(2+idim,2+jdim) = -dt/(xloc(i+1,1)-xloc(i,1)) *   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) *  Dedd_dF_1d(idim,1,jdim,1)      *Rayon(idim,1,1)/Rayon(idim,2,3)
                 enddo
                 a1d(2+idim,2+idim) = a1d(2+idim,2+idim) + dt/(xloc(i+1,1)-xloc(i,1))*c*lmp(1,1)*eps(1,1)                                                *Rayon(idim,1,1)/Rayon(idim,2,3)
              else
                 a1d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) * (DeddM(idim,1,1,1)+Dedd_dE_1d(idim,1,1))*dSloc(1,1) *Rayon(idim,1,1)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a1d(2+idim,2+jdim) = -dt/dV(i,j,k)*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) *  Dedd_dF_1d(idim,1,jdim,1)*dSloc(1,1)       *Rayon(idim,1,1)/Rayon(idim,2,3)
                 enddo
                 a1d(2+idim,2+idim) = a1d(2+idim,2+idim) + dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*dSloc(1,1)                                                *Rayon(idim,1,1)/Rayon(idim,2,3)
              endif
           enddo

           ! Temperature equation
           a1d(1,nvar_ray+2)  =  dt/dV(i,j,k)*c*lmp(2,1)*eps(2,1)*conv*dSloc(2,1) !ErayP(2)   
           a1d(1,nvar_ray+3)  = -dt/dV(i,j,k)*   lm(2,1)         *conv*dSloc(2,1) !FrayP(1,2) 

           ! Radiative energy equation
           a1d(2,nvar_ray+2)  =  dt/dV(i,j,k)*c*lmp(2,1)*eps(2,1)*dSloc(2,1) !*ErayP(2)
           a1d(2,nvar_ray+3)  = -dt/dV(i,j,k)*   lm(2,1)         *dSloc(2,1) !*FrayP(1,2)

           ! Radiative flux equation
           do idim=1,ndim
              if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 a1d(2+idim,nvar_ray+2) = dt/(xloc(i+1,1)-xloc(i,1))*   c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * (DeddP(idim,1,2,1)+Dedd_dE_1d(idim,1,2)) *Rayon(idim,2,1)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a1d(2+idim,nvar_ray+2+jdim) = dt/(xloc(i+1,1)-xloc(i,1))* c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * Dedd_dF_1d(idim,1,jdim,2)          *Rayon(idim,2,1)/Rayon(idim,2,3)
                 enddo
                 a1d(2+idim,nvar_ray+2+idim) = a1d(2+idim,nvar_ray+2+idim) + dt/(xloc(i+1,1)-xloc(i,1))*c* lmp(2,1)*eps(2,1)                                     *Rayon(idim,2,1)/Rayon(idim,2,3)
              else
                 a1d(2+idim,nvar_ray+2) = dt/dV(i,j,k)*   c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * (DeddP(idim,1,2,1)+Dedd_dE_1d(idim,1,2))*dSloc(2,1) *Rayon(idim,2,1)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a1d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)* c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * Dedd_dF_1d(idim,1,jdim,2)*dSloc(2,1)          *Rayon(idim,2,1)/Rayon(idim,2,3)
                 enddo
                 a1d(2+idim,nvar_ray+2+idim) = a1d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c* lmp(2,1)*eps(2,1)*dSloc(2,1)                                     *Rayon(idim,2,1)/Rayon(idim,2,3)
              endif
           enddo

           if(ndim.gt.1) then
              ! coupling with j-1 terms
              a2d(1,1) = 0.
              a2d(1,2) =  dt/dV(i,j,k)*c*lmp(1,2)*eps(1,2)*conv*dSloc(1,2)
              a2d(1,3) = 0.
              a2d(1,4) = -dt/dV(i,j,k)*   lp(1,2)         *conv*dSloc(1,2)

              a2d(2,1) = 0.
              a2d(2,2) =  dt/dV(i,j,k)*c*lmp(1,2)*eps(1,2)*dSloc(1,2)
              a2d(2,3) = 0.
              a2d(2,4) = -dt/dV(i,j,k)*   lp(1,2)         *dSloc(1,2)

              do idim=1,ndim
                 if(idim.eq.2 .and. geom_dir(2)=='the_s') then
                    a2d(2+idim,2) = -dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * (DeddM(idim,2,1,2)+Dedd_dE_2d(idim,2,1))  *Rayon(idim,1,2)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a2d(2+idim,2+jdim) = -dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,1)         *Rayon(idim,1,2)/Rayon(idim,2,3)
                    enddo
                    a2d(2+idim,2+idim) = a2d(2+idim,2+idim) + dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*c* lmp(1,2)*eps(1,2)                                                *Rayon(idim,1,2)/Rayon(idim,2,3)
                 else
                    a2d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * (DeddM(idim,2,1,2)+Dedd_dE_2d(idim,2,1))*dSloc(1,2)      *Rayon(idim,1,2)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a2d(2+idim,2+jdim) = -dt/dV(i,j,k)*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,1)*dSloc(1,2)             *Rayon(idim,1,2)/Rayon(idim,2,3)
                    enddo
                    a2d(2+idim,2+idim) = a2d(2+idim,2+idim) + dt/dV(i,j,k)*c* lmp(1,2)*eps(1,2)*dSloc(1,2)                                                    *Rayon(idim,1,2)/Rayon(idim,2,3)
                 endif
              enddo

              ! coupling with j+1 terms
              a2d(1,nvar_ray+1) = 0.
              a2d(1,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,2)*eps(2,2)*conv*dSloc(2,2)
              a2d(1,nvar_ray+3) = 0.
              a2d(1,nvar_ray+4) = -dt/dV(i,j,k)*   lm(2,2)         *conv*dSloc(2,2)

              a2d(2,nvar_ray+1) = 0.
              a2d(2,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,2)*eps(2,2)*dSloc(2,2)
              a2d(2,nvar_ray+3) = 0.
              a2d(2,nvar_ray+4) = -dt/dV(i,j,k)*   lm(2,2)         *dSloc(2,2)

              do idim=1,ndim
                 if(idim.eq.2 .and. geom_dir(2)=='the_s') then
                    a2d(2+idim,nvar_ray+2) = dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * (DeddP(idim,2,2,2)+Dedd_dE_2d(idim,2,2))*Rayon(idim,2,2)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a2d(2+idim,nvar_ray+2+jdim) = dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,2)       *Rayon(idim,2,2)/Rayon(idim,2,3)
                    enddo
                    a2d(2+idim,nvar_ray+2+idim) = a2d(2+idim,nvar_ray+2+idim) + dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*c* lmp(2,2)*eps(2,2)                                    *Rayon(idim,2,2)/Rayon(idim,2,3)
                 else
                    a2d(2+idim,nvar_ray+2) = dt/dV(i,j,k)*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * (DeddP(idim,2,2,2)+Dedd_dE_2d(idim,2,2))*dSloc(2,2)  *Rayon(idim,2,2)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a2d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,2)*dSloc(2,2)         *Rayon(idim,2,2)/Rayon(idim,2,3)
                    enddo
                    a2d(2+idim,nvar_ray+2+idim) = a2d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c* lmp(2,2)*eps(2,2)*dSloc(2,2)                                      *Rayon(idim,2,2)/Rayon(idim,2,3)
                 endif
              enddo

              if(ndim.gt.2) then
                 ! coupling with k-1 terms
                 a3d(1,1) = 0.
                 a3d(1,2) =  dt/dV(i,j,k)*c*lmp(1,3)*eps(1,3)*conv*dSloc(1,3)
                 a3d(1,3) = 0.
                 a3d(1,4) = 0.
                 a3d(1,5) = -dt/dV(i,j,k)*   lp(1,3)         *conv*dSloc(1,3)

                 a3d(2,1) = 0.
                 a3d(2,2) =  dt/dV(i,j,k)*c*lmp(1,3)*eps(1,3)*dSloc(1,3)
                 a3d(2,3) = 0.
                 a3d(2,4) = 0.
                 a3d(2,5) = -dt/dV(i,j,k)  * lp(1,3)         *dSloc(1,3)

                 do idim=1,ndim
                    a3d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,3)*eps(1,3)**2+(1.-eps(1,3)**2)/2 ) * (DeddM(idim,3,1,3)+Dedd_dE_3d(idim,3,1))*dSloc(1,3)      *Rayon(idim,1,3)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a3d(2+idim,2+jdim) =  -dt/dV(i,j,k)*   c2*( lp(1,3)*eps(1,3)**2+(1.-eps(1,3)**2)/2 ) * Dedd_dF_3d(idim,3,jdim,1)*dSloc(1,3)            *Rayon(idim,1,3)/Rayon(idim,2,3)
                    enddo
                    a3d(2+idim,2+idim) = a3d(2+idim,2+idim) + dt/dV(i,j,k)*c * lmp(1,3)*eps(1,3)*dSloc(1,3)                                                   *Rayon(idim,1,3)/Rayon(idim,2,3)
                 enddo

                 ! coupling with k+1 terms
                 a3d(1,nvar_ray+1) = 0.
                 a3d(1,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,3)*eps(2,3)*conv*dSloc(2,3)
                 a3d(1,nvar_ray+3) = 0.
                 a3d(1,nvar_ray+4) = 0.
                 a3d(1,nvar_ray+5) = -dt/dV(i,j,k)  * lm(2,3)         *conv*dSloc(2,3)

                 a3d(2,nvar_ray+1) = 0.
                 a3d(2,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,3)*eps(2,3)*dSloc(2,3)
                 a3d(2,nvar_ray+3) = 0.
                 a3d(2,nvar_ray+4) = 0.
                 a3d(2,nvar_ray+5) = -dt/dV(i,j,k)  * lm(2,3)         *dSloc(2,3)

                 do idim=1,ndim
                    a3d(2+idim,nvar_ray+2) =  dt/dV(i,j,k)*   c2*(-lm(2,3)*eps(2,3)**2+(1.-eps(2,3)**2)/2 )  *(DeddP(idim,3,2,3)+Dedd_dE_3d(idim,3,2))*dSloc(2,3) *Rayon(idim,2,3)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a3d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)*   c2*(-lm(2,3)*eps(2,3)**2+(1.-eps(2,3)**2)/2 )  *Dedd_dF_3d(idim,3,jdim,2)*dSloc(2,3)         *Rayon(idim,2,3)/Rayon(idim,2,3)
                    enddo
                    a3d(2+idim,nvar_ray+2+idim) = a3d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c * lmp(2,3)*eps(2,3)*dSloc(2,3)                                     *Rayon(idim,2,3)/Rayon(idim,2,3)
                 enddo
              endif
           endif

           !***********************************************************
           !              Boundary conditions
           !***********************************************************
           if(i==1) then
              if(icl_ray(1)==1) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,     1:     nvar_ray)
              else if(icl_ray(1)==3) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,     1:     nvar_ray)
                 mat(1:nvar_ray,3     ) = mat(1:nvar_ray,3     ) - 2.* a1d(1:nvar_ray,     3          )
              endif
              if(icl_ray(1).ne.2 .and. icl_ray(1).ne.0) a1d(1:nvar_ray,1:nvar_ray) = 0.
           else if(i==nx(1)) then
              if(icl_ray(2)==1) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
              else if(icl_ray(2)==3) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                 mat(1:nvar_ray,3     ) = mat(1:nvar_ray,3     ) - 2.* a1d(1:nvar_ray,nvar_ray+3          )
              endif
              if(icl_ray(2).ne.2 .and. icl_ray(2).ne.0) a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
           endif
           if(ndim.gt.1) then
              if(j==1) then
                 if(icl_ray(3)==1) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,     1:     nvar_ray)
                 else if(icl_ray(3)==3) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,     1:     nvar_ray)
                    mat(1:nvar_ray,4     ) = mat(1:nvar_ray,4     ) - 2.* a2d(1:nvar_ray,     4          )
                 endif
                 if(icl_ray(3).ne.2 .and. icl_ray(3).ne.0) a2d(1:nvar_ray,1:nvar_ray) = 0.
              else if(j==nx(2)) then
                 if(icl_ray(4)==1) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                 else if(icl_ray(4)==3) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                    mat(1:nvar_ray,4     ) = mat(1:nvar_ray,4     ) - 2.* a2d(1:nvar_ray,nvar_ray+4          )
                 endif
                 if(icl_ray(4).ne.2 .and. icl_ray(4).ne.0) a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
              endif
           endif
           if(ndim.gt.2) then
              if(k==1) then
                 if(icl_ray(5)==1) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,     1:     nvar_ray)
                 else if(icl_ray(5)==3) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,     1:     nvar_ray)
                    mat(1:nvar_ray,5     ) = mat(1:nvar_ray,5     ) - 2.* a3d(1:nvar_ray,     5          )
                 endif
                 if(icl_ray(5).ne.2 .and. icl_ray(5).ne.0) a3d(1:nvar_ray,1:nvar_ray) = 0.
              else if(k==nx(3)) then
                 if(icl_ray(6)==1) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                 else if(icl_ray(6)==3) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                    mat(1:nvar_ray,5     ) = mat(1:nvar_ray,5     ) - 2.* a3d(1:nvar_ray,nvar_ray+5          )
                 endif
                 if(icl_ray(6).ne.2 .and. icl_ray(6).ne.0) a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
              endif
           endif

           !**************  Normalised equations   ****************!
           !*******************************************************!
           !**        eqi / Xoldi et Aij=dFi/dXj *Xoldj          **!
           !**       to get     Aij (deltaX/Xold)j = eqi         **!
           !*******************************************************!
           !*******************************************************!
           vect(1  ) = vect(1  )/   Tgazold(i,j,k)
           vect(2  ) = vect(2  )/   Erayold(i,j,k)
           vect(3:5) = vect(3:5)/(c*Erayold(i,j,k))

           mat(1  ,1:5) = mat(1  ,1:5)/   Tgazold(i,j,k)
           mat(2  ,1:5) = mat(2  ,1:5)/   Erayold(i,j,k)
           mat(3:5,1:5) = mat(3:5,1:5)/(c*Erayold(i,j,k))

           mat(1:5,1  ) = mat(1:5,1  ) *   Tgazold(i,j,k)
           mat(1:5,2  ) = mat(1:5,2  ) *   Erayold(i,j,k)
           mat(1:5,3:5) = mat(1:5,3:5) * c*Erayold(i,j,k)

           if(precond_ray) then
              call essai_inv5x5(mat,inv)
              vect(1:nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),vect(1:nvar_ray))
           endif

           do ivar = 1,nvar_ray
           
              b((i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray+ivar) =  -vect(ivar)
!*******************************************************!
!!$              !* System A X/Xold=b and not A DX/Xold = b...
!!$              if(i-1.ge.1) then
!!$                 b(im) = b(im) + a(i1,1) + a(i1,2) + a(i1,3)*Frayold(i-1,j,k,1)/c/Erayold(i-1,j,k)
!!$              endif
!!$                 b(im) = b(im) + a(i1,4) + a(i1,5) + a(i1,6)*Frayold(i  ,j,k,1)/c/Erayold(i  ,j,k)
!!$              if(i+1.le.nx(1)) then
!!$                 b(im) = b(im) + a(i1,7) + a(i1,8) + a(i1,9)*Frayold(i+1,j,k,1)/c/Erayold(i+1,j,k)
!!$              endif
!*******************************************************!
           enddo


        enddo
     enddo
  enddo      ! end of loop over all cells


  ! Array deallocation
  deallocate(dx,dxc)
  deallocate(ds,dv)
  deallocate(xloc,xcloc)

  if (verbose) write(*,*) 'End of cal_b3d'
  
  return

end subroutine cal_b3d
