!> \file
!! Contains subroutines cal_x3d(), calmoinsun_x3d() and
!! allocate_gmres()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CAL_X3D
!
!> Computes the X member of the GMRES inversion algorithm.
!<
subroutine cal_x3d

  use parameters, only : nvar_ray,nx,ndim
  use varray    , only : Eray,Fray,Erayold,Frayold
  use var       , only : Tgaz
  use varold    , only : Tgazold
  use unites    , only : c
  use tabgmres  , only : xin,b,imin,imax,jmin,jmax,kmin,kmax

  implicit none

  integer :: i,j,k,index,idim

  do k = 1,nx(3)
  do j = 1,nx(2)
  do i = 1,nx(1)
  
     index = (i-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray
     xin(index+1) = ( Tgaz(i,j,k) - Tgazold(i,j,k) ) / Tgazold(i,j,k) 
     xin(index+2) = ( Eray(i,j,k) - Erayold(i,j,k) ) / Erayold(i,j,k) 
    
     do idim = 1,ndim
        xin(index+2+idim) = ( Fray(i,j,k,idim) - Frayold(i,j,k,idim) )/(c*Erayold(i,j,k))
     enddo
     
  enddo
  enddo
  enddo
  
  return

end subroutine cal_x3d

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CALMOINSUN_X3D
!
!> Computes the X member of the GMRES inversion algorithm.
!<
subroutine calmoinsun_x3d

  use parameters, only : nvar_ray,nx,ndim
  use varray    , only : Eray,Fray,Erayold,Frayold
  use var       , only : Tgaz
  use varold    , only : Tgazold
  use unites    , only : c
  use tabgmres  , only : xout,imin,imax,jmin,jmax,kmin,kmax
  use cputime   , only : temps_comm

  implicit none

#if PARA==1
  include 'mpif.h'
#endif

  integer :: i,j,k, idim,index

  do k = 1,nx(3)
  do j = 1,nx(2)
  do i = 1,nx(1)
  
     index = (i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray
     Tgaz(i,j,k) = (xout(index+1) + 1.) * Tgazold(i,j,k) 
     Eray(i,j,k) = (xout(index+2) + 1.) * Erayold(i,j,k)
     
     do idim = 1,ndim
        Fray(i,j,k,idim) = (xout(index+2+idim) + Frayold(i,j,k,idim)/c/Erayold(i,j,k) ) * c*Erayold(i,j,k) 
     enddo
     
  enddo
  enddo
  enddo

  call communications_ray(temps_comm)
  call limites_ray
  
  return

end subroutine calmoinsun_x3d

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine ALLOCATE_GMRES
!
!> Allocates the arrays required by the GMRES inversion
!! algorithm.
!<
subroutine allocate_gmres

  use parameters, only : nx,nvar_ray
  use tabgmres  , only : xin,xout,b,imin,imax,jmin,jmax,kmin,kmax

  implicit none

  imin = 0 ; imax = nx(1)+1
  if(nx(2).eq.1) then
     jmin = 1 ; jmax = 1
  else
     jmin = 0 ; jmax = nx(2)+1
  endif
  if(nx(3).eq.1) then
     kmin = 1 ; kmax = 1
  else
     kmin = 0 ; kmax = nx(3)+1
  endif

  allocate(xin(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+(kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray))
  allocate(xout(1: nx(1)*nx(2)*nx(3)*nvar_ray))
  allocate(b   (1: nx(1)*nx(2)*nx(3)*nvar_ray))

  xin = 0. ; xout = 0. ; b = 0.
  
  return

end subroutine allocate_gmres
