!> \file
!! Contains subroutines pasdt_ray() and pasdt_ray_exp()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PASDT_RAY
!
!> Computes the maximum allowed implicit timestep for the M1
!! radiative transfer.
!<
subroutine pasdt_ray

  use parameters
  use varold     , only : Tgazold
  use var        , only : Tgaz
  use unites     , only : c
  use geom       , only : x,geom_dir
  use divers_ray 
  use varray     , only : Eray,Erayold,Fray,Frayold
  use divers     , only : dt_max,temps,iout,tout,verbose,mu,dt,terminal_output,noutput
  use para       , only : mype,x_start

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info
  real    :: varmax_glob
#endif

  integer, dimension(1:ndim) :: idfray,jdfray,kdfray
  integer                    :: i,j,k,idim,ixmin,ixmax,iymin,iymax,izmin,izmax
  integer                    :: idtgaz,jdtgaz,kdtgaz,ideray,jderay,kderay

  real   , dimension(1:ndim) :: dfray,dfraymax,dfraymax_glob,dt_fray
  real                       :: dtgaz,dtgazmax,dtgazmax_glob,deray,deraymax,deraymax_glob
  real                       :: dt_tgaz,dt_eray,vardt_ray,varmax,dt_ray_imp_extrapole
  
  if(verbose) write(*,*) 'Start of pasdt_ray'

  if (dt_ray == 0.) then
     dt_ray     = dt_ray_exp
     dt_ray_old = dt_ray

     EXP_RAY = .true.
     IMP_RAY = .false.
     return
  endif

  dt_ray_old = dt_ray

!   if(mype==0) write(*,*) 'dt_ray =',dt_ray

  ! Relative variations of primary variables during the timestep
  dtgazmax = 0.
  deraymax = 0.
  dfraymax = 0. ; dfray = 0. ; idfray  = 1 ; jdfray = 1 ; kdfray = 1

  ixmin =  0 ; ixmax = nx(1)+1
  iymin =  1 ; iymax = 1
  izmin =  1 ; izmax = 1
  if(ndim.gt.1) then
     iymin =  0 ; iymax = nx(2)+1
  endif
  if(ndim.gt.2) then
     izmin =  0 ; izmax = nx(3)+1
  endif

  do k = izmin,izmax
  do j = iymin,iymax
  do i = ixmin,ixmax

     dtgaz    = Abs(Tgaz(i,j,k)/Tgazold(i,j,k)-1.)
     dtgazmax = Max(dtgazmax    ,dtgaz   )
     If (dtgaz == dtgazmax) then
        idtgaz = i ; jdtgaz = j ; kdtgaz = k 
     endif

     dEray = Abs(Eray(i,j,k)/Erayold(i,j,k)-1.)
     deraymax = Max(deraymax ,deray)
     If (deray == deraymax) then
        ideray = i ; jderay = j ; kderay = k 
     endif

     dfray = 0.           
     do idim = 1,ndim
        if (abs(Frayold(i,j,k,idim)) > 1.0E-2*c*Erayold(i,j,k) ) dFray(idim) = Abs(Fray(i,j,k,idim)/Frayold(i,j,k,idim) - 1.)
        dfraymax(idim)= Max(dfraymax(idim),dfray(idim))
        If (dfray(idim) == dfraymax(idim)) then
           idfray(idim) = i ; jdfray(idim) = j ; kdfray(idim) = k 
        endif
     enddo
     
  enddo
  enddo
  enddo
  
  varmax = max(deraymax,dtgazmax)
  do idim = 1,ndim
     varmax = max(varmax,dfraymax(idim))
  enddo
  
#if PARA==1

  ! Reduction of maxima variations
  info = 0
  call MPI_ALLREDUCE(dtgazmax,dtgazmax_glob,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)

  info = 0
  call MPI_ALLREDUCE(deraymax,deraymax_glob,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)

  do idim = 1,ndim
     info = 0
     call MPI_ALLREDUCE(dfraymax(idim),dfraymax_glob(idim),1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
  enddo
  
#else

  dtgazmax_glob = dtgazmax
  deraymax_glob = deraymax
  do idim = 1,ndim
     dfraymax_glob(idim) = dfraymax(idim)
  enddo
  
#endif

  if( (dtgazmax_glob == 0. .and. mype==0) .or. (dtgazmax_glob.ne.0. .and. dtgazmax_glob == dtgazmax)) then
     noutput = noutput + 1
     Write(terminal_output(noutput),601) idtgaz+x_start(1)-1,jdtgaz+x_start(2)-1,kdtgaz+x_start(3)-1,Tgaz(idtgaz,jdtgaz,kdtgaz),dtgazmax*100.
  endif
  if((deraymax_glob == 0. .and. mype==0) .or. (deraymax_glob.ne.0. .and. deraymax_glob == deraymax)) then
     noutput = noutput + 1
     Write(terminal_output(noutput),602) ideray+x_start(1)-1,jderay+x_start(2)-1,kderay+x_start(3)-1,Eray(ideray,jderay,kderay),deraymax*100.
  endif
  do idim = 1,ndim
     if((dfraymax_glob(idim) == 0. .and. mype==0) .or. (dfraymax_glob(idim).ne.0. .and. dfraymax_glob(idim) == dfraymax(idim))) then
        noutput = noutput + 1
        Write(terminal_output(noutput),603) idfray(idim)+x_start(1)-1,jdfray(idim)+x_start(2)-1,kdfray(idim)+x_start(3)-1, &
             &       Fray(idfray(idim),jdfray(idim),kdfray(idim),idim),dfraymax(idim)*100.,                                &
             &       Fray(idfray(idim),jdfray(idim),kdfray(idim),idim)/c/Eray(idfray(idim),jdfray(idim),kdfray(idim))
     endif
  enddo

601  Format('     Tgaz : ',I4,I4,I4,' Tgaz =',E12.4,'  dTgaz =',f9.4,' %'                 )
602  Format('     Eray : ',I4,I4,I4,' Eray =',E12.4,'  dEray =',f9.4,' %'                 )
603  Format('     Fray : ',I4,I4,I4,' Fray =',E12.4,'  dFray =',f9.4,' %','  F/cE  =',f9.4)

  noutput = noutput + 1
  write(terminal_output(noutput),'(a)') '**************************************************************************'

     !  Timestep gestion in function of relative variations
                                                                   vardt_ray = 10.**(1./30.) 
  if (varmax.le.varrel_ray/10.  .and. varmax.gt.varrel_ray/100. )  vardt_ray = 10.**(1./20.)
  if (varmax.le.varrel_ray/100. .and. varmax.gt.varrel_ray/1000.)  vardt_ray = 10.**(1./10.)
  if (varmax.le.varrel_ray/1000.                                )  vardt_ray = 10.**(1./5.)

  dt_tgaz  = vardt_ray*varrel_gaz*dt_ray/Max(vardt_ray*dtgazmax ,varrel_gaz)
  dt_eray  = vardt_ray*varrel_ray*dt_ray/Max(vardt_ray*deraymax ,varrel_ray)
  do idim = 1,ndim
     dt_fray(idim) = vardt_ray*varrel_ray*dt_ray/Max(vardt_ray*dfraymax(idim),varrel_ray)
  enddo

  dt_ray_imp = min(dt_eray,dt_tgaz)
  do idim = 1,ndim
     dt_ray_imp = min(dt_ray_imp,dt_fray(idim))
  enddo
 

#if PARA==1

  ! Timestep reduction
  info = 0
  call MPI_ALLREDUCE(dt_ray_imp, dt_ray_glob,1,MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,info)
  dt_ray_imp = dt_ray_glob

  info = 0
  call MPI_ALLREDUCE(varmax, varmax_glob,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)

  varmax = varmax_glob

#endif


  ! Implicit timestep that can be extrapolated
  dt_ray_imp_extrapole = dt_ray * varrel_ray/(varmax+1.0E-10)


  if (dt_ray_imp_extrapole > alpha_explicite * dt_ray_exp) then
     dt_ray = dt_ray_imp

     IMP_RAY = .true.
     EXP_RAY = .false.
  else
     dt_ray     = dt_ray_exp

     IMP_RAY = .false.
     EXP_RAY = .true.
  endif

  dt_ray = min(dt_ray,dt_max)

!  !Modification of timestep in order get temps.eq.tout(iout) (maximum variation of 20% allowed)
!  if(temps.ne.tout(iout)) then
!     if(temps+1.2*dt_ray.gt.tout(iout)) dt_ray=tout(iout)-temps
!  endif

  if(verbose) write(*,*) 'End of pasdt_ray'
  
  return
  
end subroutine pasdt_ray

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PASDT_RAY_EXP
!
!> Computes the maximum allowed explicit timestep for the M1
!! radiative transfer.
!<
subroutine pasdt_ray_exp

  use parameters
  use unites    , only : c,seconde
  use geom      , only : x,geom_dir
  use divers_ray 
  use para      , only : mype
  use divers    , only : verbose
 
  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info
#endif

  integer :: i,j,k,idim,ixmin,ixmax,iymin,iymax,izmin,izmax

  real, dimension(1:ndim) :: dx
  real                    :: xx
  
  if(verbose) write(*,*) 'Start of pasdt_ray_exp'

  dt_ray_exp = 1.0E100

  ixmin =  0 ; ixmax = nx(1)+1
  iymin =  1 ; iymax = 1
  izmin =  1 ; izmax = 1
  if(ndim.gt.1) then
     iymin =  0 ; iymax = nx(2)+1
  endif
  if(ndim.gt.2) then
     izmin =  0 ; izmax = nx(3)+1
  endif

  do k = izmin,izmax
  do j = iymin,iymax
  do i = ixmin,ixmax

                      dx(1) = x(i+1,1)-x(i,1)
     if (ndim .gt. 1) dx(2) = x(j+1,2)-x(j,2)
     if (ndim .gt. 2) dx(3) = x(k+1,3)-x(k,3)

     if (geom_dir(2) == 'phi_c') dx(2) = 0.5*(x(i,1)+x(i+1,1))*dx(2)
     if (geom_dir(2) == 'the_s') dx(2) = 0.5*(x(i,1)+x(i+1,1))*dx(2)
     if (geom_dir(3) == 'phi_s') dx(3) = 0.5*(x(i,1)+x(i+1,1))*sin(0.5*(x(j,2)+x(j+1,2))) * dx(3)

     xx = 0.
     do idim=1,ndim
        xx = xx + c /dx(idim)
     enddo

     dt_ray_exp = Min(1./xx, dt_ray_exp)
     
  enddo
  enddo
  enddo
  
  dt_ray_exp = 0.8 * dt_ray_exp 

#if PARA==1
  ! Timestep reduction
  info = 0
  call MPI_ALLREDUCE(dt_ray_exp, dt_ray_glob,1,MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,info)
  dt_ray_exp = dt_ray_glob
#endif

  if (mype == 0) write(*,*) 'dt_ray_exp =',dt_ray_exp/seconde, 's'

  if(verbose) write(*,*) 'End of pasdt_ray_exp'

  return
  
end subroutine pasdt_ray_exp
