!     ---------------
subroutine INIT
!     ---------------
  use parameters
  use unites
  use divers
  use var
  use para
  use cl
  use geom  , only : x,x_glob,shift_gr,lbox,box_min,box_max,Cartesien, Cylindrique, Spherique 
  implicit none

  integer :: i,j,k,idim,i1,i2,nx_dim
  real, dimension(1:3) :: dx

  real :: vol

  namelist/condinit/Lbox,box_min,box_max
!----------------------
  
! Begin time
  temps = 0.

  mu=1.

  box_min = 0. ; box_max = 1. ; Lbox = 1.

  ! Reading the namelist (previous values are erased)
  if(mype == 0) write(*,'(a)') 'reading namelist CONDINIT'
  open  (1,file='liste.nml')
  rewind(1                 )
  read  (1,NML=condinit    )
  close (1                 )

  tend   = tend   * seconde
  dt_max = dt_max * seconde

  Lbox(2) = Pi/2.  ! To simplify the namelist
  Lbox(3) = Pi/2.
  
  box_min(1) = 0.00
  box_min(2) = Pi*(1./2. - 1./4.) 
  box_min(2) = 0.

!  Global mesh
!  ~~~~~~~~~~~
  do idim = 1,3
     nx_dim = nx_glob(idim)
     
     x_glob(       1,idim) = box_min(idim)
     x_glob(nx_dim+1,idim) = box_min(idim) + Lbox(idim)
     
     dx(idim) = (x_glob(nx_dim+1,idim) - x_glob(1,idim))/float(nx_dim)
     do i=2,nx_dim
        x_glob(i,idim) = x_glob(i-1,idim) + dx(idim)
     enddo
  enddo

  shift_gr(1:ndim) = 0.

!  Local mesh
!  ~~~~~~~~~~
  
  do idim = 1,3
     i1 = 1
     if (icl(2*idim-1) == 0) i1 = 1-nbuf
     i2 = nx(idim)+1
     if (icl(2*idim  ) == 0) i2 = nx(idim)+1+nbuf

     do i=i1,i2
        x(i,idim) = x_glob(x_start(idim)+i-1,idim)
     enddo
  enddo

  ! Sedov setup
  ! ~~~~~~~~~~~ 

  ! Normalization of the energy depending on the geometry

  ! rc 
  if (cylindrique)  vol = pi*x_glob(2,1)**2

  ! x  
  if (cartesien)    vol = dx(1)
   
  ! rs 
  if(spherique)     vol = 4.*pi*x_glob(2,1)**3/3. 

  write(*,*) 'Vol =',vol

  do k=1,nx(3)
  do j=1,nx(2)
  do i=1,nx(1)
     if (x(i,1) < 0.8*dx(1)) then 
        E(i,j,k) = 1./vol
     else
        E(i,j,k) = 1.0E-12/vol
     endif
     rhou(i,j,k,1:ndim) = 0.
     rho (i,j,k       ) = 1.
  enddo
  enddo
  enddo

  return
  
end subroutine INIT

!*******************************************************************************************************************

!     ---------------------
subroutine DIMENSIONS
!     ---------------------
  use unites
  use gammas

  implicit none

  ! code units
  unitM = 1.0E+00 *gramme
  unitL = 1.0E+00 *centimetre
  unitT = 1.0E+00 *seconde
  unitD = unitM/unitL**3

  ! Specific heats ratio Cp/Cv
  gamma = 1.4
  
  return
  
end subroutine DIMENSIONS

