!     ---------------
subroutine INIT
!     ---------------
  use parameters 
  use unites
  use divers
  use var
  use geom        , only : x, x_glob, lbox
  use para        , only : mype, x_start
  use gravity
  use cl
  
  implicit none

#if PARA==1
  include 'mpif.h'
#endif

  integer :: i, j, k, idim, nx_dim, i1, i2, nb_case_loc, nb_case, distribution_type, info
  Real, dimension(3) :: dx, dx_from_m, xc
  real :: dv, xm, ym, zm, a,  m
  

  namelist/condinit/lbox, xm, ym, zm, a, c, m, distribution_type
  !----------------------

  if (verbose)  write(*,*) 'debut de init'

  ! distance in kilometre, mass in solar mass 
  !(xm,ym,zm) centre of mass of the distribution of total mass m
  ! for sphere radius = a, for spheroid radii long = a small = c 
  
  lbox(1) = 10.
  lbox(2) = 10.
  lbox(3) = 10.
  xm      = 5.
  ym      = 5.
  zm      = 5.
  a       = 2.5
  c       = 1.25
  m       = 1.
  
  ! distribution_type : 0=ponctual   , 1=sphere     , 2=spheroid
  distribution_type = 0
  

  ! Reading the namelist (previous values are erased)
  if(mype == 0) write(*,'(a)') 'reading namelist CONDINIT'
  open  (1,file='liste.nml')
  rewind(1                 )
  read  (1,NML=condinit    )
  close (1                 )

  tend   = tend   * seconde
  dt_max = dt_max * seconde

  !  Units 
  !  ~~~~~
  
  lbox(1) = lbox(1) * 1000 * metre
  lbox(2) = lbox(2) * 1000 * metre
  lbox(3) = lbox(3) * 1000 * metre
  xm      = xm      * 1000 * metre
  ym      = ym      * 1000 * metre
  zm      = zm      * 1000 * metre
  a       = a       * 1000 * metre
  c       = c       * 1000 * metre
  m       = m       * Msol
  tend    = tend    * seconde
  
!  maillage global
!  ~~~~~~~~~~~~~~
  do idim = 1,3
     nx_dim = nx_glob(idim)
     
     x_glob(       1,idim) = 0
     x_glob(nx_dim+1,idim) = Lbox(idim)
     
     dx(idim) = (x_glob(nx_dim+1,idim) - x_glob(1,idim))/float(nx_dim)
     do i=2,nx_dim
        x_glob(i,idim) = x_glob(i-1,idim) + dx(idim)
     enddo

  enddo

!  maillage local
!  ~~~~~~~~~~~~~~
  
  do idim = 1,3

     i1 = 1
     if (icl(2*idim-1) == 0) i1 = 1-nbuf
     i2 = nx(idim)+1
     if (icl(2*idim  ) == 0) i2 = nx(idim)+1+nbuf

     do i=i1,i2
        x(i,idim) = x_glob(x_start(idim)+i-1,idim)
     enddo

  enddo


! Ponctual mass 
! ~~~~~~~~~~~~~~
  
if (distribution_type==0) then
   
     dx          = 0.
     dv          = 0.

     do k=1,nx(3)
        do j=1,nx(2)
           do i=1,nx(1)
           
              dx(1)=x(i+1,1)-x(i,1)
              dx(2)=x(j+1,2)-x(j,2)
              dx(3)=x(k+1,3)-x(k,3)
           
              dv=dx(1)*dx(2)*dx(3)
            
              if ((xm.LT.x(i+1,1)).AND.(xm.GE.x(i,1)).AND.&
                  (ym.LT.x(j+1,2)).AND.(ym.GE.x(j,2)).AND.&
                  (zm.LT.x(k+1,3)).AND.(zm.GE.x(k,3)))then
                
                 rho (i,j,k) = m/(dv)
             
              else
              
                 rho(i,j,k) = 0.

              endif
           enddo
        enddo
     enddo

  ! Spheroid 
  ! ~~~~~~~~

  else if (distribution_type==2) then

     dx          = 0.
     dv          = 0.
     nb_case_loc = 0
     nb_case     = 0
 
     do k=1,nx(3)
        do j=1,nx(2)
           do i=1,nx(1)
              
              xc(1)=0.5*(x(i,1)+x(i+1,1))
              xc(2)=0.5*(x(j,2)+x(j+1,2))
              xc(3)=0.5*(x(k,3)+x(k+1,3))

              dx_from_m(1)=abs(xc(1)-xm)
              dx_from_m(2)=abs(xc(2)-ym)
              dx_from_m(3)=abs(xc(3)-zm)
           
              if (dx_from_m(1)**2/a**2+(dx_from_m(2))**2/a**2+(dx_from_m(3))**2/c**2.LT.1) then
                 
                 nb_case_loc = nb_case_loc + 1  
              
              endif
              
           enddo
        enddo
     enddo

#if PARA==1
     info = 0 
     call MPI_ALLREDUCE(nb_case_loc, nb_case ,1,MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,info)
#else
     nb_case = nb_case_loc
#endif
     
     do k=1,nx(3)
        do j=1,nx(2)
           do i=1,nx(1)
              
              dx(1)=x(i+1,1)-x(i,1)
              dx(2)=x(j+1,2)-x(j,2)
              dx(3)=x(k+1,3)-x(k,3)
              
              xc(1)=0.5*(x(i,1)+x(i+1,1))
              xc(2)=0.5*(x(j,2)+x(j+1,2))
              xc(3)=0.5*(x(k,3)+x(k+1,3))
           
              dx_from_m(1)=abs(xc(1)-xm)
              dx_from_m(2)=abs(xc(2)-ym)
              dx_from_m(3)=abs(xc(3)-zm)
           
              dv=dx(1)*dx(2)*dx(3)
           
              if (dx_from_m(1)**2/a**2+(dx_from_m(2))**2/a**2+(dx_from_m(3))**2/c**2.LT.1) then
                 
                 rho (i,j,k) = m/(nb_case*dv)
                     
              else
              
                 rho (i,j,k) = 0
        
              endif
           enddo
        enddo
     enddo
     


  ! Sphere 
  ! ~~~~~~~~

  else if (distribution_type==1) then

     dx          = 0.
     dv          = 0.
     nb_case_loc = 0
     nb_case     = 0
 
     do k=1,nx(3)
        do j=1,nx(2)
           do i=1,nx(1)
              
              xc(1)=0.5*(x(i,1)+x(i+1,1))
              xc(2)=0.5*(x(j,2)+x(j+1,2))
              xc(3)=0.5*(x(k,3)+x(k+1,3))

              dx_from_m(1)=abs(xc(1)-xm)
              dx_from_m(2)=abs(xc(2)-ym)
              dx_from_m(3)=abs(xc(3)-zm)
              
              if (dx_from_m(1)**2/a**2+(dx_from_m(2))**2/a**2+(dx_from_m(3))**2/a**2.LT.1) then
                 
                 nb_case_loc = nb_case_loc + 1  
              
              endif
              
           enddo
        enddo
     enddo

#if PARA==1
     info = 0 
     call MPI_ALLREDUCE(nb_case_loc, nb_case ,1,MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,info)
#else
     nb_case = nb_case_loc
#endif

     do k=1,nx(3)
        do j=1,nx(2)
           do i=1,nx(1)
     
              dx(1)=x(i+1,1)-x(i,1)
              dx(2)=x(j+1,2)-x(j,2)
              dx(3)=x(k+1,3)-x(k,3)

              xc(1)=0.5*(x(i,1)+x(i+1,1))
              xc(2)=0.5*(x(j,2)+x(j+1,2))
              xc(3)=0.5*(x(k,3)+x(k+1,3))
           
              dx_from_m(1)=abs(xc(1)-xm)
              dx_from_m(2)=abs(xc(2)-ym)
              dx_from_m(3)=abs(xc(3)-zm)
           
              dv=dx(1)*dx(2)*dx(3)
              
              if (dx_from_m(1)**2/a**2+(dx_from_m(2))**2/a**2+(dx_from_m(3))**2/a**2.LT.1) then
                
                 rho (i,j,k) = m/(nb_case*dv)
                     
              else
              
                 rho (i,j,k) = 0
        
              endif
           enddo
        enddo
     enddo
     
  else
     
     call close_program(mype,'Not a valid mass distribution')

  endif
    

if (verbose)  write(*,*) 'fin de init'

  return
  
end subroutine INIT

!*******************************************************************************************************************

!     ---------------------
subroutine DIMENSIONS
!     ---------------------
  use unites
  use gammas

  implicit none

  ! code units
  unitM = 1.0E+00 *gramme
  unitL = 1.0E+00 *centimetre
  unitT = 1.0E+00 *seconde
  unitD = unitM/unitL**3

  ! Specific heats ratio Cp/Cv
  gamma = 5.0/3.0
  
  return
  
end subroutine DIMENSIONS
