!     ---------------
subroutine INIT
!     ---------------
  use parameters
  use unites
  use divers
  use var
  use gammas
  use para
  use cl
  use geom  , only : x,x_glob,lbox
  implicit none

  integer :: i,j,k,idim,i1,i2
  real, dimension(1:3) :: dx

  real :: xx

  integer :: ii,nx_dim,idir

  real    :: rho1,u1,P1,rho2,u2,P2,E1,E2

  namelist/condinit/rho1,u1,P1,rho2,u2,P2,Lbox,idir,isotherme
!----------------------

! Begin time
  temps = 0.

  mu     =  1.
  P1     =  1.
  P2     =  1.

  !direction of the test
  idir=1

  ! Sod
  ! ~~~
  rho1   =  1.
  u1     =  0.   
  P1     =  1.   

  rho2   =  0.125 
  u2     =  0.   
  P2     =  0.1   

  isotherme=.false.
  cs_iso   = 1.

  ! Reading the namelist (previous values are erased)
  if(mype == 0) write(*,'(a)') 'reading namelist CONDINIT'
  open  (1,file='liste.nml')
  rewind(1                 )
  read  (1,NML=condinit    )
  close (1                 )

  tend   = tend   * seconde
  dt_max = dt_max * seconde

  Lbox = Lbox*centimetre

  if(isotherme) then
     cs_iso2  = cs_iso**2
     P1     =  rho1*cs_iso2
     P2     =  rho2*cs_iso2
     E1     =  P1/(gamma-1.) + 0.5*rho1*u1*u1 
     E2     =  P2/(gamma-1.) + 0.5*rho2*u2*u2 
  else
     E1     =  P1/(gamma-1.) + 0.5*rho1*u1*u1 
     E2     =  P2/(gamma-1.) + 0.5*rho2*u2*u2 
  endif

  !  Global mesh
  !  ~~~~~~~~~~~
  do idim = 1,3
     nx_dim = nx_glob(idim)
     
     x_glob(       1,idim) = 0. 
     x_glob(nx_dim+1,idim) = Lbox(idim)
     
     dx(idim) = (x_glob(nx_dim+1,idim) - x_glob(1,idim))/float(nx_dim)
     do i=2,nx_dim
        x_glob(i,idim) = x_glob(i-1,idim) + dx(idim)
     enddo

  enddo

  !  Local mesh
  !  ~~~~~~~~~~
  do idim = 1,3

     i1 = 1
     if (icl(2*idim-1) == 0) i1 = 1-nbuf
     i2 = nx(idim)+1
     if (icl(2*idim  ) == 0) i2 = nx(idim)+1+nbuf 
     do i=i1,i2
        x(i,idim) = x_glob(x_start(idim)+i-1,idim)
     enddo

  enddo

  do k=1,nx(3)
  do j=1,nx(2)
  do i=1,nx(1)
     if(idir==1) then
        ii=i
     else if(idir==2) then
        ii=j
     else if(idir==3) then
        ii=k
     endif
     xx = 0.5*(x(ii,idir)+x(ii+1,idir))

     if (xx < 0.5) then
        rhou(i,j,k,idir) = rho1*u1
        E   (i,j,k     ) = E1
        rho (i,j,k     ) = rho1
        !fx  (i,j,k,1   ) = 1
        !fx  (i,j,k,2   ) = 0
     else
        rhou(i,j,k,idir) = rho2*u2
        E   (i,j,k     ) = E2
        rho (i,j,k     ) = rho2
        !fx  (i,j,k,2   ) = 1
        !fx  (i,j,k,1   ) = 0
     endif
  enddo
  enddo
  enddo

!  write(*,*) 'Mach = ',u1/sqrt(gamma*P1/rho1)

  return
  
end subroutine INIT

!*******************************************************************************************************************

!     ---------------------
subroutine DIMENSIONS
!     ---------------------
  use unites
  use gammas

  implicit none

  ! code units
  unitM = 1.0E+00 *gramme
  unitL = 1.0E+00 *centimetre
  unitT = 1.0E+00 *seconde
  unitD = unitM/unitL**3

  ! Specific heats ratio Cp/Cv
  gamma = 7.0/5.0
  
  return
  
end subroutine DIMENSIONS
