import numpy as np
from numpy import cos, sin, pi
from enthought.tvtk.api import tvtk
from enthought.mayavi.scripts import mayavi2

from data import rd_heracles

class Heracles3d:
    
    def __init__(self,idump=1,form='hdf',opt=[1,1,1,1,1,1,1],verbose=True,endian='@', header_prec='i',dir='.',extend='no'):
        
        try:
            self.get_geometry(idump,form,verbose,endian,header_prec,dir,extend)
        except:
            return
        try:
            self.get_data(idump,form,opt,verbose,endian,header_prec,dir,extend)
        except:
            return
        
    #Get the geomtry
    def get_geometry(self,idump=1,form='hdf',verbose=True,endian='@', header_prec='i',dir='.',extend='no'):
        data=rd_heracles.HeraclesData(idump,form,[1,0,0,0,0,0,0],verbose,endian,header_prec,dir)
    
        #Extend the mesh if extend
        if extend=='yes':
            data=extend_mesh(data)

        #Get the mesh dimensions
        nx = (data.nxglob,data.nyglob,data.nzglob)
        if data.ndim==1:
            print('ndim=1, not very useful to use Mayavi...')
            raise Exception
        elif data.ndim==2 and extend=='no':
            nx_plusone = (data.nxglob+1,data.nyglob+1,data.nzglob)
        elif data.ndim==3 or (data.ndim==2 and extend=='yes'):
            nx_plusone = (data.nxglob+1,data.nyglob+1,data.nzglob+1)
            
        #Project the mesh into the Cartesian coordinates system of Mayavi
        pts = generate_points(data.xg, data.yg, data.zg, nx_plusone, data.geometry, extend)
        self.sgrid = tvtk.StructuredGrid(dimensions=nx_plusone)
        self.sgrid.points = pts

        del(pts,data)

    #Get the data
    def get_data(self,idump=1,form='hdf',opt=[1,1,1,1,1,1,1],verbose=True,endian='@', header_prec='i',dir='.',extend='no'):
        data=rd_heracles.HeraclesData(idump,form,opt,verbose,endian,header_prec,dir)

        #Extend the mesh if extend
        if extend=='yes':
            data=extend_mesh(data)
            data=extend_data(data,opt)
    
        #Get the mesh dimensions
        nx = (data.nxglob,data.nyglob,data.nzglob)
        if data.ndim==1:
            print('ndim=1, not very useful to use Mayavi...')
            raise Exception
        elif data.ndim==2 and extend=='no':
            nx_plusone = (data.nxglob+1,data.nyglob+1,data.nzglob)
        elif data.ndim==3 or (data.ndim==2 and extend=='yes'):
            nx_plusone = (data.nxglob+1,data.nyglob+1,data.nzglob+1)
            
        #Fill the data
        count=0

        #Go in reverse order to make the variables appear in right order in Mayavi
        if opt[6]==1 and data.mhd:
            count=count+1
            proj=project_vector(data.xcg, data.ycg, data.zcg, nx, data.B, data.geometry, extend)
            s = np.empty([nx[0]*nx[1]*nx[2],3])
            for i in range(nx[0]):
                for j in range(nx[1]):
                    for k in range(nx[2]):
                        for ii in range(3):
                            s[i+nx[0]*j+nx[0]*nx[1]*k,ii]=proj[i,j,k,ii]
                    
            if count==1:
                self.sgrid.cell_data.vectors = s.copy()
                self.sgrid.cell_data.vectors.name = 'B'
            else:
                id=self.sgrid.cell_data.add_array(s.copy())
                self.sgrid.cell_data.get_array(id).name = 'B'
                self.sgrid.cell_data.set_active_vectors('B')

        if opt[5]==1 and data.diffu:
            count=count+1
            s = np.empty(nx[0]*nx[1]*nx[2])
            for i in range(nx[0]):
                for j in range(nx[1]):
                    for k in range(nx[2]):
                        s[i+nx[0]*j+nx[0]*nx[1]*k]=data.ed[i,j,k]

            if count==1:
                self.sgrid.cell_data.scalars = s.copy()
                self.sgrid.cell_data.scalars.name = 'Ed'
            else:
                id=self.sgrid.cell_data.add_array(s.copy())
                self.sgrid.cell_data.get_array(id).name = 'Ed'
                self.sgrid.cell_data.set_active_scalars('Ed')
                
        if opt[4]==1 and data.ray:
            count=count+1
            s = np.empty(nx[0]*nx[1]*nx[2])
            for i in range(nx[0]):
                for j in range(nx[1]):
                    for k in range(nx[2]):
                        s[i+nx[0]*j+nx[0]*nx[1]*k]=data.er[i,j,k]
                        
            if count==1:
                self.sgrid.cell_data.scalars = s.copy()
                self.sgrid.cell_data.scalars.name = 'Er'
            else:
                id=self.sgrid.cell_data.add_array(s.copy())
                self.sgrid.cell_data.get_array(id).name = 'Er'
                self.sgrid.cell_data.set_active_scalars('Er')

            proj=project_vector(data.xcg, data.ycg, data.zcg, nx, data.fr, data.geometry, extend)
            s = np.empty([nx[0]*nx[1]*nx[2],3])
            for i in range(nx[0]):
                for j in range(nx[1]):
                    for k in range(nx[2]):
                        for ii in range(3):
                            s[i+nx[0]*j+nx[0]*nx[1]*k,ii]=proj[i,j,k,ii]
                    
            if count==1:
                self.sgrid.cell_data.vectors = s.copy()
                self.sgrid.cell_data.vectors.name = 'Fr'
            else:
                id=self.sgrid.cell_data.add_array(s.copy())
                self.sgrid.cell_data.get_array(id).name = 'Fr'
                self.sgrid.cell_data.set_active_vectors('Fr')


        if opt[3]==1 and data.nfx>0:
            count=count+1
            for ifx in range(data.nfx):
                s = np.empty(nx[0]*nx[1]*nx[2])
                for i in range(nx[0]):
                    for j in range(nx[1]):
                        for k in range(nx[2]):
                            s[i+nx[0]*j+nx[0]*nx[1]*k]=data.fx[i,j,k,ifx]
                        
                if count==1:
                    self.sgrid.cell_data.scalars = s.copy()
                    self.sgrid.cell_data.scalars.name = 'fx('+str(ifx)+')'
                else:
                    id=self.sgrid.cell_data.add_array(s.copy())
                    self.sgrid.cell_data.get_array(id).name = 'fx('+str(ifx)+')'
                    self.sgrid.cell_data.set_active_scalars('fx('+str(ifx)+')')
                    
        if opt[2]==1:
            count=count+1
            s = np.empty(nx[0]*nx[1]*nx[2])
            for i in range(nx[0]):
                for j in range(nx[1]):
                    for k in range(nx[2]):
                        s[i+nx[0]*j+nx[0]*nx[1]*k]=data.E[i,j,k]
                                
            if count==1:
                self.sgrid.cell_data.scalars = s.copy()
                self.sgrid.cell_data.scalars.name = 'E'
            else:
                id=self.sgrid.cell_data.add_array(s.copy())
                self.sgrid.cell_data.get_array(id).name = 'E'
                self.sgrid.cell_data.set_active_scalars('E')
                
        if opt[1]==1:
            count=count+1
            proj=project_vector(data.xcg, data.ycg, data.zcg, nx, data.u, data.geometry, extend)
            s = np.empty([nx[0]*nx[1]*nx[2],3])
            for i in range(nx[0]):
                for j in range(nx[1]):
                    for k in range(nx[2]):
                        for ii in range(3):
                            s[i+nx[0]*j+nx[0]*nx[1]*k,ii]=proj[i,j,k,ii]
                            
            if count==1:
                self.sgrid.cell_data.vectors = s.copy()
                self.sgrid.cell_data.vectors.name = 'v'
            else:
                id=self.sgrid.cell_data.add_array(s.copy())
                self.sgrid.cell_data.get_array(id).name = 'v'
                self.sgrid.cell_data.set_active_vectors('v')
                
        if opt[0]==1:
            count=count+1
            s = np.empty(nx[0]*nx[1]*nx[2])
            for i in range(nx[0]):
                for j in range(nx[1]):
                    for k in range(nx[2]):
                        s[i+nx[0]*j+nx[0]*nx[1]*k]=data.rho[i,j,k]
                        
            if count==1:
                self.sgrid.cell_data.scalars = s.copy()
                self.sgrid.cell_data.scalars.name = 'rho'
            else:
                id=self.sgrid.cell_data.add_array(s.copy())
                self.sgrid.cell_data.get_array(id).name = 'rho'
                self.sgrid.cell_data.set_active_scalars('rho')
                
        del(s,data)
        if(opt[6]==1 or opt[4]==1 or opt[1]==1):
            del(proj)

    # View the data.
    @mayavi2.standalone
    def view(self):
        from enthought.mayavi.sources.vtk_data_source import VTKDataSource
        from enthought.mayavi.modules.api import Outline, GridPlane
    
        mayavi.new_scene()
        
        src = VTKDataSource(data=self.sgrid)
        mayavi.add_source(src)
        mayavi.add_module(Outline())
        

    def reload(self,idump=1,form='hdf',opt=[1,1,1,1,1,1,1],verbose=True,endian='@', header_prec='i',dir='.', extend='no'):
        try:
            self.get_data(idump,form,opt,verbose,endian,header_prec,dir,extend)
        except:
            return
        self.sgrid.modified()


#Project the mesh into the Cartesian coordinates system
def generate_points(x, y, z, nx, geometry, extend):
            
    points = np.empty([nx[0]*nx[1]*nx[2],3])

    if geometry == 'xy   ' and extend == 'no':
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    points[i+nx[0]*j+nx[0]*nx[1]*k,0]=x[i]
                    points[i+nx[0]*j+nx[0]*nx[1]*k,1]=y[j]
                    points[i+nx[0]*j+nx[0]*nx[1]*k,2]=0.
                    
    elif geometry == 'xyz  ' or (geometry == 'xy   ' and extend == 'yes'):
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    points[i+nx[0]*j+nx[0]*nx[1]*k,0]=x[i]
                    points[i+nx[0]*j+nx[0]*nx[1]*k,1]=y[j]
                    points[i+nx[0]*j+nx[0]*nx[1]*k,2]=z[k]
                    
    elif geometry == 'rz   ' and extend == 'no':
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    points[i+nx[0]*j+nx[0]*nx[1]*k,0]=x[i]
                    points[i+nx[0]*j+nx[0]*nx[1]*k,1]=y[j]
                    points[i+nx[0]*j+nx[0]*nx[1]*k,2]=0.
                    
    elif geometry == 'rp   ' and extend == 'no':
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    points[i+nx[0]*j+nx[0]*nx[1]*k,0]=x[i]*cos(y[j])
                    points[i+nx[0]*j+nx[0]*nx[1]*k,1]=x[i]*sin(y[j])
                    points[i+nx[0]*j+nx[0]*nx[1]*k,2]=0.
                    
    elif geometry == 'rpz  ' or ((geometry == 'rz   ' or geometry == 'rp   ') and extend == 'yes'):
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    points[i+nx[0]*j+nx[0]*nx[1]*k,0]=x[i]*cos(y[j])
                    points[i+nx[0]*j+nx[0]*nx[1]*k,1]=x[i]*sin(y[j])
                    points[i+nx[0]*j+nx[0]*nx[1]*k,2]=z[k]
                    
    elif geometry == 'rthe ' and extend == 'no':
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    points[i+nx[0]*j+nx[0]*nx[1]*k,0]=x[i]*sin(y[j])
                    points[i+nx[0]*j+nx[0]*nx[1]*k,1]=x[i]*cos(y[j])
                    points[i+nx[0]*j+nx[0]*nx[1]*k,2]=0.
    
    elif geometry == 'rtp  ' or (geometry == 'rthe ' and extend == 'yes'):
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    points[i+nx[0]*j+nx[0]*nx[1]*k,0]=x[i]*sin(y[j])*cos(z[k])
                    points[i+nx[0]*j+nx[0]*nx[1]*k,1]=x[i]*sin(y[j])*sin(z[k])
                    points[i+nx[0]*j+nx[0]*nx[1]*k,2]=x[i]*cos(y[j])
                    
    return points

#Project vector into the Cartesian coordinates system
def project_vector(x, y, z, nx, vector, geometry, extend):
    proj = np.empty([nx[0],nx[1],nx[2],3])
    
    if geometry == 'xy   ' and extend == 'no':
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    proj[i,j,k,0]=vector[i,j,k,0]
                    proj[i,j,k,1]=vector[i,j,k,1]
                    proj[i,j,k,2]=0.
                    
    elif geometry == 'xyz  ' or (geometry == 'xy   ' and extend == 'yes'):
        proj=vector

    elif geometry == 'rz   ' and extend == 'no':
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    proj[i,j,k,0]=vector[i,j,k,0]
                    proj[i,j,k,1]=vector[i,j,k,1]
                    proj[i,j,k,2]=0.
                    
    elif geometry == 'rp   ' and extend == 'no':
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    proj[i,j,k,0]=vector[i,j,k,0]*cos(y[j])-vector[i,j,k,1]*sin(y[j])
                    proj[i,j,k,1]=vector[i,j,k,0]*sin(y[j])+vector[i,j,k,1]*cos(y[j])
                    proj[i,j,k,2]=0.
                    
    elif geometry == 'rpz  ' or ((geometry == 'rz   ' or geometry == 'rp   ') and extend == 'yes'):
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    proj[i,j,k,0]=vector[i,j,k,0]*cos(y[j])-vector[i,j,k,1]*sin(y[j])
                    proj[i,j,k,1]=vector[i,j,k,0]*sin(y[j])+vector[i,j,k,1]*cos(y[j])
                    proj[i,j,k,2]=vector[i,j,k,2]
                    
    elif geometry == 'rthe ' and extend == 'no':
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    proj[i,j,k,0]=vector[i,j,k,0]*sin(y[j])+vector[i,j,k,1]*cos(y[j])
                    proj[i,j,k,1]=vector[i,j,k,0]*cos(y[j])-vector[i,j,k,1]*sin(y[j])
                    proj[i,j,k,2]=0.
    
    elif geometry == 'rtp  ' or (geometry == 'rthe ' and extend == 'yes'):
        for i in range(nx[0]):
            for j in range(nx[1]):
                for k in range(nx[2]):
                    proj[i,j,k,0]=vector[i,j,k,0]*sin(y[j])*cos(z[k])+vector[i,j,k,1]*cos(y[j])*cos(z[k])-vector[i,j,k,2]*sin(z[k])
                    proj[i,j,k,1]=vector[i,j,k,0]*sin(y[j])*sin(z[k])+vector[i,j,k,1]*cos(y[j])*sin(z[k])+vector[i,j,k,2]*cos(z[k])
                    proj[i,j,k,2]=vector[i,j,k,0]*cos(y[j])          -vector[i,j,k,1]*sin(y[j])
                    
    return proj
                


def extend_mesh(data):
    
    if data.geometry == 'xyz  ' or data.geometry == 'rpz  ' or data.geometry == 'rtp  ':
        return data

    elif data.geometry == 'xy   ':
        if data.nxglob <= data.nyglob:
            data.nzglob = data.nxglob

            data.zcg=data.xcg
            data.zg =data.xg
        else:
            data.nzglob = data.nyglob

            data.zcg=data.ycg
            data.zg =data.yg

    elif data.geometry == 'rz   ':
        data.nzglob = min(data.nxglob,data.nyglob)

        dx=2*np.pi/data.nzglob

        data.zcg=np.empty(data.nzglob)
        for i in range(data.nzglob):
            data.zcg[i]=(i+0.5)*dx

        data.zg=np.empty(data.nzglob+1)
        for i in range(data.nzglob+1):
            data.zg[i]=i*dx

        temp=data.nyglob
        data.nyglob=data.nzglob
        data.nzglob=temp

        temp=data.ycg
        data.ycg=data.zcg
        data.zcg=temp

        temp=data.yg
        data.yg=data.zg
        data.zg=temp

    elif data.geometry == 'rp   ':
        data.nzglob = data.nxglob

        data.zcg=data.xcg
        data.zg =data.xg

    elif data.geometry == 'rthe ':
        data.nzglob = min(data.nxglob,data.nyglob)

        dx=2*np.pi/data.nzglob

        data.zcg=np.empty(data.nzglob)
        for i in range(data.nzglob):
            data.zcg[i]=(i+0.5)*dx

        data.zg=np.empty(data.nzglob+1)
        for i in range(data.nzglob+1):
            data.zg[i]=i*dx
    else:
        print('Extend set to yes but geometry = '+data.geometry)
        print('Extend not yet implemented for this geometry')
        sys.exit()

    return data

def extend_data(data,opt):

    if data.geometry == 'xyz  ' or data.geometry == 'rpz  ' or data.geometry == 'rtp  ':
        return data

    elif data.geometry == 'rz   ':
        #Warning: data.nyglob and data.nzglob have already been switched in extend_geom
        imin=0  ; imax=data.nxglob

        def j1min(i):
            return i
        def j1max(i,nyglob,nzglob):
            return i+1
        def k1min(i):
            return 0
        def k1max(i,nyglob,nzglob):
            return nzglob

        j2min=0 ; j2max=data.nzglob
        k2min=0 ; k2max=1

        comp=[0,2,1]
        
        range_i = data.nyglob
    elif data.geometry == 'xy   ' or data.geometry == 'rp   ' or data.geometry == 'rthe ':
        imin=0  ; imax=data.nxglob

        def j1min(i):
            return 0
        def j1max(i,nyglob,nzglob):
            return nyglob
        def k1min(i):
            return i
        def k1max(i,nyglob,nzglob):
            return i+1

        j2min=0 ; j2max=data.nyglob
        k2min=0 ; k2max=1

        comp=[0,1,2]

        range_i = data.nzglob

    if opt[6]==1 and data.mhd:
        B=np.zeros([data.nxglob,data.nyglob,data.nzglob,3])
        for i in range(range_i):
            for idim in range(3):
                B[imin:imax,j1min(i):j1max(i,data.nyglob,data.nzglob),k1min(i):k1max(i,data.nyglob,data.nzglob),comp[idim]]=data.B[imin:imax,j2min:j2max,k2min:k2max,idim]
                
        data.B=B
        del(B)

    if opt[5]==1 and data.diffu:
        Ed=np.zeros([data.nxglob,data.nyglob,data.nzglob])
        for i in range(range_i):
            Ed[imin:imax,j1min(i):j1max(i,data.nyglob,data.nzglob),k1min(i):k1max(i,data.nyglob,data.nzglob)]=data.ed[imin:imax,j2min:j2max,k2min:k2max]
            
        data.ed=Ed
        del(Ed)

    if opt[4]==1 and data.ray:
        Er=np.zeros([data.nxglob,data.nyglob,data.nzglob])
        for i in range(range_i):
            Er[imin:imax,j1min(i):j1max(i,data.nyglob,data.nzglob),k1min(i):k1max(i,data.nyglob,data.nzglob)]=data.er[imin:imax,j2min:j2max,k2min:k2max]
            
        data.er=Er
        del(Er)

        Fr=np.zeros([data.nxglob,data.nyglob,data.nzglob,3])
        for i in range(range_i):
            for idim in range(2):
                Fr[imin:imax,j1min(i):j1max(i,data.nyglob,data.nzglob),k1min(i):k1max(i,data.nyglob,data.nzglob),comp[idim]]=data.fr[imin:imax,j2min:j2max,k2min:k2max,idim]

        data.fr=Fr
        del(Fr)

    if opt[3]==1 and data.nfx>0:
        fx=np.zeros([data.nxglob,data.nyglob,data.nzglob,nfx])
        for i in range(range_i):
            fx[imin:imax,j1min(i):j1max(i,data.nyglob,data.nzglob),k1min(i):k1max(i,data.nyglob,data.nzglob),:]=data.fx[imin:imax,j2min:j2max,k2min:k2max,:]
            
        data.fx=fx
        del(fx)

    if opt[2]==1:
        E=np.zeros([data.nxglob,data.nyglob,data.nzglob])
        for i in range(range_i):
            E[imin:imax,j1min(i):j1max(i,data.nyglob,data.nzglob),k1min(i):k1max(i,data.nyglob,data.nzglob)]=data.E[imin:imax,j2min:j2max,k2min:k2max]

        data.E=E
        del(E)

    if opt[1]==1:
        u=np.zeros([data.nxglob,data.nyglob,data.nzglob,3])
        for i in range(range_i):
            for idim in range(2):
                u[imin:imax,j1min(i):j1max(i,data.nyglob,data.nzglob),k1min(i):k1max(i,data.nyglob,data.nzglob),comp[idim]]=data.u[imin:imax,j2min:j2max,k2min:k2max,idim]

        data.u=u
        del(u)
        
    if opt[0]==1:
        r=np.empty([data.nxglob,data.nyglob,data.nzglob])
        for i in range(range_i):
            r[imin:imax,j1min(i):j1max(i,data.nyglob,data.nzglob),k1min(i):k1max(i,data.nyglob,data.nzglob)]=data.rho[imin:imax,j2min:j2max,k2min:k2max]
        
        data.rho=r
        del(r)

    return data
